with Ada.Text_IO, Strings, Ints, Flts, Globals, Reps, Params, ScalVectors;
use Ada.Text_IO, Strings, Ints, Flts, Globals, Reps, Params;
with Params.Names,Matrices,Vectors,Cylinders,Cylinders.Fix;

pragma Elaborate_All (Ada.Text_IO,Strings,Ints,Flts,Globals,Reps,ScalVectors);
pragma Elaborate_All (Cylinders,Cylinders.Fix);

procedure Approx_FixPt(P: in Param) is

   package PN is new Params.Names(P => P, Trunc => True);
   use PN;

   subtype Scalar is Rep;

   package SM is new Matrices (Scalar => Scalar);
   package SV is new Vectors (Component => Scalar);
   package SSV is new ScalVectors(Scalar => Scalar, SM => SM, SV => SV);

   package Cyls is new Cylinders(DegX => DegX, DegY => DegY, LDeg => LDeg, WFac => WFac, RadX => RadX,
                                 EllR => EllR, Scalar => Scalar, SM => SM, SV => SV, Ref => Ref, Symm => Symm);
   package Fix is new Cyls.Fix(Mat_PackSize => Mat_PackSize, NN => NN, RR => RR, Map => Map, SSV => SSV);

   use Cyls,Fix,Cyls.Cyl;
   use Flt_EF;

   Epsilon: constant Radius := 1.0e-15;
   F: Cylinder;
   M: QC_Matrix;
   Pi: Flt := 3.141592653;

begin

   Verbosity := 2;
   Show("",P);

   Read(Input_Dir & Fix_Name,F);
   --  Read(Input_Dir & "NCS2-fix.approx",F);
   --  Copy(F,G1);
   --  Rotate(Pi,G1,G2);
   --  Sub(G2,F);
   --  Rotate(Pi,G1,G2);
   --  Add(G2,F);
   --  Rotate(-Pi/2.0,G1,G2);
   --  Sub(G2,F);
   --  Rotate(-Pi/6.0,G1,G2);
   --  Add(G2,F);
   --  Rotate(Pi/2.0,G1,G2);
   --  Add(G2,F);
   --  Rotate(Pi/3.0,G1,G2);
   --  Sub(G2,F);
   --  Rotate(Pi,G1,G2);
   --  Sub(G2,F);
   --  Rotate(-Pi/3.0,G1,G2);
   --  Sub(G2,F);

   if Ref=1 then ProjEven(F); end if;
   if Ref=(-1) then ProjOdd(F); end if;
   if Symm>1 then ProjSymm(Symm,F); end if;
   if Symm<0 then ProjASymm(-Symm,F); end if;
   if Symm=0 then ProjConstant(F); end if;
   --  Read(Input_Dir & Eigen_Name,G);
   --  Gmap(F,G1);
   --  Sub(F,G1);
   --  Show1("diff",MaxNorm(G1));
   --  Plot3(F,Pic_Dir & Plot_Name);
   --  Show1("norm",MaxNorm(f));
   --  Write(Output_Dir & Fix_Name,F);

   ApproxFix(F,M,Epsilon);
   if Symm<0 then ProjASymm(-Symm,F); end if;
   if Ref=(-1) then ProjOdd(F); end if;

   --  for K in 1..100 loop
   --     Read(Input_Dir & Eigen_Name,G);
   --     Mult(Flt(K)*0.0025,G);
   --     Sum(F,G,H);
   --     Gmap(H,G);
   --     Sub(H,G);
   --     Show1("k",K);
   --     Show1("Norm F",MaxNorm(G));
   --  end loop;

   Show1("Norm F",MaxNorm(F));
   Plot3(F,Pic_Dir & Plot_Name);
   Write(Output_Dir & Fix_Name,F);
   SM.Write(Input_Dir & Mat_Name,M);

end Approx_FixPt;
