with Ada.Text_IO, Strings, Ints, Flts, Globals, Reps, Params, ScalVectors.Eigenv.RNum;
use Ada.Text_IO, Strings, Ints, Flts, Globals, Reps, Params;
with Params.Names,Matrices,Vectors,Cylinders,Cylinders.Fix;

pragma Elaborate_All (Ada.Text_IO,Strings,Ints,Flts,Globals,Reps,ScalVectors.Eigenv.RNum);
pragma Elaborate_All (Cylinders,Cylinders.Fix);

procedure Approx_Morse(P: in Param) is

   package PN is new Params.Names(P => P, Trunc => True);
   use PN;

   subtype Scalar is Rep;

   package SM is new Matrices (Scalar => Scalar);
   package SV is new Vectors (Component => Scalar);
   package SSV is new ScalVectors(Scalar => Scalar, SM => SM, SV => SV);

   package Cyls is new Cylinders(DegX => DegX, DegY => DegY, LDeg => LDeg, WFac => WFac, RadX => RadX,
                                 EllR => EllR, Scalar => Scalar, SM => SM, SV => SV, Ref => Ref,
                                 Symm => Symm, Full => True);
   package Fix is new Cyls.Fix(Mat_PackSize => Mat_PackSize, NN => NN, RR => RR, Map => Map, SSV => SSV);

   use Cyls,Fix,Cyls.Cyl;
   use Flt_EF;

   Epsilon: constant Radius := 1.0e-15;
   F: Cylinder;
   A,Vr,Vi: QC_Matrix;
   Wr,Wi: QC_Vector;
   Count: Natural := 0;
   package SVE is new SSV.Eigenv;
   package SVEN is new SVE.RNum;
   use SV,SSV,SVEN;

begin

   Verbosity := 2;
   Show("",P);

   Read(Input_Dir & Fix_Name,F);
   if Ref=1 then ProjEven(F); end if;
   if Ref=(-1) then ProjOdd(F); end if;
   if Symm>1 then ProjSymm(Symm,F); end if;
   if Symm<=-1 then ProjASymm(-Symm,F); end if;
   if Symm=0 then ProjConstant(F); end if;
   DerMatrix(F,A);
   Eigen(A,Wr,Wi,Vr,Vi);
   for I in 1..IMin(WR'Last,50) loop
      if Approx(Wr(I))>One then
         Count := Count+1;
         Show2("ev:",Approx(Wr(I)),Approx(Wi(I)));
      end if;
   end loop;
   Show1("Morse: ",Count);

end Approx_Morse;
