with Strings;
use Strings;

pragma Elaborate_All (Strings);

package body BigRats is

  function IsOne(B: BigRat) return Boolean is
  begin
    for I in 1 .. Size loop
      if B(I) /= 0 then return False; end if;
    end loop;
    return True;
  end IsOne;

  function Num(B: BigRat) return BigRat is
    BNum: BigRat;
  begin
    for I in 1 .. Size loop
      BNum(I) := Integer'Max(0,B(I));
    end loop;
    return BNum;
  end Num;

  function Den(B: BigRat) return BigRat is
    BDen: BigRat;
  begin
    for I in 1 .. Size loop
      BDen(I) := Integer'Max(0,-B(I));
    end loop;
    return BDen;
  end Den;

  procedure GCD(B1: in BigRat; B2: in out BigRat) is
  begin
    for I in 1 .. Size loop
      B2(I) := Integer'Min(B1(I),B2(I));
    end loop;
  end GCD;


  procedure GCD(B: in BigRat_Vec; BG: out BigRat) is
  begin
    BG := B(B'First);
    for K in B'First+1 .. B'Last loop
      GCD(B(K),BG);
    end loop;
  end GCD;

  procedure Mult(B1: in BigRat; B2: in out BigRat) is
  begin
    for I in 1 .. Size loop
      B2(I) := B2(I)+B1(I);
    end loop;
  end Mult;

  procedure Prod(B1,B2: in BigRat; B3: out BigRat) is
  begin
    for I in 1 .. Size loop
      B3(I) := B1(I)+B2(I);
    end loop;
  end Prod;

  procedure Inv(B: in out BigRat) is
  begin
    for I in 1 .. Size loop
      B(I) := -B(I);
    end loop;
  end Inv;

  procedure Inv(B1: in BigRat; B2: out BigRat) is
  begin
    for I in 1 .. Size loop
      B2(I) := -B1(I);
    end loop;
  end Inv;

  procedure Div(B1: in BigRat; B2: in out BigRat) is
  begin
    for I in 1 .. Size loop
      B2(I) := B2(I)-B1(I);
    end loop;
  end Div;

  procedure Quot(B1,B2: in BigRat; B3: out BigRat) is
  begin
    for I in 1 .. Size loop
      B3(I) := B1(I)-B2(I);
    end loop;
  end Quot;

  procedure Power(K: in Integer; B: in out BigRat) is
  begin
    for I in 1 .. Size loop
      B(I) := K*B(I);
    end loop;
  end Power;

  procedure Power(K: in Integer; B1: in BigRat; B2: out BigRat) is
  begin
    for I in 1 .. Size loop
      B2(I) := K*B1(I);
    end loop;
  end Power;

  procedure Sqrt(B: in out BigRat; B_Rem: out BigRat) is
    --- B_Rem is a positive integer
  begin
    for I in 1 .. Size loop
      B_Rem(I) := B(I) mod 2;
      B(I) := (B(I)-B_Rem(I))/2;
    end loop;
  end Sqrt;

  procedure DivGCD(B: in out BigRat_Vec; BG: out BigRat) is
  begin
    GCD(B,BG);
    for K in B'Range loop
      Div(BG,B(K));
    end loop;
  end DivGCD;

  -----------------

  procedure Assign(N: in Positive; B: out BigRat) is
  begin
    for I in 1 .. Size loop B(I) := 0; end loop;
    if N>1 then
      declare
        M: Positive := N;
        I: Positive;
      begin
        while M>1 loop
          I := PrimeFactorIndex(M);
          B(I) := B(I)+1;
          M := M/Prime(I);
        end loop;
      end;
    end if;
  end Assign;

  procedure Show1(N: in String; B: in BigRat; NewLine: in Boolean := True) is
  begin
    Show0(N,False);
    if IsOne(B) then Show0("1"); return; end if;
    for I in 1 .. Size loop
      if B(I) /= 0 then
        Show0(Strng(Prime(I)) & "^" & Strng(B(I)) & "  ",False);
      end if;
    end loop;
    Show0(NewLine);
  end Show1;

  function Approx(B: BigRat) return Long_Long_Float is
    subtype Real is Long_Long_Float;
    R: Real := 1.0;
  begin
    for I in 1 .. Size loop
      if B(I) /= 0 then R := R*Real(Prime(I))**B(I); end if;
    end loop;
    return R;
  end Approx;

  function Make_Factorials return Factorial_Type is
    B: BigRat;
    Fac: Factorial_Type;
  begin
    Assign(1,Fac(0));
    Assign(1,Fac(1));
    Assign(2,Fac(2));
    for N in 3 .. NMax loop
      Assign(N,B);
      Prod(Fac(N-1),B,Fac(N));
    end loop;
    return Fac;
  end Make_Factorials;

end BigRats;
