with Ada.Text_IO, Strings, Ints, Flts, Globals, Reps.Balls, Params;
use Ada.Text_IO, Strings, Ints, Flts, Globals, Reps.Balls, Params;
with Params.Names, Matrices, Vectors,Cylinders;

pragma Elaborate_All (Ada.Text_IO,Strings,Ints,Flts,Globals,Reps.Balls);
pragma Elaborate_All (Params.Names,Matrices,Vectors,Cylinders);

procedure Check_Misc(P: in Param) is

  package PN is new Params.Names(P => P, Trunc => False);
  use PN;

  subtype Scalar is Ball;

  package SM is new Matrices (Scalar => Scalar);
  package SV is new Vectors (Component => Scalar);
  package Cyls is new Cylinders(DegX => DegX, DegY => DegY, LDeg => LDeg, WFac => WFac, RadX => RadX,
                      EllR => EllR, Scalar => Scalar, SM => SM, SV => SV, Ref => Ref, Symm => Symm);
  use Cyls,Cyls.Cyl;

  C: constant Cylinder := Read(Input_Dir & Fix_Name);
  VMin: constant FltVec := Read(Input_Dir & Prefix & "-min.approx"); -- approx local minima
  VMax: constant FltVec := Read(Input_Dir & Prefix & "-max.approx"); -- approx local maxima
  NMin: constant Natural := VMin'Length/3;                           -- number of approx local minima
  NMax: constant Natural := VMax'Length/3;                           -- number of approx local maxima

  ApproxPi:  constant := 3.14159_26535_89793_23846_26433_83279_50288_41971_69399_37511;
  ApproxPi2: constant Flt := Two*ApproxPi;
  Pi1:       constant Scalar := Pi;

  procedure Horizontal_Values is
    --- range of C at X=0 and X=Pi
    Steps: constant Positive := 2**16;
    MinInf,MaxInf,MinSup,MaxSup: Flt;
  begin
    Show0("=============================== Horizontal_Values");
    JValAt0(C,Zero,EllR,MinInf,MaxInf,MinSup,MaxSup,Steps);
    Show2("bounds on MIN at X=0  ",MinInf,MinSup);
    Show2("bounds on MAX at X=0  ",MaxInf,MaxSup);

    JValAtPi(C,Zero,EllR,MinInf,MaxInf,MinSup,MaxSup,Steps);
    Show2("bounds on MIN at X=Pi  ",MinInf,MinSup);
    Show2("bounds on MAX at X=Pi  ",MaxInf,MaxSup);
  end Horizontal_Values;

  procedure VMin_Point(I: in Positive; X,Y,Z: out Flt) is
    --- I-th point in VMin
  begin
    X := VMin(3*I-2);
    Y := VMin(3*I-1);
    Z := VMin(3*I);
  end VMin_Point;

  procedure VMax_Point(I: in Positive; X,Y,Z: out Flt) is
    --- I-th point in VMax
  begin
    X := VMax(3*I-2);
    Y := VMax(3*I-1);
    Z := VMax(3*I);
  end VMax_Point;

  function InfMax return Flt is
    --- Inf of true values at VMax points
    XC,YC,ZC: Flt;
    MinVal: Flt := 1.0E+999;
    X,Y,Z: Scalar;
  begin
    for I in 1 .. NMax loop
      VMax_Point(I,XC,YC,ZC);
      Assign(XC,X);
      Assign(YC,Y);
      Val(C,X,Y,Z);
      MinVal := RMin(MinVal,Inf(Z));
    end loop;
    return MinVal;
  end InfMax;

  function Segment(X: in Flt) return Integer is
    --- determine segment (0..11) containing X
    PiOver12: constant Scalar := Pi1/Flt(12);  --- 2Pi/24
    KMin: constant Integer := -8;
    KMax: constant Integer := 13;
    KR: Integer := KMin;
    KL: Integer := KMax;
    S: Scalar;
  begin
    loop
      Prod(Flt(2*KL-1),PiOver12,S);      -- left endpoint of interval # KL
      exit when X >= Sup(S);             -- X in this interval
      KL := KL-1;
      if KL<KMin then return -1; end if; -- should not happen
    end loop;
    if KL=KMax then
      Show1("out of range ",X);
      return -1;
    end if;
    while KL>11 loop KL := KL-12; end loop;
    while KL<0 loop KL := KL+12; end loop;
    ---
    loop
      Prod(Flt(2*KR+1),PiOver12,S);      -- right endpoint of interval # KR
      exit when X <= Inf(S);             -- X in this interval
      KR := KR+1;
      if KR>KMax then return -1; end if; -- should not happen
    end loop;
    if KR=KMin then
      Show1("out of range ",X);
      return -1;
    end if;
    while KR>11 loop KR := KR-12; end loop;
    while KR<0 loop KR := KR+12; end loop;
    ---
    if KL=KR then return KL; end if;
    return -1;
  end Segment;

  IL: Positive  := 2**24;  -- approx steps in X direction
  JL: Positive  := 2**16;  -- approx steps in Y direction
  NDX: Positive := 256;    -- DX := ApproxPi2/Flt(NDX)
  NDY: Positive := 64;     -- DY := EllR/Flt(NDY)

  procedure Check_Min(I: in Positive) is
    SI: constant String := Strng(I,2);
    DX: constant Radius := ApproxPi2/Flt(NDX);
    DY: constant Radius := EllR/Flt(NDY);
    KL,KR: Integer;
    XC,XL,XR,YC,YL,YR,ZC: Flt;
  begin
    if NMin=0 then
      Show0("=============================== Check_Min none");
    else
      Show0("=============================== Check_Min " & SI);
      VMin_Point(I,XC,YC,ZC);
      XL := -(DX-XC);
      XR := XC+DX;
      YL := -(DY-YC);
      YR := YC+DY;
      CheckMin(C,XC,XL,XR,YC,YL,YR,IL,JL);

      Show0("------------------------------- checking segment");
      KL := Segment(XL);
      KR := Segment(XR);
      if KL<0 or KR<0 then
        if KL<0 then
          Show1("Error: segment failed. XL =",XL);
        end if;
        if KR<0 then
          Show1("Error: segment failed. XR =",XR);
        end if;
      elsif KL/=KR then
        Show2("Error: segments " & Strng(KL) & "," & Strng(KR) & " are different for",XL,XR);
      else
        Show1("OK: Min " & SI & " in segment ",KL);
      end if;
    end if;
  end Check_Min;

  procedure Check_Min is
  begin
    if NMin=0 then
      Show0("=============================== Check_Min none");
    else
      Show1("=============================== Check_Min 1 ..",NMin);
      for I in 1 .. NMin loop Check_Min(I); end loop;
    end if;
  end Check_Min;

  procedure Check_Max(I: in Positive) is
    SI: constant String := Strng(I,2);
    DX: constant Radius := ApproxPi2/Flt(NDX);
    DY: constant Radius := EllR/Flt(NDY);
    KL,KR: Integer;
    XC,XL,XR,YC,YL,YR,ZC: Flt;
  begin
    if NMax=0 then
      Show0("=============================== Check_Max none");
    else
      Show0("=============================== Check_Max " & SI);
      VMax_Point(I,XC,YC,ZC);
      XL := -(DX-XC);
      XR := XC+DX;
      YL := -(DY-YC);
      YR := YC+DY;
      CheckMax(C,XC,XL,XR,YC,YL,YR,IL,JL);

      Show0("------------------------------- checking segment");
      KL := Segment(XL);
      KR := Segment(XR);
      if KL<0 or KR<0 then
        if KL<0 then
          Show1("Error: segment failed. XL =",XL);
        end if;
        if KR<0 then
          Show1("Error: segment failed. XR =",XR);
        end if;
      elsif KL/=KR then
        Show2("Error: segments",KL,KR,False);
        Show2("are different for ",XL,XR);
      else
        Show1("OK: Max " & SI & " in segment ",KL);
      end if;
    end if;
  end Check_Max;

  procedure Check_Max is
  begin
    if NMax=0 then
      Show0("=============================== Check_Max none");
    else
      Show1("=============================== Check_Max 1 ..",NMax);
      for I in 1 .. NMax loop Check_Max(I); end loop;
    end if;
  end Check_Max;

  procedure Level_Stat is
    Steps: constant Integer := 2**18;
    Eps: constant Flt := Half**6;
    XC,YC,ZC,Cut: Flt;
    X,Y,Z: Scalar;
  begin
    if NMax=0 then
      Show0("=============================== Level_Stat max none");
    else
      Show1("=============================== Level_Stat max 1 ..",NMax);
      VMax_Point(1,XC,YC,ZC);
      Assign(XC,X);
      Assign(YC,Y);
      Val(C,X,Y,Z);
      Cut := Inf(Z)-Eps;
      LevelStat(C,YC,Cut,Steps);
    end if;
  end Level_Stat;

  procedure Check_Injective is
    --- needed only for near-Cassini
  begin
    if Map=6 then
      Show0("=============================== Check_Injective");
      Near_Cassini_Injective(RR);
    end if;
  end Check_Injective;

begin
  NTasks := 6;
  Verbosity := 2;
  Show("",P);
  Show2("        IL NDX: ",IL,NDX);
  Show2("        JL NDY: ",IL,NDY);

  Horizontal_Values; -- just extra information
  Check_Min;
  Check_Max;
  Level_Stat;
  Check_Injective;

end Check_Misc;
