with Primes;
use Primes;

pragma Elaborate_All (Primes);

package body Comb is

  procedure Int_Mult(B: in BigRat; S,Tmp3,Tmp4,Tmp5: in out Scalar) is
    --- assuming all B(I) >= 0
    SPrime: Scalar renames Tmp3;
    SPower: Scalar renames Tmp4;
    R: Flt;
  begin
    Mult(Two**B(1),S,Tmp5);
    for I in 2 .. Size loop
      R := Flt(Prime(I));
      if B(I)>1 then
        R := R**B(I);
        if R<IntegerLast then -- result is exact
          Mult(R,S,Tmp5);
        else
          Assign(Prime(I),SPrime);
          IPower(B(I),SPrime,SPower,Tmp5);
          Mult(SPower,S);
        end if;
      elsif B(I)=1 then
        Mult(R,S,Tmp5);
      end if;
    end loop;
  end Int_Mult;

  procedure AbsChoose(P: in Integer; Q: in Positive; C: in out Scalar_Vector) is
    --- |{P/Q choose N}|
  begin
    if C'First /= 0 then raise Be_Serious; end if;
    Assign(1,C(0));
    if C'Last>0 then
      declare
        QplusP: constant Integer := Q+P;
        INum,IDen: Integer;
        SNum,SDen,Tmp3,Tmp4,Tmp5: Scalar;
        R,Rat: BigRat;
      begin
        Assign(1,Rat);
        for N in 1 .. C'Last loop
          IDen := Q*N;
          INum := abs(IDen-QPLusP);
          Assign(INum,R);
          Mult(R,Rat);                            -- Rat := Rat*abs(Q*N-Q-P)
          Assign(IDen,R);
          Div(R,Rat);                             -- Rat := Rat/(Q*N)

          Assign(1,SNum);
          Int_Mult(Num(Rat),SNum,Tmp3,Tmp4,Tmp5); -- SNum := Num(Rat)
          Assign(1,SDen);
          Int_Mult(Den(Rat),SDen,Tmp3,Tmp4,Tmp5); -- SDen := Den(Rat)
          Quot(SNum,SDen,C(N));                   -- C(I) := SNum/SDen
        end loop;
      end;
    end if;
  end AbsChoose;

  function AbsChoose(P: Integer; Q: Positive; LMax: Natural) return Scalar_Vector is
    C: Scalar_Vector(0 .. LMax);
  begin
    AbsChoose(P,Q,C);
    return C;
  end AbsChoose;

end Comb;
