with Strings, Globals, Flt_Vectors, Protected_Counters;
with Rounding, Rounding.FPU;
use Strings, Globals, Flt_Vectors, Protected_Counters;

pragma Elaborate_All (Strings,Globals,Flt_Vectors,Protected_Counters,Rounding.FPU);

package body CpCosSins1 is

  procedure Show_Cp_Rounding is
    use Rounding,Rounding.FPU;
    M: constant Rounding_Mode := Get_Rounding_Mode;
    Prec: constant Positive := Get_Precision(SHalf/Three);
  begin
    Show1("Cp:     Precision = ",Prec);        --- set precision before elaboration
    Show0("Cp: Rounding_Mode =  " & Strng(M));
  end Show_Cp_Rounding;

  procedure Show_Cp_Param(Full: in Boolean := True) is
  begin
    Show1("Cp:          FDeg = ",FDeg);
    Show1("Cp:          WFac = ",WFac);
    Show1("Cp:          WPow = ",WPow);
    Show1("DCp: efault_Pariy = ",Default_Parity);
    if Full then
      Show0("Cp:        RepType =  " & SInfo.RepType);
      Show1("Cp:         STrunc =  ",STrunc);
      Show1("Cp:         NTasks = ",NTasks);
      Show_Cp_Rounding;
    end if;
  end Show_Cp_Param;

  function Info(Dummy: CpCosSin1) return Scalar_Info is
    pragma Unreferenced(Dummy);
    PInfo: Scalar_Info := SInfo;
  begin
    PInfo.NReps := SInfo.NReps*(FDeg+1);
    return PInfo;
  end Info;

  function Get_Precision(P: CpCosSin1) return Positive is
  begin
    return Get_Precision(P.C(1));
  end Get_Precision;

  ---------------------------

  function QuasiDeg(P: Parity; C: FPoly1) return Natural is
  begin
    for I in reverse P .. FDeg loop
      if not IsZero(C(I)) then return I; end if;
    end loop;
    return 0;
  end QuasiDeg;

  function QuasiDeg(C: FPoly1) return Natural is
  begin
    for D in reverse 1 .. FDeg loop
      if not IsZero(C(D)) then return D; end if;
    end loop;
    return 0;
  end QuasiDeg;

  function QuasiDeg1(C: FPoly1) return Natural is
  begin
    for I in reverse 2 .. FDeg loop
      if not IsZero(C(I)) then return I; end if;
    end loop;
    return 1;
  end QuasiDeg1;

  function QuasiDeg(E: Errs1) return Natural is
  begin
    if Not_STrunc then
      for I in reverse 1 .. EDeg loop
        if E(I) /= Zero then return I; end if;
      end loop;
    end if;
    return 0;
  end QuasiDeg;

  function Is_Zero(P: Parity; C: FPoly1) return Boolean is
  begin
    for I in P .. FDeg loop
      if not IsZero(C(I)) then return False; end if;
    end loop;
    return True;
  end Is_Zero;

  function Is_Zero(E: Errs1) return Boolean is
  begin
    if Not_STrunc then
      for I in EFreq loop
        if E(I) /= Zero then return False; end if;
      end loop;
    end if;
    return True;
  end Is_Zero;

  procedure LiftE0(P: in out CpCosSin1) is
    --- use only if Not_STrunc
    E: Errs1 renames P.E;
  begin
    if (E(0) /= Zero) then
      E(1) := E(1)+E(0);
      E(0) := Zero;
    end if;
  end LiftE0;

  ----------------------------------

  function IsZero(P: CpCosSin1) return Boolean is
    C: FPoly1 renames P.C;
  begin
    for I in P.P .. FDeg loop
      if not IsZero(C(I)) then return False; end if;
    end loop;
    if Not_STrunc then
      declare
        E: Errs1 renames P.E;
      begin
        for I in EFreq loop
          if E(I) /= Zero then return False; end if;
        end loop;
      end;
    end if;
    return True;
  end IsZero;

  function IsSharp(P: CpCosSin1) return Boolean is
  begin
    if Not_STrunc then
      declare
        C: FPoly1 renames P.C;
        E: Errs1 renames P.E;
      begin
        for I in P.P .. FDeg loop
          if not IsSharp(C(I)) then return False; end if;
        end loop;
        for I in EFreq loop
          if E(I) /= Zero then return False; end if;
        end loop;
      end;
    end if;
    return True;
  end IsSharp;

  function IsPoly(P: CpCosSin1) return Boolean is
  begin
    if Not_STrunc then
      declare
        E: Errs1 renames P.E;
      begin
        for I in EFreq loop
          if E(I) /= Zero then return False; end if;
        end loop;
      end;
    end if;
    return True;
  end IsPoly;

  function "="(P1,P2: CpCosSin1) return Boolean is
    C1: FPoly1 renames P1.C;
    C2: FPoly1 renames P2.C;
  begin
    if P1.P /= P2.P then return False; end if;
    for I in P1.P .. FDeg loop
      if C1(I) /= C2(I) then return False; end if;
    end loop;
    if not STrunc then
      declare
        E1: Errs1 renames P1.E;
        E2: Errs1 renames P2.E;
      begin
        for I in EFreq loop
          if E1(I) /= E2(I) then return False; end if;
        end loop;
      end;
    end if;
    return True;
  end "=";

  procedure FillZero(P: in out CpCosSin1) is
    C: FPoly1 renames P.C;
  begin
    for I in FFreq loop SetZero(C(I)); end loop;
    if Not_STrunc then P.E := (others => Zero); end if;
  end FillZero;

  procedure SetZero(Par: in Parity; P: in out CpCosSin1) is
    C: FPoly1 renames P.C;
  begin
    P.P := Par;
    for I in FFreq loop SetZero(C(I)); end loop;
    if Not_STrunc then P.E := (others => Zero); end if;
  end SetZero;

  procedure SetZero(P: in out CpCosSin1) is
    C: FPoly1 renames P.C;
  begin
    P.P := Default_Parity;
    for I in FFreq loop SetZero(C(I)); end loop;
    if Not_STrunc then P.E := (others => Zero); end if;
  end SetZero;

  procedure Copy(P1: in CpCosSin1; P2: in out CpCosSin1) is
    C1: FPoly1 renames P1.C;
    C2: FPoly1 renames P2.C;
  begin
    P2.P := P1.P;
    for I in P1.P .. FDeg loop Copy(C1(I),C2(I)); end loop;
    if Not_STrunc then P2.E := P1.E; end if;
  end Copy;

  procedure Swap(P1,P2: in out CpCosSin1) is
    C1: FPoly1 renames P1.C;
    C2: FPoly1 renames P2.C;
  begin
    if P1.P /= P2.P then raise Not_Implemented; end if;
    for I in P1.P .. FDeg loop Swap(C1(I),C2(I)); end loop;
    if Not_STrunc then
      declare
        E1: Errs1 renames P1.E;
        E2: Errs1 renames P2.E;
        R: Flt;
      begin
        for I in EFreq loop
          R := E1(I); E1(I) := E2(I); E2(I) := R;
        end loop;
      end;
    end if;
  end Swap;

  ----------------------------------

  function Center0(P: CpCosSin1) return Boolean is
    --- symmetric about zero
    C: FPoly1 renames P.C;
  begin
    for I in P.P .. FDeg loop
      if not Center0(C(I)) then return False; end if;
    end loop;
    return True;
  end Center0;

  function Contains0(P: CpCosSin1) return Logical is
    --- very crude
    C: FPoly1 renames P.C;
    L: Logical := True;
    LI: Logical;
  begin
    for I in P.P .. FDeg loop
      LI := Contains0(C(I));
      if LI=False then return False; end if;
      if LI=Uncertain then L := Uncertain; end if;
    end loop;
    return L;
  end Contains0;

  procedure BallAt0(R: in Flt; Par: in Parity; P: in out CpCosSin1) is
  begin
    SetZero(Par,P);
    if Not_STrunc then
      P.E(0) := abs(R);
      if P.P=1 then LiftE0(P); end if;
    end if;
  end BallAt0;

  procedure BallAt0(R: in Flt; P: in out CpCosSin1) is
  begin
    BallAt0(R,Default_Parity,P);
  end BallAt0;

  function BallAt0(R: Flt) return CpCosSin1 is
    P: CpCosSin1;
  begin
    BallAt0(R,P);
    return P;
  end BallAt0;

  procedure ToErr(P: in out CpCosSin1) is
    --- zero-centered set containing P
    C: FPoly1 renames P.C;
  begin
    for I in P.P .. FDeg loop ToErr(C(I)); end loop;
  end ToErr;

  procedure ToErr(P1: in CpCosSin1; P2: in out CpCosSin1) is
    --- zero-centered set containing P1
    C1: FPoly1 renames P1.C;
    C2: FPoly1 renames P2.C;
  begin
    P2.P := P1.P;
    for I in P1.P .. FDeg loop ToErr(C1(I),C2(I)); end loop;
    if Not_STrunc then P2.E := P1.E; end if;
  end ToErr;

  function ToErr(P: CpCosSin1) return CpCosSin1 is
    --- zero-centered set containing P
    Q: CpCosSin1;
  begin
    ToErr(P,Q);
    return Q;
  end ToErr;

  procedure Center(P: in out CpCosSin1) is
    C: FPoly1 renames P.C;
  begin
    if Not_STrunc then
      for I in P.P .. FDeg loop Center(C(I)); end loop;
      P.E := (others => Zero);
    end if;
  end Center;

  procedure Center(P1: in CpCosSin1; P2: in out CpCosSin1) is
    C1: FPoly1 renames P1.C;
    C2: FPoly1 renames P2.C;
  begin
    P2.P := P1.P;
    for I in P1.P .. FDeg loop Center(C1(I),C2(I)); end loop;
    if Not_STrunc then P2.E := (others => Zero); end if;
  end Center;

  function Center(P: CpCosSin1) return CpCosSin1 is
    Q: CpCosSin1;
  begin
    Center(P,Q);
    return Q;
  end Center;

  procedure ModCenter(P: in out CpCosSin1) is
    C: FPoly1 renames P.C;
  begin
    for I in P.P .. FDeg loop ModCenter(C(I)); end loop;
  end ModCenter;

  procedure ModCenter(P1: in CpCosSin1; P2: in out CpCosSin1) is
    C1: FPoly1 renames P1.C;
    C2: FPoly1 renames P2.C;
  begin
    P2.P := P1.P;
    for I in P1.P .. FDeg loop ModCenter(C1(I),C2(I)); end loop;
    if Not_STrunc then P2.E := P1.E; end if;
  end ModCenter;

  function ModCenter(P: CpCosSin1) return CpCosSin1 is
    Q: CpCosSin1;
  begin
    ModCenter(P,Q);
    return Q;
  end ModCenter;

  procedure SplitErr(P: in CpCosSin1; PC,PE: in out CpCosSin1) is
    --- split P into a sharp PC and an error PE
  begin
    Center(P,PC);
    ModCenter(P,PE);
  end SplitErr;

  procedure ErrMult(R: in Radius; P: in out CpCosSin1) is
    C: FPoly1 renames P.C;
  begin
    if Not_STrunc then
      for I in P.P .. FDeg loop ErrMult(R,C(I)); end loop;
      declare
        E: Errs1 renames P.E;
      begin
        for I in EFreq loop E(I) := R*E(I); end loop;
      end;
    end if;
  end ErrMult;

  procedure Union(P1: in CpCosSin1; P2: in out CpCosSin1) is
    C1: FPoly1 renames P1.C;
    C2: FPoly1 renames P2.C;
  begin
    if P1.P /= P2.P then raise Parity_Error; end if;
    for I in P1.P .. FDeg loop Union(C1(I),C2(I)); end loop;
    if Not_STrunc then
      declare
        E1: Errs1 renames P1.E;
        E2: Errs1 renames P2.E;
      begin
        for I in EFreq loop E2(I) := RMax(E1(I),E2(I)); end loop;
      end;
    end if;
  end Union;

  function Union(P1,P2: CpCosSin1) return CpCosSin1 is
    Q: CpCosSin1 := P2;
  begin
    Union(P1,Q);
    return Q;
  end Union;

  ------------------------------------------------------------

  procedure Neg(P: in out CpCosSin1) is
    C: FPoly1 renames P.C;
  begin
    for I in P.P .. FDeg loop Neg(C(I)); end loop;
  end Neg;

  procedure Neg(P1: in CpCosSin1; P2: in out CpCosSin1) is
    C1: FPoly1 renames P1.C;
    C2: FPoly1 renames P2.C;
  begin
    P2.P := P1.P;
    for I in P1.P .. FDeg loop Neg(C1(I),C2(I)); end loop;
    if Not_STrunc then P2.E := P1.E; end if;
  end Neg;

  function "-"(P: CpCosSin1) return CpCosSin1 is
    Q: CpCosSin1;
  begin
    Neg(P,Q);
    return Q;
  end "-";

  procedure Add(I: in Integer; P: in out CpCosSin1) is
  begin
    if P.P=1 then raise Parity_Error; end if;
    Add(I,P.C(0));
  end Add;

  procedure Add(S: in Scalar; P: in out CpCosSin1) is
  begin
    if P.P=1 then raise Parity_Error; end if;
    Add(S,P.C(0));
  end Add;

  procedure Add(P1: in CpCosSin1; P2: in out CpCosSin1) is
    C1: FPoly1 renames P1.C;
    C2: FPoly1 renames P2.C;
  begin
    if P1.P /= P2.P then raise Parity_Error; end if;
    for I in P1.P .. FDeg loop Add(C1(I),C2(I)); end loop;
    if Not_STrunc then
      declare
        E1: Errs1 renames P1.E;
        E2: Errs1 renames P2.E;
      begin
        for I in EFreq loop E2(I) := E2(I)+E1(I); end loop;
      end;
    end if;
  end Add;

  procedure Sum(P1,P2: in CpCosSin1; P3: in out CpCosSin1) is
    C1: FPoly1 renames P1.C;
    C2: FPoly1 renames P2.C;
    C3: FPoly1 renames P3.C;
  begin
    if P1.P /= P2.P then raise Parity_Error; end if;
    P3.P := P1.P;
    for I in P1.P .. FDeg loop Sum(C1(I),C2(I),C3(I)); end loop;
    if Not_STrunc then
      declare
        E1: Errs1 renames P1.E;
        E2: Errs1 renames P2.E;
        E3: Errs1 renames P3.E;
      begin
        for I in EFreq loop E3(I) := E1(I)+E2(I); end loop;
      end;
    end if;
  end Sum;

  function "+"(P1,P2: CpCosSin1) return CpCosSin1 is
    P3: CpCosSin1;
  begin
    Sum(P1,P2,P3);
    return P3;
  end "+";

  procedure Sub(P1: in CpCosSin1; P2: in out CpCosSin1) is
    C1: FPoly1 renames P1.C;
    C2: FPoly1 renames P2.C;
  begin
    if P1.P /= P2.P then raise Parity_Error; end if;
    for I in P1.P .. FDeg loop Sub(C1(I),C2(I)); end loop;
    if Not_STrunc then
      declare
        E1: Errs1 renames P1.E;
        E2: Errs1 renames P2.E;
      begin
        for I in EFreq loop E2(I) := E2(I)+E1(I); end loop;
      end;
    end if;
  end Sub;

  procedure Diff(P1,P2: in CpCosSin1; P3: in out CpCosSin1) is
    C1: FPoly1 renames P1.C;
    C2: FPoly1 renames P2.C;
    C3: FPoly1 renames P3.C;
  begin
    if P1.P /= P2.P then raise Parity_Error; end if;
    P3.P := P1.P;
    for I in P1.P .. FDeg loop Diff(C1(I),C2(I),C3(I)); end loop;
    if Not_STrunc then
      declare
        E1: Errs1 renames P1.E;
        E2: Errs1 renames P2.E;
        E3: Errs1 renames P3.E;
      begin
        for I in EFreq loop E3(I) := E1(I)+E2(I); end loop;
      end;
    end if;
  end Diff;

  function "-"(P1,P2: CpCosSin1) return CpCosSin1 is
    P3: CpCosSin1;
  begin
    Diff(P1,P2,P3);
    return P3;
  end "-";

  procedure Mult(R: in Flt; P: in out CpCosSin1) is
    C: FPoly1 renames P.C;
    Tmp: Scalar;
  begin
    if R=Zero then
      FillZero(P);
    else
      for I in P.P .. FDeg loop Mult(R,C(I),Tmp); end loop;
      if Not_STrunc then
        declare
          E: Errs1 renames P.E;
          A: constant Flt := abs(R);
        begin
          for I in EFreq loop E(I) := A*E(I); end loop;
        end;
      end if;
    end if;
  end Mult;

  procedure Prod(R: in Flt; P1: in CpCosSin1; P2: in out CpCosSin1) is
  begin
    if R=One then
      Copy(P1,P2);
    elsif R=NegOne then
      Neg(P1,P2);
    else
      P2.P := P1.P;
      if R=Zero then
        FillZero(P2);
      else
        declare
          C1: FPoly1 renames P1.C;
          C2: FPoly1 renames P2.C;
        begin
          for I in P1.P .. FDeg loop Prod(R,C1(I),C2(I)); end loop;
        end;
        if Not_STrunc then
          declare
            E1: Errs1 renames P1.E;
            E2: Errs1 renames P2.E;
            A: constant Flt := abs(R);
          begin
            for I in EFreq loop E2(I) := A*E1(I); end loop;
          end;
        end if;
      end if;
    end if;
  end Prod;

  function "*"(R: Flt; P: CpCosSin1) return CpCosSin1 is
    Q: CpCosSin1;
  begin
    Prod(R,P,Q);
    return Q;
  end "*";

  procedure AddProd(R: in Flt; P1: in CpCosSin1; P2: in out CpCosSin1) is
  begin
    if R=Zero then
      null;
    elsif R=One then
      Add(P1,P2);
    elsif R=NegOne then
      Sub(P1,P2);
    else
      if P1.P /= P2.P then raise Parity_Error; end if;
      declare
        C1: FPoly1 renames P1.C;
        C2: FPoly1 renames P2.C;
        Tmp: Scalar;
      begin
        for I in P1.P .. QuasiDeg1(C1) loop AddProd(R,C1(I),C2(I),Tmp); end loop;
      end;
      if Not_STrunc then
        declare
          E1: Errs1 renames P1.E;
          E2: Errs1 renames P2.E;
          A: constant Flt := abs(R);
        begin
          for I in EFreq loop E2(I) := E2(I)+A*E1(I); end loop;
        end;
      end if;
    end if;
  end AddProd;

  procedure Mult(Q: in Rational; P: in out CpCosSin1) is
    QNum: constant LInt := LNum(Q);
    C: FPoly1 renames P.C;
  begin
    if QNum=0 then
      FillZero(P);
    else
      for I in P.P .. QuasiDeg1(C) loop Mult(Q,C(I)); end loop;
      if not IsPoly(P) then
        declare
          R: constant Flt := Flt(abs(QNum))/Flt(LDen(Q));
          E: Errs1 renames P.E;
        begin
          for I in EFreq loop E(I) := R*E(I); end loop;
        end;
      end if;
    end if;
  end Mult;

  procedure Mult(S: in Scalar; P: in out CpCosSin1) is
    C: FPoly1 renames P.C;
  begin
    if IsZero(S) then
      FillZero(P);
    else
      for I in P.P .. FDeg loop Mult(S,C(I)); end loop;
      if Not_STrunc then
        declare
          A: constant Flt := MaxNorm(S);
          E: Errs1 renames P.E;
        begin
          for I in EFreq loop E(I) := A*E(I); end loop;
        end;
      end if;
    end if;
  end Mult;

  procedure Prod(S: in Scalar; P1: in CpCosSin1; P2: in out CpCosSin1) is
    C1: FPoly1 renames P1.C;
    C2: FPoly1 renames P2.C;
  begin
    P2.P := P1.P;
    if IsZero(S) then
      FillZero(P2);
    else
      for I in P1.P .. FDeg loop Prod(S,C1(I),C2(I)); end loop;
      if Not_STrunc then
        declare
          A: constant Flt := MaxNorm(S);
          E1: Errs1 renames P1.E;
          E2: Errs1 renames P2.E;
        begin
          for I in EFreq loop E2(I) := A*E1(I); end loop;
        end;
      end if;
    end if;
  end Prod;

  function "*"(S: Scalar; P: CpCosSin1) return CpCosSin1 is
    Q: CpCosSin1;
  begin
    Prod(S,P,Q);
    return Q;
  end "*";

  procedure AddProd(S: in Scalar; P1: in CpCosSin1; P2: in out CpCosSin1) is
  begin
    if not IsZero(S) then
      if P1.P /= P2.P then raise Parity_Error; end if;
      declare
        C1: FPoly1 renames P1.C;
        C2: FPoly1 renames P2.C;
        Tmp: Scalar;
      begin
        for I in P1.P .. QuasiDeg1(C1) loop AddProd(S,C1(I),C2(I),Tmp); end loop;
      end;
      if Not_STrunc then
        declare
          A: constant Flt := MaxNorm(S);
          E1: Errs1 renames P1.E;
          E2: Errs1 renames P2.E;
        begin
          for I in EFreq loop E2(I) := E2(I)+A*E1(I); end loop;
        end;
      end if;
    end if;
  end AddProd;

  procedure Mult(P1: in CpCosSin1; P2,Tmp: in out CpCosSin1) is
  begin
    Prod(P1,P2,Tmp);
    Copy(Tmp,P2);
  end Mult;

  procedure Mult(P1: in CpCosSin1; P2: in out CpCosSin1) is
    Tmp: LP_Pointer;
  begin
    PPool.Allocate(Tmp);
    Prod(P1,P2,Tmp.all.Data);
    Copy(Tmp.all.Data,P2);
    PPool.Recycle(Tmp);
  end Mult;

  protected body Add_Type is
    procedure Add(P: in CpCosSin1; Q: in out CpCosSin1) is
      PC: FPoly1 renames P.C;
      QC: FPoly1 renames Q.C;
    begin
      for I in FFreq loop Add(PC(I),QC(I)); end loop;
      if Not_STrunc then
        declare
          PE: Errs1 renames P.E;
          QE: Errs1 renames Q.E;
        begin
          for I in EFreq loop QE(I) := PE(I)+QE(I); end loop;
        end;
      end if;
    end Add;
  end Add_Type;

  procedure CCProdCosCos2(C1,C2: in FPoly1; A1,A2: in CAbs; P3: in out CpCosSin1) is
    Task_Error: Boolean := False;
    PC: Protected_Counter(0);
    Update: Add_Type;

    procedure Prod_Loop(P: in out CpCosSin1) is
      C: FPoly1 renames P.C;
      E: Errs1 renames P.E;
      K,M: Integer;
      Tmp: Scalar;
    begin
      loop                                         --- 2*cos(m)*cos(n) = cos(m-n) + cos(m+n)
        PC.Next(M);
        exit when M>FDeg;
        if A1(M) /= Zero then
          for N in 0 .. FDeg loop
            if A2(N) /= Zero then
              K := M-N;                             --- freq diff
              AddProd(C1(M),C2(N),C(abs(K)),Tmp);
              K := M+N;                             --- freq sum
              if K <= FDeg then
                AddProd(C1(M),C2(N),C(K),Tmp);
              elsif Not_STrunc then
                E(K) := E(K)+A1(M)*A2(N)*NormFac(K);
              end if;
            end if;
          end loop;
        end if;
      end loop;
    end Prod_Loop;

    procedure Prod_Loop is
      Tmp: LP_Pointer;
    begin
      PPool.Allocate(Tmp);
      declare
        P: CpCosSin1 renames Tmp.all.Data;
      begin
        SetZero(0,P);
        Prod_Loop(P);
        Update.Add(P,P3);
      end;
      PPool.Recycle(Tmp);
    exception
      when others => Task_Error := True; raise;
    end Prod_Loop;

    task type Prod_Task_Type is end Prod_Task_Type;
    task body Prod_Task_Type is begin Proper_Rounding; Prod_Loop; end;

    NT: constant Natural := Reserve_Tasks(FDeg,Prod_Parallel);
  begin
    if NT=0 then
      Prod_Loop(P3);
    else
      declare
        Prod_Task: array(1 .. NT) of Prod_Task_Type;
        pragma Warnings (Off,Prod_Task);
      begin
        Prod_Loop;
      end;
      Free_Tasks(NT);
      if Task_Error then raise Sorry with "Exception in CCProdCosCos2"; end if;
    end if;
  end CCProdCosCos2;

  procedure CCProdCosSin2(C1,C2: in FPoly1; A1,A2: in CAbs; P3: in out CpCosSin1) is
    Task_Error: Boolean := False;
    PC: Protected_Counter(0);
    Update: Add_Type;

    procedure Prod_Loop(P: in out CpCosSin1) is
      C: FPoly1 renames P.C;
      E: Errs1 renames P.E;
      K,M: Integer;
      Tmp: Scalar;
    begin
      loop                                       --- 2*cos(m)*sin(n) = -sin(m-n) + sin(m+n)
        PC.Next(M);
        exit when M>FDeg;
        if A1(M) /= Zero then
          for N in 1 .. FDeg loop
            if A2(N) /= Zero then
              K := M-N;                          --- freq diff
              if K>0 then
                SubProd(C1(M),C2(N),C(K),Tmp);   ---    M>N
              elsif K<0 then
                AddProd(C1(M),C2(N),C(-K),Tmp);  ---    M<N
              end if;
              K := M+N;                          --- freq sum
              if K <= FDeg then
                AddProd(C1(M),C2(N),C(K),Tmp);
              elsif Not_STrunc then
                E(K) := E(K)+A1(M)*A2(N)*NormFac(K);
              end if;
            end if;
          end loop;
        end if;
      end loop;
    end Prod_Loop;

    procedure Prod_Loop is
      Tmp: LP_Pointer;
    begin
      PPool.Allocate(Tmp);
      declare
        P: CpCosSin1 renames Tmp.all.Data;
      begin
        SetZero(1,P);
        Prod_Loop(P);
        Update.Add(P,P3);
      end;
      PPool.Recycle(Tmp);
    exception
      when others => Task_Error := True; raise;
    end Prod_Loop;

    task type Prod_Task_Type is end Prod_Task_Type;
    task body Prod_Task_Type is begin Proper_Rounding; Prod_Loop; end;

    NT: constant Natural := Reserve_Tasks(FDeg,Prod_Parallel);
  begin
    if NT=0 then
      Prod_Loop(P3);
    else
      declare
        Prod_Task: array(1 .. NT) of Prod_Task_Type;
        pragma Warnings (Off,Prod_Task);
      begin
        Prod_Loop;
      end;
      Free_Tasks(NT);
      if Task_Error then raise Sorry with "Exception in CCProdCosSin2"; end if;
    end if;
  end CCProdCosSin2;

  procedure CCProdSinCos2(C1,C2: in FPoly1; A1,A2: in CAbs; P3: in out CpCosSin1) is
    Task_Error: Boolean := False;
    PC: Protected_Counter(1);
    Update: Add_Type;

    procedure Prod_Loop(P: in out CpCosSin1) is
      C: FPoly1 renames P.C;
      E: Errs1 renames P.E;
      K,M: Integer;
      Tmp: Scalar;
    begin
      loop                                      --- 2*sin(m)*cos(n) = sin(m-n) + sin(m+n)
        PC.Next(M);
        exit when M>FDeg;
        if A1(M) /= Zero then
          for N in 0 .. FDeg loop
            if A2(N) /= Zero then
              K := M-N;                          --- freq diff
              if K>0 then
                AddProd(C1(M),C2(N),C(K),Tmp);   ---    M>N
              elsif K<0 then
                SubProd(C1(M),C2(N),C(-K),Tmp);  ---    M<N
              end if;
              K := M+N;                          --- freq sum
              if K <= FDeg then
                AddProd(C1(M),C2(N),C(K),Tmp);
              elsif Not_STrunc then
                E(K) := E(K)+A1(M)*A2(N)*NormFac(K);
              end if;
            end if;
          end loop;
        end if;
      end loop;
    end Prod_Loop;

    procedure Prod_Loop is
      Tmp: LP_Pointer;
    begin
      PPool.Allocate(Tmp);
      declare
        P: CpCosSin1 renames Tmp.all.Data;
      begin
        SetZero(1,P);
        Prod_Loop(P);
        Update.Add(P,P3);
      end;
      PPool.Recycle(Tmp);
    exception
      when others => Task_Error := True; raise;
    end Prod_Loop;

    task type Prod_Task_Type is end Prod_Task_Type;
    task body Prod_Task_Type is begin Proper_Rounding; Prod_Loop; end;

    NT: constant Natural := Reserve_Tasks(FDeg,Prod_Parallel);
  begin
    if NT=0 then
      Prod_Loop(P3);
    else
      declare
        Prod_Task: array(1 .. NT) of Prod_Task_Type;
        pragma Warnings (Off,Prod_Task);
      begin
        Prod_Loop;
      end;
      Free_Tasks(NT);
      if Task_Error then raise Sorry with "Exception in CCProdSinCos2"; end if;
    end if;
  end CCProdSinCos2;

  procedure CCProdSinSin2(C1,C2: in FPoly1; A1,A2: in CAbs; P3: in out CpCosSin1) is
    Task_Error: Boolean := False;
    PC: Protected_Counter(1);
    Update: Add_Type;

    procedure Prod_Loop(P: in out CpCosSin1) is
      C: FPoly1 renames P.C;
      E: Errs1 renames P.E;
      K,M: Integer;
      Tmp: Scalar;
    begin
      loop                                         --- 2*sin(m)*sin(n) =  cos(m-n) - cos(m+n)
        PC.Next(M);
        exit when M>FDeg;
        if A1(M) /= Zero then
          for N in 1 .. FDeg loop
            if A2(N) /= Zero then
              K := M-N;                             --- freq diff
              AddProd(C1(M),C2(N),C(abs(K)),Tmp);
              K := M+N;                             --- freq sum
              if K <= FDeg then
                SubProd(C1(M),C2(N),C(K),Tmp);
              elsif Not_STrunc then
                E(K) := E(K)+A1(M)*A2(N)*NormFac(K);
              end if;
            end if;
          end loop;
        end if;
      end loop;
    end Prod_Loop;

    procedure Prod_Loop is
      Tmp: LP_Pointer;
    begin
      PPool.Allocate(Tmp);
      declare
        P: CpCosSin1 renames Tmp.all.Data;
      begin
        SetZero(0,P);
        Prod_Loop(P);
        Update.Add(P,P3);
      end;
      PPool.Recycle(Tmp);
    exception
      when others => Task_Error := True; raise;
    end Prod_Loop;

    task type Prod_Task_Type is end Prod_Task_Type;
    task body Prod_Task_Type is begin Proper_Rounding; Prod_Loop; end;

    NT: constant Natural := Reserve_Tasks(FDeg,Prod_Parallel);
  begin
    if NT=0 then
      Prod_Loop(P3);
    else
      declare
        Prod_Task: array(1 .. NT) of Prod_Task_Type;
        pragma Warnings (Off,Prod_Task);
      begin
        Prod_Loop;
      end;
      Free_Tasks(NT);
      if Task_Error then raise Sorry with "Exception in CCProdSinSin2"; end if;
    end if;
  end CCProdSinSin2;

  procedure CCProd2(P1P,P2P: in Parity; C1,C2: in FPoly1; A1,A2: in CAbs; P3: in out CpCosSin1) is
    --- add twice C1*C2
    --- does not set P3.P
  begin
    if P1P=0 then
      if P2P=0 then
        CCProdCosCos2(C1,C2,A1,A2,P3); --  2*cos(m)*cos(n) =  cos(m-n) + cos(m+n)
      else
        CCProdCosSin2(C1,C2,A1,A2,P3); --  2*cos(m)*sin(n) = -sin(m-n) + sin(m+n)
      end if;
    else
      if P2P=0 then
        CCProdSinCos2(C1,C2,A1,A2,P3); --  2*sin(m)*cos(n) =  sin(m-n) + sin(m+n)
      else
        CCProdSinSin2(C1,C2,A1,A2,P3); --  2*sin(m)*sin(n) =  cos(m-n) - cos(m+n)
      end if;
    end if;
  end CCProd2;

  procedure CEProd2(P1P: in Parity; A1: in CAbs; E2: in Errs1; E3: in out Errs1) is
    --- add twice A1*E2
    E3F: Radius renames E3(0);
    E3L: Radius renames E3(EDeg);
    K: Integer;
  begin
    for M in P1P .. FDeg loop
      if A1(M) /= Zero then
        for N in 0 .. EDeg loop
          if E2(N) /= Zero then
            K := N-M;                                --- freq diff
            if K<0 then                              --- M>N
              K := -K;
              if K>N then
                E3F := E3F+A1(M)*E2(N)*NormFac(K-N); --- w(k+j) <= w(n+j)*w(k-n)
              else
                E3F := E3F+A1(M)*E2(N);              --- w(k+j) <= w(n+j)
              end if;
            else                                     --- M<N
              E3(K) := E3(K)+A1(M)*E2(N);            --- w(k+j) <= w(n+j)
            end if;
            K := N+M;                                --- freq sum
            if K <= EDeg then
              E3(K) := E3(K)+A1(M)*E2(N)*NormFac(M);
            else
              E3L := E3L+A1(M)*E2(N)*NormFac(M);
            end if;
          end if;
        end loop;
      end if;
    end loop;
  end CEProd2;

  procedure EEProd2(E1,E2: in Errs1; E3: in out Errs1) is
    --- add twice E1*E2
    --- slow version, more accurate
    E3F: Radius renames E3(0);
    E3L: Radius renames E3(EDeg);
    K: Integer;
  begin
    for M in 0 .. EDeg loop
      if E1(M) /= Zero then
        for N in 0 .. EDeg loop
          if E2(N) /= Zero then
            E3F := E3F+E1(M)*E2(N);           --- freq diff
            K := N+M;                         --- freq sum
            if K <= EDeg then
              E3(K) := E3(K)+E1(M)*E2(N);
            else
              E3L := E3L+E1(M)*E2(N);
            end if;
          end if;
        end loop;
      end if;
    end loop;
  end EEProd2;

  procedure Prod(P1,P2: in CpCosSin1; P3: in out CpCosSin1) is
    C1: FPoly1 renames P1.C;
    C2: FPoly1 renames P2.C;
    E1: Errs1 renames P1.E;
    E2: Errs1 renames P2.E;
    E3: Errs1 renames P3.E;
    ZC1: constant Boolean := Is_Zero(P1.P,C1);
    ZE1: constant Boolean := Is_Zero(E1);
    ZC2,ZE2: Boolean;
    A1,A2: CAbs;
  begin
    FillZero(P3);
    P3.P := (P1.P+P2.P) mod 2;
    if ZC1 and then ZE1 then return; end if; -- P1=0
    ZC2 := Is_Zero(P2.P,C2);
    ZE2 := Is_Zero(E2);
    if ZC2 then
      if ZE2 then return; end if;            -- P2=0
    else
      for I in P2.P .. FDeg loop A2(I) := MaxNorm(C2(I)); end loop;
    end if;
    if not ZC1 then
      for I in P1.P .. FDeg loop A1(I) := MaxNorm(C1(I)); end loop;
      if not ZC2 then
        CCProd2(P1.P,P2.P,C1,C2,A1,A2,P3); --- parallelized
      end if;
    end if;
    if not (ZC1 or ZE2) then
      CEProd2(P1.P,A1,E2,E3); --- could be parallelized
    end if;
    if not (ZC2 or ZE1) then
      CEProd2(P2.P,A2,E1,E3); --- could be parallelized
    end if;
    if not (ZE1 or ZE2) then
      EEProd2(E1,E2,E3);      --- could be parallelized
    end if;
    if P3.P=1 then LiftE0(P3); end if;
    Mult(SHalf,P3);
  end Prod;

  function "*"(P1,P2: CpCosSin1) return CpCosSin1 is
    P3: CpCosSin1;
  begin
    Prod(P1,P2,P3);
    return P3;
  end "*";

  procedure AddProd(P1,P2: in CpCosSin1; P3: in out CpCosSin1) is
    Tmp: LP_Pointer;
  begin
    PPool.Allocate(Tmp);
    Prod(P1,P2,Tmp.all.Data);
    Add(Tmp.all.Data,P3);
    PPool.Recycle(Tmp);
  end AddProd;

  procedure AddProd(P1,P2: in CpCosSin1; P3,Tmp: in out CpCosSin1) is
  begin
    Prod(P1,P2,Tmp);
    Add(Tmp,P3);
  end AddProd;

  procedure SubProd(P1,P2: in CpCosSin1; P3: in out CpCosSin1) is
    Tmp: LP_Pointer;
  begin
    PPool.Allocate(Tmp);
    Prod(P1,P2,Tmp.all.Data);
    Sub(Tmp.all.Data,P3);
    PPool.Recycle(Tmp);
  end SubProd;

  procedure SubProd(P1,P2: in CpCosSin1; P3,Tmp: in out CpCosSin1) is
  begin
    Prod(P1,P2,Tmp);
    Sub(Tmp,P3);
  end SubProd;

  procedure Div(R: in Flt; P: in out CpCosSin1) is
    C: FPoly1 renames P.C;
    Tmp: Scalar;
  begin
    for I in P.P .. FDeg loop Div(R,C(I),Tmp); end loop;
    if Not_STrunc then
      declare
        E: Errs1 renames P.E;
        A: constant Flt := abs(R);
      begin
        for I in EFreq loop E(I) := E(I)/A; end loop;
      end;
    end if;
  end Div;

  procedure Quot(P1: in CpCosSin1; R: in Flt; P2: in out CpCosSin1) is
    C1: FPoly1 renames P1.C;
    C2: FPoly1 renames P2.C;
  begin
    P2.P := P1.P;
    for I in P1.P .. FDeg loop Quot(C1(I),R,C2(I)); end loop;
    if Not_STrunc then
      declare
        E1: Errs1 renames P1.E;
        E2: Errs1 renames P2.E;
        A: constant Flt := abs(R);
      begin
        for I in EFreq loop E2(I) := E1(I)/A; end loop;
      end;
    end if;
  end Quot;

  function "/"(P: CpCosSin1; R: Flt) return CpCosSin1 is
    Q: CpCosSin1;
  begin
    Quot(P,R,Q);
    return Q;
  end "/";

  ------------------------------------------------

  procedure Assign(I: in Integer; P: in out CpCosSin1) is
  begin
    SetZero(0,P);
    Assign(I,P.C(0));
  end Assign;

  procedure Assign(Q: in Rational; P: in out CpCosSin1) is
  begin
    SetZero(0,P);
    Assign(Q,P.C(0));
  end Assign;

  procedure Assign(R: in Flt; P: in out CpCosSin1) is
  begin
    SetZero(0,P);
    Assign(R,P.C(0));
  end Assign;

  procedure Assign(S: in Scalar; P: in out CpCosSin1) is
  begin
    SetZero(0,P);
    Copy(S,P.C(0));
  end Assign;

  procedure Show1(N: in String; P: in CpCosSin1; Hide0: in Boolean := True) is
    use SV;
  begin
    if not IsZero(P) then
      if P.P=0 then
        Show1(N & "Cos ",P.C(P.P .. FDeg));
      else
        Show1(N & "Sin ",P.C(P.P .. FDeg));
      end if;
      if not IsPoly(P) then
        Show1(N & "Err ",P.E);
      end if;
    elsif not Hide0 then
      Show0(N & "0");
    end if;
  end Show1;

  procedure Show2(N: in String; P1,P2: in CpCosSin1; Hide0: in Boolean := True) is
    use SV;
  begin
    if P1.P /= P2.P then raise Not_Implemented; end if;
    if not (IsZero(P1) and IsZero(P2)) then
      if P1.P=0 then
        Show2(N & "Cos ",P1.C(P1.P .. FDeg),P2.C(P2.P .. FDeg));
      else
        Show2(N & "Sin ",P1.C(P1.P .. FDeg),P2.C(P2.P .. FDeg));
      end if;
      if not (IsPoly(P1) and IsPoly(P2)) then
        Show2(N & "Err ",P1.E,P2.E);
      end if;
    elsif not Hide0 then
      Show0(N & "0 0");
    end if;
  end Show2;

  procedure Put(F: in File_Type; P: in CpCosSin1; Decimal: in Boolean := False) is
    C: FPoly1 renames P.C;
    E: Errs1 renames P.E;
    DC: constant Integer := QuasiDeg(P.P,C);
    DE: Integer;
  begin
    Put(F,WFac,Decimal);
    Put(F,WPow,Decimal);
    Put(F,P.P);
    Put(F,DC);
    for I in P.P .. DC loop
      Put(F,C(I),Decimal);
    end loop;
    if STrunc then
      Ints.Put(F,0);
      Put(F,Zero,Decimal);
    else
      DE := QuasiDeg(E);
      Put(F,DE);
      for I in 0 .. DE loop
        Put(F,E(I),Decimal);
      end loop;
    end if;
  end Put;

  procedure Get(F: in File_Type; P: in out CpCosSin1; Decimal: in Boolean := False) is
    DC,DE: Integer := 0;
    WF,WP,R: Flt := Zero;
    S: Scalar;
  begin
    FillZero(P);
    Get(F,WF,Decimal);
    Get(F,WP,Decimal);
    Get(F,P.P);
    Get(F,DC);
    for I in P.P .. DC loop
      Get(F,S,Decimal);
      AddCoeff(I,S,P);
    end loop;
    Get(F,DE);
    for I in 0 .. DE loop
      Get(F,R,Decimal);
      AddErr(WF,WP,I,R,P);
    end loop;
    if Not_STrunc then
      if P.P=1 then LiftE0(P); end if;
      if  WF<WFac or else WP<WPow then raise Domain_Error; end if;
    end if;
  end Get;

  procedure Write(FileName: in String; P: in CpCosSin1; Decimal: in Boolean := False) is
    F: File_Type;
  begin
    if Verbosity>0 then Show0("Writing " & FileName); end if;
    Create(F,Out_File,FileName);
    Put(F,P,Decimal);
    Close(F);
  end Write;

  procedure Read(FileName: in String; P: in out CpCosSin1; Decimal: in Boolean := False) is
    F: File_Type;
  begin
    if Verbosity>0 then Show0("Reading " & FileName); end if;
    Open(F,In_File,FileName);
    Get(F,P,Decimal);
    Close(F);
  end Read;

  function Read(FileName: String; Decimal: Boolean := False) return CpCosSin1 is
    P: CpCosSin1;
  begin
    Read(FileName,P,Decimal);
    return P;
  end Read;

  -----------------------------------------

  function IsReal(P: CpCosSin1) return Boolean is
    C: FPoly1 renames P.C;
  begin
    if SInfo.IsComplex then
      for I in P.P .. FDeg loop
        if not IsReal(C(I)) then return False; end if;
      end loop;
      if not IsPoly(P) then raise Not_Certain; end if;
    end if;
    return True;
  end IsReal;

  procedure Real_Part(P: in out CpCosSin1) is
    C: FPoly1 renames P.C;
  begin
    if SInfo.IsComplex then
      for I in P.P .. FDeg loop Real_Part(C(I)); end loop;
    end if;
  end Real_Part;

  procedure Imag_Part(P: in out CpCosSin1) is
    C: FPoly1 renames P.C;
  begin
    for I in P.P .. FDeg loop Imag_Part(C(I)); end loop;
    if not (STrunc or else SInfo.IsComplex) then
      P.E := (others => Zero);
    end if;
  end Imag_Part;

  procedure Adjoint(P: in out CpCosSin1) is
    C: FPoly1 renames P.C;
  begin
    if SInfo.IsComplex then
      for I in P.P .. FDeg loop Adjoint(C(I)); end loop;
    end if;
  end Adjoint;

  procedure Adjoint(P1: in CpCosSin1; P2: in out CpCosSin1) is
    C1: FPoly1 renames P1.C;
    C2: FPoly1 renames P2.C;
  begin
    P2.P := P1.P;
    for I in P1.P .. FDeg loop Adjoint(C1(I),C2(I)); end loop;
    if Not_STrunc then P2.E := P1.E; end if;
  end Adjoint;

  procedure Simple_Random(P: in out CpCosSin1) is
    --- assumes P.P has been set
    C: FPoly1 renames P.C;
  begin
    FillZero(P);
    for I in P.P .. FDeg loop
      Simple_Random(C(I));
      Mult(InvNormFac(I),C(I));
    end loop;
    if Not_STrunc then P.E := (others => Zero); end if;
  end Simple_Random;

  procedure Simple_Random(Par: in Parity; P: in out CpCosSin1) is
  begin
    P.P := Par;
    Simple_Random(P);
  end Simple_Random;

  function MaxNorm(P: CpCosSin1) return Radius is
    C: FPoly1 renames P.C;
    N: Flt := Zero;
  begin
    if Not_STrunc then
      declare
        E: Errs1 renames P.E;
      begin
        for I in reverse EFreq loop N := N+E(I); end loop;
      end;
    end if;
    for I in reverse 1 .. FDeg loop
      N := N+NormFac(I)*MaxNorm(C(I));
    end loop;
    if P.P=0 then N := N+MaxNorm(C(0)); end if;
    return N;
  end MaxNorm;

  procedure Norm(P: in CpCosSin1; S: in out Scalar) is
    C: FPoly1 renames P.C;
    NC,Tmp: Scalar;
  begin
    if IsPoly(P) then
      SetZero(S);
    else
      declare
        E: Errs1 renames P.E;
        NE: Flt := Zero;
      begin
        for I in reverse EFreq loop NE := NE+E(I); end loop;
        BallAt0(NE,S);
      end;
    end if;
    for I in reverse 1 .. FDeg loop
      Norm(C(I),NC);
      AddProd(NormFac(I),NC,S,Tmp);
    end loop;
    Norm(C(0),NC);
    Add(NC,S);
  end Norm;

  function Norm(P: CpCosSin1) return Scalar is
    S: Scalar;
  begin
    Norm(P,S);
    return S;
  end Norm;

  procedure YRange(P: in CpCosSin1; YInf,YSup: out Flt) is
    --- cheap
    C: FPoly1 renames P.C;
  begin
    YSup := Zero;
    if Not_STrunc then
      declare
        E: Errs1 renames P.E;
      begin
        for I in reverse EFreq loop YSup := YSup+E(I); end loop;
      end;
    end if;
    for I in reverse 1 .. FDeg loop
      YSup := YSup+MaxNorm(C(I));
    end loop;
    if P.P=0 then
      YInf := -(YSup-Inf(C(0)));
      YSup := YSup+Sup(C(0));
    else
      YInf := -YSup;
    end if;
  end YRange;

  function Inf(P: CpCosSin1) return Flt is
    YInf,YSup: Flt;
  begin
    YRange(P,YInf,YSup);
    return YInf;
  end Inf;

  function Sup(P: CpCosSin1) return Flt is
    YInf,YSup: Flt;
  begin
    YRange(P,YInf,YSup);
    return YSup;
  end Sup;

  function LInftyNorm(P: CpCosSin1) return Radius is
    YInf,YSup: Flt;
  begin
    YRange(P,YInf,YSup);
    return RMax(abs(YInf),abs(YSup));
  end LInftyNorm;

  procedure Average(P: in CpCosSin1; S: in out Scalar) is
  begin
    if P.P=1 then
      SetZero(S);
    elsif (STrunc or else P.E(0)=Zero) then
      Copy(P.C(0),S);
    else
      BallAt0(P.E(0),S);
      Add(P.C(0),S);
    end if;
  end Average;

  function Average(P: CpCosSin1) return Scalar is
    S: Scalar;
  begin
    Average(P,S);
    return S;
  end Average;

  procedure SetAverage0(P: in out CpCosSin1) is
  begin
    SetZero(P.C(0)); -- should be 0 already if P.P=1
    if Not_STrunc then LiftE0(P); end if;
  end SetAverage0;

  procedure SplitOffAverage(P: in out CpCosSin1; S: in out Scalar) is
    C: Scalar renames P.C(0);
    E: Errs1 renames P.E;
  begin
    if P.P=1 then
      SetZero(S);
      if Not_STrunc then
        E(1) := E(1)+E(0);
        E(0) := Zero;
      end if;
    else
      if STrunc or else E(0)=Zero then
        Copy(C,S);
      else
        BallAt0(E(0),S);
        Add(C,S);
        E(1) := E(1)+E(0);
        E(0) := Zero;
      end if;
      SetZero(C);
    end if;
  end SplitOffAverage;

  procedure Coeff(I: in Natural; P: in CpCosSin1; S: in out Scalar) is
  begin
    if I=0 then
      Average(P,S);
    elsif IsPoly(P) then
      if I<=FDeg then Copy(P.C(I),S); else SetZero(S); end if;
    else
      declare
        E: Errs1 renames P.E;
        ESum: Flt := Zero;
      begin
        for K in reverse 0 .. IMin(I,EDeg) loop ESum := ESum+E(K); end loop;
        BallAt0(ESum*InvCpWeight(WFac,WPow,I),S);
      end;
      if I<=FDeg then Add(P.C(I),S); end if;
    end if;
  end Coeff;

  function Coeff(I: Natural; P: CpCosSin1) return Scalar is
    S: Scalar;
  begin
    Coeff(I,P,S);
    return S;
  end Coeff;

  procedure AddCoeff(I: in Natural; S: in Scalar; P: in out CpCosSin1) is
  begin
    if not IsZero(S) then
      if I <= FDeg then
        if I<P.P then raise Undefined; end if;
        Add(S,P.C(I));
      elsif Not_STrunc then
        if I<=EDeg then
          P.E(I) := P.E(I)+NormFac(I)*MaxNorm(S);
        elsif I<=NFLast then
          P.E(EDeg) := P.E(EDeg)+NormFac(I)*MaxNorm(S);
        else
          P.E(EDeg) := P.E(EDeg)+CpWeight(WFac,WPow,I)*MaxNorm(S);
        end if;
      end if;
    end if;
  end AddCoeff;

  procedure AddErr(WF,WP: in Radius; I: in Natural; R: in Flt; P: in out CpCosSin1) is
    --- check WF,WP beforehand
  begin
    if Not_STrunc then
      declare
        EI: Flt renames P.E(IMin(I,EDeg));
      begin
        if WF=WFac and then WP=WPow then
          EI := EI+abs(R);
        else
          EI := EI+CpWeight(WF,WP,I)*InvCpWeight(WFac,WPow,I)*abs(R);
        end if;
      end;
    end if;
  end AddErr;

  procedure AddErr(I: in Natural; R: in Flt; P: in out CpCosSin1) is
  begin
    if Not_STrunc then
      if I <= EDeg then
        P.E(I) := P.E(I)+abs(R);
      else
        P.E(EDeg) := P.E(EDeg)+abs(R);
      end if;
    end if;
  end AddErr;

  procedure AddBall(R: in Radius; P: in out CpCosSin1) is
  begin
    if not STrunc then
      P.E(0) := P.E(0)+R;
      if P.P=1 then LiftE0(P); end if;
    end if;
  end AddBall;

  procedure ValAt0(P: in CpCosSin1; S: in out Scalar; D: in Natural; Is_Poly: in Boolean) is
    --- D is assumed to be QuasiDeg(P.P,P.C)
    C: FPoly1 renames P.C;
  begin
    if P.P=1 then
      SetZero(S);
    else
      if Is_Poly then
        SetZero(S);
      else
        declare
          E: Errs1 renames P.E;
          ESum: Flt := Zero;
        begin
          for I in reverse EFreq loop
            ESum := ESum+InvNormFac(I)*E(I);
          end loop;
          BallAt0(ESum,S);
        end;
      end if;
      for I in reverse 0 .. D loop
        Add(C(I),S);
      end loop;
    end if;
  end ValAt0;

  procedure ValAt0(P: in CpCosSin1; S: in out Scalar) is
  begin
    ValAt0(P,S,QuasiDeg(P.P,P.C),IsPoly(P));
  end ValAt0;

  procedure CSPowers(X: in Scalar; C,S: in out FPoly1) is
    --- C(I)=Cos(I*X)  S(I)=Sin(I*X)
    Tmp1: Scalar renames C(0);
    Tmp2: Scalar renames S(0);
    I,J: Integer;
  begin
    Cos(X,C(1));
    Sin(X,S(1));
    for N in 2 .. FDeg loop
      I := N/2;
      J := N-N/2;
      Prod(C(I),C(J),Tmp1);
      Prod(S(I),S(J),Tmp2);
      Diff(Tmp1,Tmp2,C(N)); -- Cn = Ci*Cj-Si*Sj
      Prod(C(I),S(J),Tmp1);
      Prod(S(I),C(J),Tmp2);
      Sum(Tmp1,Tmp2,S(N));  -- Sn = Ci*Sj+Si*Cj
    end loop;
    Assign(One,C(0));
    SetZero(S(0));
  end CSPowers;

  procedure Val(P: in CpCosSin1; X: in Scalar; Y: in out Scalar; Tmp1,Tmp2: in out FPoly1) is
    C: FPoly1 renames P.C;
    D: constant Natural := QuasiDeg(P.P,C);
    Is_Poly: constant Boolean := IsPoly(P);
  begin
    if IsZero(X) then ValAt0(P,Y,D,Is_Poly); return; end if;
    if Is_Poly then
      if D=0 then
        if P.P=0 then Copy(C(0),Y); else SetZero(Y); end if;
        return;
      end if;
      SetZero(Y);
    else
      if not IsReal(X) then raise Domain_Error; end if;
      declare
        E: Errs1 renames P.E;
        ESum: Flt := Zero;
      begin
        for I in reverse EFreq loop
          ESum := ESum+InvNormFac(I)*E(I);
        end loop;
        BallAt0(ESum,Y);
      end;
    end if;
    declare
      CP: FPoly1 renames Tmp1;
      SP: FPoly1 renames Tmp2;
      S: Scalar renames SP(0); -- useless anyhow
    begin
      CSPowers(X,CP,SP);
      if P.P=0 then
        for I in reverse 1 .. D loop
          Prod(C(I),CP(I),S);
          Add(S,Y);
        end loop;
        Add(C(0),Y);
      else
        for I in reverse 1 .. D loop
          Prod(C(I),SP(I),S);
          Add(S,Y);
        end loop;
      end if;
    end;
  end Val;

  procedure Val(P: in CpCosSin1; X: in Scalar; Y: in out Scalar) is
    Tmp1,Tmp2: LP_Pointer;
  begin
    if IsZero(X) then
      ValAt0(P,Y);
    else
      PPool.Allocate(Tmp1);
      PPool.Allocate(Tmp2);
      Val(P,X,Y,Tmp1.all.Data.C,Tmp2.all.Data.C);
      PPool.Recycle(Tmp1);
      PPool.Recycle(Tmp2);
    end if;
  end Val;

  function Val(P: CpCosSin1; S: Scalar) return Scalar is
    V: Scalar;
  begin
    Val(P,S,V);
    return V;
  end Val;

  procedure IVal(P: in CpCossin1; XL,XR: in Flt; MinInf,MaxInf,MinSup,MaxSup: out Flt; Steps: in Positive) is
    --- evaluate on interval [XL,XR]
    DX: constant Flt := (XR-XL)/Flt(Steps);
    X1,X2,InfZ,SupZ: Flt;
    X,Z: Scalar;
    Tmp1,Tmp2: FPoly1;
  begin
    MinInf :=  Flt'Last;
    MaxInf := -Flt'Last;
    MinSup :=  Flt'Last;
    MaxSup := -Flt'Last;
    X2 := XL;
    for I in 0 .. Steps loop
      X1 := X2;
      if I=Steps then X2 := XR; else X2 := X2+DX; end if;
      Enclose(X1,X2,X);
      Val(P,X,Z,Tmp1,Tmp2);
      InfZ := Inf(Z);
      SupZ := Sup(Z);
      MinInf := RMin(InfZ,MinInf);
      MaxInf := RMax(InfZ,MaxInf);
      MinSup := RMin(SupZ,MinSup);
      MaxSup := RMax(SupZ,MaxSup);
    end loop;
  end IVal;

  procedure IVal(P: in CpCossin1; XL,XR: in Flt; Z: in out Scalar; Steps: in Positive) is
    MinInf,MaxInf,MinSup,MaxSup: Flt;
  begin
    IVal(P,XL,XR,MinInf,MaxInf,MinSup,MaxSup,Steps);
    Enclose(MinInf,MaxSup,Z);
  end IVal;

  function IVal_Range(P: CpCossin1; XL,XR,ZMin,ZMax: Flt; Steps: Positive) return Boolean is
    --- check on [XL,XR] whether all values are in (ZMin,ZMax)
    --- Steps is an initial guess
    N: constant LInt := 2**36;            -- max number of steps
    MaxWait: constant Positive := 32;
    DX: constant Radius := (XR-XL)/Flt(N);
    OK: Boolean;
    Count: Integer := 0;                 -- count number of steps
    Wait: Natural := MaxWait;            -- wait before trying to decrease Fac
    N0: LInt := Lint(Steps);
    Fac: LInt := 1;
    MaxSupZ: Flt := ZMin;
    MinInfZ: Flt := ZMax;
    X1,X2,SupZ,InfZ: Flt;
    X,Z: Scalar;
    Tmp1,Tmp2: FPoly1;
  begin
    while N0<N loop N0 := 2*N0; Fac := 2*Fac; end loop; -- Steps determines initial Fac
    X1 := XL;
    loop
      X2 := X1+Flt(Fac)*DX;                  -- increase X by Fac*DX
      if X2>XR then X2 := XR; end if;
      Enclose(X1,X2,X);
      Val(P,X,Z,Tmp1,Tmp2);
      SupZ := Sup(Z);
      InfZ := Inf(Z);
      OK := SupZ<ZMax and then InfZ>ZMin;
      if OK then                              -- [X1,X2] successful
        if SupZ>MaxSupZ then MaxSupZ := SupZ; end if;
        if InfZ<MinInfZ then MinInfZ := InfZ; end if;
        Count := Count+1;
        exit when X2=XR;
        X1 := X2;                             -- next X1
        Wait := Wait-1;
        if Wait=0 then                        -- try larger step
          Fac := 2*Fac;
          Wait := MaxWait;
        end if;
      else                                    -- [X1,X2] failed
        Fac := Fac/2;                         -- try smaller step
        exit when Fac=0;
        Wait := MaxWait;
      end if;
    end loop;
    if not OK then
      Show2("CpCossins1, STUCK at X1 X2",X1,X2);
      Show2("InfZ SupZ",InfZ,SupZ);
    end if;
    if Verbosity>2 then
      Show1("CpCossins1, actual steps:",Count);
      Show2("MinInfZ ZMin",MinInfZ,ZMin);
      Show2("MaxSupZ ZMax",MaxSupZ,ZMax);
    end if;
    return OK;
  end IVal_Range;

  procedure TranslatePi(P: in out CpCosSin1) is
    --- C(odd) := -C(odd)
    C: FPoly1 renames P.C;
    I: Integer := 1;
  begin
    while I <= FDeg loop Neg(C(I)); I := I+2; end loop;
  end TranslatePi;

  procedure TranslatePi(P1: in CpCosSin1; P2: in out CpCosSin1) is
    C1: FPoly1 renames P1.C;
    C2: FPoly1 renames P2.C;
    Odd: Boolean := True;
  begin
    P2.P := P1.P;
    if P1.P=0 then Copy(C1(0),C2(0)); else SetZero(C2(0)); end if;
    for I in 1 .. FDeg loop
      if Odd then Neg(C1(I),C2(I)); else Copy(C1(I),C2(I)); end if;
      Odd := not Odd;
    end loop;
    if Not_STrunc then P2.E := P1.E; end if;
  end TranslatePi;

  procedure NegTranslatePi(P1: in CpCosSin1; P2: in out CpCosSin1) is
    C1: FPoly1 renames P1.C;
    C2: FPoly1 renames P2.C;
    Odd: Boolean := True;
  begin
    P2.P := P1.P;
    SetZero(C2(0));
    for I in 1 .. FDeg loop
      if Odd then Copy(C1(I),C2(I)); else Neg(C1(I),C2(I)); end if;
      Odd := not Odd;
    end loop;
    if Not_STrunc then P2.E := P1.E; end if;
  end NegTranslatePi;

  procedure AntiDer(P: in out CpCosSin1) is
    C: FPoly1 renames P.C;
    Tmp: Scalar renames C(0);
  begin
    if P.P=0 then
      P.P := 1;
      if not IsZero(C(0)) then raise Undefined; end if;
      for I in 2 .. QuasiDeg1(C) loop
        Div(Flt(I),C(I),Tmp);
      end loop;
    else
      P.P := 0;
      Neg(C(1));
      for I in 2 .. QuasiDeg1(C) loop
        Div(Flt(-I),C(I),Tmp);
      end loop;
    end if;
    if not IsPoly(P) then
      declare
        E: Errs1 renames P.E;
      begin
        LiftE0(P);
        for I in 2 .. EDeg loop
          E(I) := E(I)/Flt(I);
        end loop;
      end;
    end if;
  end AntiDer;

  procedure AntiDer(P1: in CpCosSin1; P2: in out CpCosSin1) is
  begin
    Copy(P1,P2);
    AntiDer(P2);
  end AntiDer;

  function Integral(P: CpCosSin1; S1,S2: Scalar) return Scalar is
    Tmp: LP_Pointer;
    V1,V2,A: Scalar;
  begin
    PPool.Allocate(Tmp);
    declare
      PTmp: CpCosSin1 renames Tmp.all.Data;
    begin
      Copy(P,PTmp);
      SetZero(PTmp.C(0));
      AntiDer(PTmp);
      Val(PTmp,S1,V1);
      Val(PTmp,S2,V2);
    end;
    PPool.Recycle(Tmp);
    Sub(V1,V2);
    Average(P,A);
    if not IsZero(A) then
      Diff(S2,S1,V1);
      Mult(A,V1);
      Add(V1,V2);
    end if;
    return V2;
  end Integral;

  -----------------------------

  procedure NumTrunc(Siz: in Natural; P1: in CpCosSin1; P2: in out CpCosSin1) is
    C1: FPoly1 renames P1.C;
    C2: FPoly1 renames P2.C;
  begin
    P2.P := P1.P;
    if Siz >= FDeg then
      for I in P1.P .. FDeg loop Center(C1(I),C2(I)); end loop;
    else
      for I in P1.P .. Siz loop Center(C1(I),C2(I)); end loop;
      for I in Siz+1 .. FDeg loop SetZero(C2(I)); end loop;
    end if;
    if Not_STrunc then P2.E := (others => Zero); end if;
  end NumTrunc;

  function NumVal(P: CpCosSin1; T: Flt) return Flt is
    use Flt_EF;
    C: FPoly1 renames P.C;
    D: constant Integer := QuasiDeg(P.P,C);
    S: Flt := Zero;
  begin
    if P.P=1 then
      if T /= Zero then
        for I in reverse 1 .. D loop S := S+Approx(C(I))*Sin(Flt(I)*T); end loop;
      end if;
    elsif T=Zero then
      for I in reverse 0 .. D loop S := S+Approx(C(I)); end loop;
    else
      for I in reverse 0 .. D loop S := S+Approx(C(I))*Cos(Flt(I)*T); end loop;
    end if;
    return S;
  end NumVal;

  procedure PutPlot(F: in File_Type; P: in CpCosSin1; Steps: in Positive := 64; PutX: in Boolean := False) is
    use FLt_IO;
    Digs: constant Positive := 4;
    Pi2: constant Flt := 6.28318530718;
    X,Y: Flt;
  begin
    for N in 0 .. Steps loop
      X := Pi2*Flt(N)/Flt(Steps);
      Y := NumVal(P,X);
      if PutX then Put(F,X,3,Digs,4); end if;
      Put(F,Y,3,Digs,4);
      New_Line(F);
    end loop;
  end PutPlot;

end CpCosSins1;
