with Ada.Text_IO, ScalInfo, Logicals, Ints, Rationals, Flts, CpWeights;
use Ada.Text_IO, ScalInfo, Logicals, Ints, Rationals, Flts, CpWeights;
with Recycling, Vectors;

pragma Elaborate_All (Ada.Text_IO,ScalInfo,Logicals,Ints,Rationals,CpWeights);
pragma Elaborate_All (Recycling,Vectors);

generic --- modification of CosSin, with C^p norm instead of analytic norm

  FDeg: in Positive;               --- max degree
  WFac: in Radius;
  WPow: in Radius;                 --- weight factor is (1+WFac*N)^WPow
  Default_Parity: in Parity := 1;

  -------------------- begin_include scalars Scalar

  type Scalar is private;

  --- basic
  with function Info(Dummy: Scalar) return Scalar_Info is <>;
  with function IsSharp(S: Scalar) return Boolean is <>;
  --  with function IsNumbers(S: Scalar) return Boolean is <>;
  with function IsZero(S: Scalar) return Boolean is <>;
  with function "="(S1,S2: Scalar) return Boolean is <>;
  with procedure SetZero(S: in out Scalar) is <>;
  with procedure Copy(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Swap(S1,S2: in out Scalar) is <>;

  --- sets
  with function Center0(S: Scalar) return Boolean is <>;
  with function Contains0(S: Scalar) return Logical is <>;
  with function Contains(S1,S2: Scalar) return Logical is <>; -- S1 constains S2
  with procedure BallAt0(R: in Flt; S: in out Scalar) is <>;
  with function BallAt0(R: Flt) return Scalar is <>;
  with procedure DiskAt0(R: in Flt; S: in out Scalar) is <>;
  with function DiskAt0(R: Flt) return Scalar is <>;
  with procedure ToErr(S: in out Scalar) is <>;
  with procedure ToErr(S1: in Scalar; S2: in out Scalar) is <>;
  with function ToErr(S: Scalar) return Scalar is <>;
  with procedure Center(S: in out Scalar) is <>;
  with procedure Center(S1: in Scalar; S2: in out Scalar) is <>;
  with function Center(S: Scalar) return Scalar is <>;
  with procedure CenterDisk(S: in out Scalar) is <>;
  with procedure CenterDisk(S1: in Scalar; S2: in out Scalar) is <>;
  with function CenterDisk(S: Scalar) return Scalar is <>;
  with procedure ModCenter(S: in out Scalar) is <>;
  with procedure ModCenter(S1: in Scalar; S2: in out Scalar) is <>;
  with function ModCenter(S: Scalar) return Scalar is <>;
  with procedure ModNumbers(S: in out Scalar) is <>;
  with procedure ModNumbers(S1: in Scalar; S2: in out Scalar) is <>;
  with function ModNumbers(S: Scalar) return Scalar is <>;
  with procedure ErrMult(R: in Radius; S: in out Scalar) is <>;
  with procedure Union(S1: in Scalar; S2: in out Scalar) is <>;
  with function Union(S1,S2: Scalar) return Scalar is <>;
  with procedure Intersection(S1: in Scalar; S2: in out Scalar; Empty: out Logical) is <>;

  --- order
  with function Sign(S: Scalar) return Integer is <>;
  with function Compare(S1,S2: Scalar) return Integer is <>;
  with function "<"(S1,S2: Scalar) return Boolean is <>;
  with function "<="(S1,S2: Scalar) return Boolean is <>;
  with function ">"(S1,S2: Scalar) return Boolean is <>;
  with function ">="(S1,S2: Scalar) return Boolean is <>;
  with procedure Min(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Min(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function Min(S1,S2: Scalar) return Scalar is <>;
  with procedure Max(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Max(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function Max(S1,S2: Scalar) return Scalar is <>;
  with function Inf(S: Scalar) return Flt is <>; -- inf of real part
  with function Sup(S: Scalar) return Flt is <>; -- sup of real part

  --- arithmetic
  with procedure Neg(S: in out Scalar) is <>;
  with procedure Neg(S1: in Scalar; S2: in out Scalar) is <>;
  with function "-"(S: Scalar) return Scalar is <>;
  with procedure Add(I: in Integer; S: in out Scalar) is <>;
  with procedure Add(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Sum(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function "+"(S1,S2: Scalar) return Scalar is <>;
  with procedure Sub(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Diff(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function "-"(S1,S2: Scalar) return Scalar is <>;
  with procedure Mult(R: in Flt; S,Tmp: in out Scalar) is <>;
  with procedure Mult(R: in Flt; S: in out Scalar) is <>;
  with procedure Prod(R: in Flt; S1: in Scalar; S2: in out Scalar) is <>;
  with function "*"(R: Flt; S: Scalar) return Scalar is <>;
  with procedure AddProd(R: in Flt; S1: in Scalar; S2,Tmp: in out Scalar) is <>;
  with procedure AddProd(R: in Flt; S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Mult(Q: in Rational; S: in out Scalar) is <>;
  with procedure Mult(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Prod(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function "*"(S1,S2: Scalar) return Scalar is <>;
  with procedure AddProd(S1,S2: in Scalar; S3,Tmp: in out Scalar) is <>;
  with procedure AddProd(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with procedure SubProd(S1,S2: in Scalar; S3,Tmp: in out Scalar) is <>;
  with procedure Div(R: in Flt; S,Tmp: in out Scalar) is <>;
  with procedure Div(R: in Flt; S: in out Scalar) is <>;
  with procedure Quot(S1: in Scalar; R: in Flt; S2: in out Scalar) is <>;
  with function "/"(S: Scalar; R: Flt) return Scalar is <>;
  with procedure Div(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Quot(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function "/"(S1,S2: Scalar) return Scalar is <>;
  with procedure Inv(S1: in Scalar; S2: in out Scalar) is <>;
  with function Inv(S: Scalar) return Scalar is <>;
  with function "**"(S: Scalar; I: Integer) return Scalar is <>;

  --- fun
  with function IsReal(S: Scalar) return Boolean is <>;
  with procedure Adjoint(S: in out Scalar) is <>;
  with procedure Adjoint(S1: in Scalar; S2: in out Scalar) is <>;
  with function Adjoint(S: Scalar) return Scalar is <>;
  with procedure Real_Part(S: in out Scalar) is <>;
  with procedure Imag_Part(S: in out Scalar) is <>;
  with procedure Eval0(S: in out Scalar) is <>;
  with function "abs"(S: Scalar) return Scalar is <>;
  with procedure Norm(S1: in Scalar; S2: in out Scalar) is <>;
  with function Norm(S: Scalar) return Scalar is <>;
  with function MaxNorm(S: Scalar) return Radius is <>;
  with procedure Sqr(S: in out Scalar) is <>;
  with function Sqr(S: Scalar) return Scalar is <>;
  with procedure Sqrt(S1: in Scalar; S2: in out Scalar) is <>;
  with function Sqrt(S: Scalar) return Scalar is <>;
  with procedure Root(K: in Positive; S1: in Scalar; S2: in out Scalar) is <>;
  with function Root(K: Positive; S: Scalar) return Scalar is <>;
  with procedure Exp(S1: in Scalar; S2: in out Scalar) is <>;
  with function Exp(S: Scalar) return Scalar is <>;
  with procedure ArcCos(S1: in Scalar; S2: in out Scalar) is <>;
  with function ArcCos(S: Scalar) return Scalar is <>;
  with procedure ArcSin(S1: in Scalar; S2: in out Scalar) is <>;
  with function ArcSin(S: Scalar) return Scalar is <>;
  with procedure Cos(S1: in Scalar; S2: in out Scalar) is <>;
  with function Cos(S: Scalar) return Scalar is <>;
  with procedure Sin(S1: in Scalar; S2: in out Scalar) is <>;
  with function Sin(S: Scalar) return Scalar is <>;
  with procedure Log(S1: in Scalar; S2: in out Scalar) is <>;
  with function Log(S: Scalar) return Scalar is <>;
  with procedure Simple_Random(S: in out Scalar) is <>;
  with procedure QPower(Q: in Rational; S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Roots2(B,C: in Scalar; U1,U2,V: in out Scalar) is <>;
  with procedure Roots3(B,C,D: in Scalar; U0,U1,U2,V: in out Scalar) is <>;

  --- conversion and i/o
  with function Approx(S: Scalar) return Flt is <>;
  with procedure Assign(K: in Integer; S: in out Scalar) is <>;
  with procedure Assign(Q: in Rational; S: in out Scalar) is <>;
  with procedure Assign(R: in Flt; S: in out Scalar) is <>;
  with function Scal(I: Integer) return Scalar is <>;
  with function Scal(Q: Rational) return Scalar is <>;
  with function Scal(R: Flt) return Scalar is <>;
  with procedure Enclose(R1,R2: in Flt; S: in out Scalar) is <>;
  with function Enclose(R1,R2: Flt) return Scalar is <>;
  with procedure Show1(N: in String; S: in Scalar; Whatever: in Boolean := True) is <>;
  with procedure Show2(N: in String; S1,S2: in Scalar; Whatever: in Boolean := True) is <>;
  with procedure Put(F: in File_Type; S: in Scalar; Decimal: in Boolean := False) is <>;
  with procedure Get(F: in File_Type; S: in out Scalar; Decimal: in Boolean := False) is <>;
  with procedure Proper_Rounding is <>;

  --- misc
  with function Get_Precision(R: Scalar) return Positive is <>;
  -------------------- end_include scalars

  with package SV is new Vectors (Component => Scalar);
  --with package Poly is new Polynoms (Scalar => Scalar, SV => SV);

package CpCosSins1 is      -- cos series and sin series in 1 variable

  EDeg: constant Positive := 2*FDeg;
  subtype FFreq is Integer range 0 .. FDeg;
  subtype EFreq is Integer range 0 .. EDeg;
  subtype SVector is SV.Vector;
  subtype FPoly1 is SVector(FFreq);
  subtype Errs1 is FltVec(EFreq);
  subtype CAbs is FltVec(FFreq);

  NFLast: constant Positive := EDeg; --- at least EDeg
  NormFac: constant FltVec := CpWeight(WFac,WPow,NFLast);
  InvNormFac: constant FltVec := InvCpWeight(WFac,WPow,NFLast);

  Prod_Parallel: Boolean := False;

  type CpCosSin1 is
    record
      P: Parity; -- 0: even  1: odd
      C: FPoly1; -- cos(K.) if P=0,  sin(K.) if P=1
      E: Errs1;  -- higher order frequency errors (norms)
    end record;

  procedure Show_Cp_Param(Full: in Boolean := True);

  --- basic
  function Info(Dummy: CpCosSin1) return Scalar_Info;
  function IsZero(P: CpCosSin1) return Boolean;
  function IsSharp(P: CpCosSin1) return Boolean;          --- polynomial with sharp coeffs
  function IsPoly(P: CpCosSin1) return Boolean;
  function "="(P1,P2: CpCosSin1) return Boolean;
  procedure FillZero(P: in out CpCosSin1);                --- does not set parity
  procedure SetZero(Par: in Parity; P: in out CpCosSin1);
  procedure SetZero(P: in out CpCosSin1);                 --- sets default parity
  procedure Copy(P1: in CpCosSin1; P2: in out CpCosSin1);
  procedure Swap(P1,P2: in out CpCosSin1);

  --- sets
  function Center0(P: CpCosSin1) return Boolean;
  function Contains0(P: CpCosSin1) return Logical;
  procedure BallAt0(R: in Flt; Par: in Parity; P: in out CpCosSin1);
  procedure BallAt0(R: in Flt; P: in out CpCosSin1);      --- sets default parity
  function BallAt0(R: Flt) return CpCosSin1;              --- sets default parity
  procedure ToErr(P: in out CpCosSin1);
  procedure ToErr(P1: in CpCosSin1; P2: in out CpCosSin1);
  function ToErr(P: CpCosSin1) return CpCosSin1;
  procedure Center(P: in out CpCosSin1);
  procedure Center(P1: in CpCosSin1; P2: in out CpCosSin1);
  function Center(P: CpCosSin1) return CpCosSin1;
  procedure ModCenter(P: in out CpCosSin1);
  procedure ModCenter(P1: in CpCosSin1; P2: in out CpCosSin1);
  function ModCenter(P: CpCosSin1) return CpCosSin1;            --- zero center
  procedure SplitErr(P: in CpCosSin1; PC,PE: in out CpCosSin1);
  procedure ErrMult(R: in Radius; P: in out CpCosSin1);
  procedure Union(P1: in CpCosSin1; P2: in out CpCosSin1);
  function Union(P1,P2: CpCosSin1) return CpCosSin1;

  --- basic arithmetic
  procedure Neg(P: in out CpCosSin1);
  procedure Neg(P1: in CpCosSin1; P2: in out CpCosSin1);
  function "-"(P: CpCosSin1) return CpCosSin1;
  procedure Add(I: in Integer; P: in out CpCosSin1);
  procedure Add(S: in Scalar; P: in out CpCosSin1);
  procedure Add(P1: in CpCosSin1; P2: in out CpCosSin1);
  procedure Sum(P1,P2: in CpCosSin1; P3: in out CpCosSin1);
  function "+"(P1,P2: CpCosSin1) return CpCosSin1;
  procedure Sub(P1: in CpCosSin1; P2: in out CpCosSin1);
  procedure Diff(P1,P2: in CpCosSin1; P3: in out CpCosSin1);
  function "-"(P1,P2: CpCosSin1) return CpCosSin1;
  procedure Mult(R: in Flt; P: in out CpCosSin1);
  procedure Prod(R: in Flt; P1: in CpCosSin1; P2: in out CpCosSin1);
  function "*"(R: Flt; P: CpCosSin1) return CpCosSin1;
  procedure AddProd(R: in Flt; P1: in CpCosSin1; P2: in out CpCosSin1);
  procedure Mult(Q: in Rational; P: in out CpCosSin1);
  procedure Mult(S: in Scalar; P: in out CpCosSin1);
  procedure Prod(S: in Scalar; P1: in CpCosSin1; P2: in out CpCosSin1);
  function "*"(S: Scalar; P: CpCosSin1) return CpCosSin1;
  procedure AddProd(S: in Scalar; P1: in CpCosSin1; P2: in out CpCosSin1);
  procedure Mult(P1: in CpCosSin1; P2,Tmp: in out CpCosSin1);
  procedure Mult(P1: in CpCosSin1; P2: in out CpCosSin1);
  procedure Prod(P1,P2: in CpCosSin1; P3: in out CpCosSin1);
  function "*"(P1,P2: CpCosSin1) return CpCosSin1;
  procedure AddProd(P1,P2: in CpCosSin1; P3: in out CpCosSin1);
  procedure AddProd(P1,P2: in CpCosSin1; P3,Tmp: in out CpCosSin1);
  procedure SubProd(P1,P2: in CpCosSin1; P3: in out CpCosSin1);
  procedure SubProd(P1,P2: in CpCosSin1; P3,Tmp: in out CpCosSin1);
  procedure Div(R: in Flt; P: in out CpCosSin1);
  procedure Quot(P1: in CpCosSin1; R: in Flt; P2: in out CpCosSin1);
  function "/"(P: CpCosSin1; R: Flt) return CpCosSin1;

  --- conversion and i/o
  function Get_Precision(P: CpCosSin1) return Positive;
  procedure Assign(I: in Integer; P: in out CpCosSin1);
  procedure Assign(Q: in Rational; P: in out CpCosSin1);
  procedure Assign(R: in Flt; P: in out CpCosSin1);
  procedure Assign(S: in Scalar; P: in out CpCosSin1);
  procedure Show1(N: in String; P: in CpCosSin1; Hide0: in Boolean := True);
  procedure Show2(N: in String; P1,P2: in CpCosSin1; Hide0: in Boolean := True);
  procedure Put(F: in File_Type; P: in CpCosSin1; Decimal: in Boolean := False);
  procedure Get(F: in File_Type; P: in out CpCosSin1; Decimal: in Boolean := False);
  procedure Write(FileName: in String; P: in CpCosSin1; Decimal: in Boolean := False);
  procedure Read(FileName: in String; P: in out CpCosSin1; Decimal: in Boolean := False);
  function Read(FileName: String; Decimal: Boolean := False) return CpCosSin1;

  --- other
  function IsReal(P: CpCosSin1) return Boolean;
  procedure Real_Part(P: in out CpCosSin1);
  procedure Imag_Part(P: in out CpCosSin1);
  procedure Adjoint(P: in out CpCosSin1);
  procedure Adjoint(P1: in CpCosSin1; P2: in out CpCosSin1);
  procedure Simple_Random(P: in out CpCosSin1);
  procedure Simple_Random(Par: in Parity; P: in out CpCosSin1);
  function MaxNorm(P: CpCosSin1) return Radius;
  procedure Norm(P: in CpCosSin1; S: in out Scalar);
  function Norm(P: CpCosSin1) return Scalar;
  procedure YRange(P: in CpCosSin1; YInf,YSup: out Flt);
  function Inf(P: CpCosSin1) return Flt;
  function Sup(P: CpCosSin1) return Flt;
  function LInftyNorm(P: CpCosSin1) return Radius;
  procedure Average(P: in CpCosSin1; S: in out Scalar);
  function Average(P: CpCosSin1) return Scalar;
  procedure SetAverage0(P: in out CpCosSin1);
  procedure SplitOffAverage(P: in out CpCosSin1; S: in out Scalar);
  procedure Coeff(I: in Natural; P: in CpCosSin1; S: in out Scalar);
  function Coeff(I: in Natural; P: CpCosSin1) return Scalar;
  procedure AddCoeff(I: in Natural; S: in Scalar; P: in out CpCosSin1);
  procedure AddErr(I: in Natural; R: in Flt; P: in out CpCosSin1);
  procedure AddBall(R: in Radius; P: in out CpCosSin1);
  procedure ValAt0(P: in CpCosSin1; S: in out Scalar);
  procedure CSPowers(X: in Scalar; C,S: in out FPoly1);
  procedure Val(P: in CpCosSin1; X: in Scalar; Y: in out Scalar; Tmp1,Tmp2: in out FPoly1);
  procedure Val(P: in CpCosSin1; X: in Scalar; Y: in out Scalar);
  function Val(P: CpCosSin1; S: Scalar) return Scalar;
  procedure IVal(P: in CpCossin1; XL,XR: in Flt; MinInf,MaxInf,MinSup,MaxSup: out Flt; Steps: in Positive);
  procedure IVal(P: in CpCossin1; XL,XR: in Flt; Z: in out Scalar; Steps: in Positive);
  function IVal_Range(P: CpCossin1; XL,XR,ZMin,ZMax: Flt; Steps: Positive) return Boolean;
  procedure TranslatePi(P: in out CpCosSin1);
  procedure TranslatePi(P1: in CpCosSin1; P2: in out CpCosSin1);
  procedure NegTranslatePi(P1: in CpCosSin1; P2: in out CpCosSin1);
  procedure AntiDer(P: in out CpCosSin1);
  procedure AntiDer(P1: in CpCosSin1; P2: in out CpCosSin1);
  function Integral(P: CpCosSin1; S1,S2: Scalar) return Scalar;

  --- numerical
  procedure NumTrunc(Siz: in Natural; P1: in CpCosSin1; P2: in out CpCosSin1);
  function NumVal(P: CpCosSin1; T: Flt) return Flt;
  procedure PutPlot(F: in File_Type; P: in CpCosSin1; Steps: in Positive := 64; PutX: in Boolean := False);

  --- misc
  function Is_Zero(P: Parity; C: FPoly1) return Boolean;
  function Is_Zero(E: Errs1) return Boolean;

  -----------------------------------------------------
  -----------------------------------------------------

  package PAlloc is new Recycling (Item => CpCosSin1);
  subtype LP_Pointer is PAlloc.LD_Pointer;
  PAC: PAlloc.Controlled_Protected_Data;
  PPool: PAlloc.Protected_Data renames PAC.P;

private

  procedure Show_Cp_Rounding;

  function QuasiDeg(P: Parity; C: FPoly1) return Natural;
  function QuasiDeg(C: FPoly1) return Natural;
  function QuasiDeg1(C: FPoly1) return Natural;
  function QuasiDeg(E: Errs1) return Natural;
  procedure LiftE0(P: in out CpCosSin1);

  protected type Add_Type is
    procedure Add(P: in CpCosSin1; Q: in out CpCosSin1);
  end Add_Type;

  procedure CCProdCosCos2(C1,C2: in FPoly1; A1,A2: in CAbs; P3: in out CpCosSin1);
  procedure CCProdCosSin2(C1,C2: in FPoly1; A1,A2: in CAbs; P3: in out CpCosSin1);
  procedure CCProdSinCos2(C1,C2: in FPoly1; A1,A2: in CAbs; P3: in out CpCosSin1);
  procedure CCProdSinSin2(C1,C2: in FPoly1; A1,A2: in CAbs; P3: in out CpCosSin1);
  procedure CCProd2(P1P,P2P: in Parity; C1,C2: in FPoly1; A1,A2: in CAbs; P3: in out CpCosSin1);
  procedure CEProd2(P1P: in Parity; A1: in CAbs; E2: in Errs1; E3: in out Errs1);
  procedure EEProd2(E1,E2: in Errs1; E3: in out Errs1);

  procedure AddErr(WF,WP: in Radius; I: in Natural; R: in Flt; P: in out CpCosSin1);
  procedure ValAt0(P: in CpCosSin1; S: in out Scalar; D: in Natural; Is_Poly: in Boolean);

  SHalf:      constant Scalar      := Scal(Half);
  SInfo:      constant Scalar_Info := Info(SHalf);
  STrunc:     constant Boolean     := SInfo.IsNumeric;
  Not_STrunc: constant Boolean     := not STrunc;

end CpCosSins1;
