with ScalVectors, Linear, Linear.Contr;

pragma Elaborate_All (ScalVectors,Linear.Contr);

generic

  Mat_PackSize: in Positive;
  Map: in Positive;
  NN: in Integer;
  RR: in Radius;

  with package SSV is new ScalVectors (Scalar => Scalar, SM => SM, SV => SV);
  with procedure Put_Numeric(N: in Positive; S: in Scalar) is <>;
  with procedure Get_Numeric(N: in Positive; S: in out Scalar) is <>;

package Cylinders.Fix is

  pragma Elaborate_Body;

  subtype SMatrix is SM.Matrix;

  procedure Lin1(F: in out Cylinder; LF: in out Cylinder);
  procedure Lin2(F: in Cylinder; LF: in out Cylinder);
  procedure NegLapGMap(F: in Cylinder; GF: in out Cylinder);
  procedure GMap(F: in Cylinder; GF: in out Cylinder);
  procedure DGMap(F,FF3,H: in Cylinder; DGH: in out Cylinder);
  procedure DContrNorm(M: in SMatrix; F: in Cylinder; PackSize: in Positive; R: out Radius);
  procedure ContrFix(M: in SMatrix; PackSize: in Positive; F: in out Cylinder);

  package SV_Ops renames SSV.SV_Ops;
  package Lin is new Linear (Scalar => Scalar, SM => SM, SV => SV, Fun => Cylinder,
                             Mode => CylinderMode, Modes => CylinderModes);
  package Lin_Contr is new Lin.Contr (SSV => SSV);

  --- numerical
  procedure DerMatrix(F: in Cylinder; M: in out SMatrix);
  procedure ContrMatrix(F: in Cylinder; M: in out SMatrix);
  procedure ApproxFix(F: in out Cylinder; M: in out SMatrix; Eps: in Radius := Zero);
  procedure IterContr(F: in out Cylinder; M: in out SMatrix);
  procedure EigenDer(Name: in String; F: in Cylinder);

  QC: CylinderModes := Make(RadX,Mat_PackSize,True,True);   --- Simple and no Err modes
  subtype QC_Vector is SV.Vector(QC'Range);
  subtype QC_Matrix is SMatrix(QC'Range,QC'Range);

  ------------ plotting

  procedure PlotWPhi(FileName: in String; XL,YL: in Positive := 64);
  procedure PlotJDensity(C: in Cylinder; FileName: in String; XL,YL: in Positive := 64);
  procedure Plot3(C: in Cylinder; FileName: in String; XL,YL: in Positive := 64);

private

  Tiny: constant Radius := Flt'Scaling(One,Flt'Machine_Emin);
  CosNX: constant FScalar := Monom(abs(NN),One);       -- cos(kx)
  SmallCosX: constant FScalar := AplusMonom(1,One,RR); -- 1+r*cos(x)
  LT: LTaylor;  -- used for Iso, Cassini. To be initialised

end Cylinders.Fix;
