with Ada.Text_IO, Logicals, Ints, Rationals, Flts, Flt_Vectors, ScalInfo;
use Ada.Text_IO, Logicals, Ints, Rationals, Flts, ScalInfo;
with Matrices, Vectors,Polynoms,Fouriers1,CPCosSins1,LTaylors,Comb;

pragma Elaborate_All (Ada.Text_IO,Logicals,Ints, Rationals,Flt_Vectors,ScalInfo);
pragma Elaborate_All (Matrices,Polynoms,Vectors,Fouriers1,CPCosSins1,LTaylors,Comb);

generic

  DegX: in Positive;  --- max degree periodic
  DegY: in Positive;  --- max degree "radial"
  LDeg: in Natural;   --- max degree LTaylor
  RadX: in Radius;    --- weight factor for X
  WFac: in Radius;    --- weight factor is (1+WFac*N)
  EllR: in Radius;    --- e^(-ell) < |z| < 1
  Ref:  in MinusOneZeroOne;
  Symm: in Integer;
  Full: in Boolean := False;

  -------------------- begin_include scalars Scalar

  type Scalar is private;

  --- basic
  with function Info(Dummy: Scalar) return Scalar_Info is <>;
  with function IsSharp(S: Scalar) return Boolean is <>;
  with function IsNumbers(S: Scalar) return Boolean is <>;
  with function IsZero(S: Scalar) return Boolean is <>;
  with function "="(S1,S2: Scalar) return Boolean is <>;
  with procedure SetZero(S: in out Scalar) is <>;
  with procedure Copy(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Swap(S1,S2: in out Scalar) is <>;

  --- sets
  with function Center0(S: Scalar) return Boolean is <>;
  with function Contains0(S: Scalar) return Logical is <>;
  with function Contains(S1,S2: Scalar) return Logical is <>; -- S1 constains S2
  with procedure BallAt0(R: in Flt; S: in out Scalar) is <>;
  with function BallAt0(R: Flt) return Scalar is <>;
  with procedure DiskAt0(R: in Flt; S: in out Scalar) is <>;
  with function DiskAt0(R: Flt) return Scalar is <>;
  with procedure ToErr(S: in out Scalar) is <>;
  with procedure ToErr(S1: in Scalar; S2: in out Scalar) is <>;
  with function ToErr(S: Scalar) return Scalar is <>;
  with procedure Center(S: in out Scalar) is <>;
  with procedure Center(S1: in Scalar; S2: in out Scalar) is <>;
  with function Center(S: Scalar) return Scalar is <>;
  with procedure CenterDisk(S: in out Scalar) is <>;
  with procedure CenterDisk(S1: in Scalar; S2: in out Scalar) is <>;
  with function CenterDisk(S: Scalar) return Scalar is <>;
  with procedure ModCenter(S: in out Scalar) is <>;
  with procedure ModCenter(S1: in Scalar; S2: in out Scalar) is <>;
  with function ModCenter(S: Scalar) return Scalar is <>;
  with procedure ModNumbers(S: in out Scalar) is <>;
  with procedure ModNumbers(S1: in Scalar; S2: in out Scalar) is <>;
  with function ModNumbers(S: Scalar) return Scalar is <>;
  with procedure ErrMult(R: in Radius; S: in out Scalar) is <>;
  with procedure Union(S1: in Scalar; S2: in out Scalar) is <>;
  with function Union(S1,S2: Scalar) return Scalar is <>;
  with procedure Intersection(S1: in Scalar; S2: in out Scalar; Empty: out Logical) is <>;

  --- order
  with function Sign(S: Scalar) return Integer is <>;
  with function Compare(S1,S2: Scalar) return Integer is <>;
  with function "<"(S1,S2: Scalar) return Boolean is <>;
  with function "<="(S1,S2: Scalar) return Boolean is <>;
  with function ">"(S1,S2: Scalar) return Boolean is <>;
  with function ">="(S1,S2: Scalar) return Boolean is <>;
  with procedure Min(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Min(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function Min(S1,S2: Scalar) return Scalar is <>;
  with procedure Max(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Max(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function Max(S1,S2: Scalar) return Scalar is <>;
  with function Inf(S: Scalar) return Flt is <>; -- inf of real part
  with function Sup(S: Scalar) return Flt is <>; -- sup of real part

  --- arithmetic
  with procedure Neg(S: in out Scalar) is <>;
  with procedure Neg(S1: in Scalar; S2: in out Scalar) is <>;
  with function "-"(S: Scalar) return Scalar is <>;
  with procedure Add(I: in Integer; S: in out Scalar) is <>;
  with procedure Add(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Sum(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function "+"(S1,S2: Scalar) return Scalar is <>;
  with procedure Sub(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Diff(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function "-"(S1,S2: Scalar) return Scalar is <>;
  with procedure Mult(R: in Flt; S,Tmp: in out Scalar) is <>;
  with procedure Mult(R: in Flt; S: in out Scalar) is <>;
  with procedure Prod(R: in Flt; S1: in Scalar; S2: in out Scalar) is <>;
  with function "*"(R: Flt; S: Scalar) return Scalar is <>;
  with procedure AddProd(R: in Flt; S1: in Scalar; S2,Tmp: in out Scalar) is <>;
  with procedure AddProd(R: in Flt; S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Mult(Q: in Rational; S: in out Scalar) is <>;
  with procedure Mult(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Prod(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function "*"(S1,S2: Scalar) return Scalar is <>;
  with procedure AddProd(S1,S2: in Scalar; S3,Tmp: in out Scalar) is <>;
  with procedure AddProd(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with procedure SubProd(S1,S2: in Scalar; S3,Tmp: in out Scalar) is <>;
  with procedure Div(R: in Flt; S,Tmp: in out Scalar) is <>;
  with procedure Div(R: in Flt; S: in out Scalar) is <>;
  with procedure Quot(S1: in Scalar; R: in Flt; S2: in out Scalar) is <>;
  with function "/"(S: Scalar; R: Flt) return Scalar is <>;
  with procedure Div(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Quot(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function "/"(S1,S2: Scalar) return Scalar is <>;
  with procedure Inv(S1: in Scalar; S2: in out Scalar) is <>;
  with function Inv(S: Scalar) return Scalar is <>;
  with function "**"(S: Scalar; I: Integer) return Scalar is <>;

  --- fun
  with function IsReal(S: Scalar) return Boolean is <>;
  with procedure Adjoint(S: in out Scalar) is <>;
  with procedure Adjoint(S1: in Scalar; S2: in out Scalar) is <>;
  with function Adjoint(S: Scalar) return Scalar is <>;
  with procedure Real_Part(S: in out Scalar) is <>;
  with procedure Imag_Part(S: in out Scalar) is <>;
  with procedure Eval0(S: in out Scalar) is <>;
  with function "abs"(S: Scalar) return Scalar is <>;
  with procedure Norm(S1: in Scalar; S2: in out Scalar) is <>;
  with function Norm(S: Scalar) return Scalar is <>;
  with function MaxNorm(S: Scalar) return Radius is <>;
  with procedure Sqr(S: in out Scalar) is <>;
  with function Sqr(S: Scalar) return Scalar is <>;
  with procedure Sqrt(S1: in Scalar; S2: in out Scalar) is <>;
  with function Sqrt(S: Scalar) return Scalar is <>;
  with procedure Root(K: in Positive; S1: in Scalar; S2: in out Scalar) is <>;
  with function Root(K: Positive; S: Scalar) return Scalar is <>;
  with procedure Exp(S1: in Scalar; S2: in out Scalar) is <>;
  with function Exp(S: Scalar) return Scalar is <>;
  with procedure ArcCos(S1: in Scalar; S2: in out Scalar) is <>;
  with function ArcCos(S: Scalar) return Scalar is <>;
  with procedure ArcSin(S1: in Scalar; S2: in out Scalar) is <>;
  with function ArcSin(S: Scalar) return Scalar is <>;
  with procedure Cos(S1: in Scalar; S2: in out Scalar) is <>;
  with function Cos(S: Scalar) return Scalar is <>;
  with procedure Sin(S1: in Scalar; S2: in out Scalar) is <>;
  with function Sin(S: Scalar) return Scalar is <>;
  with procedure Log(S1: in Scalar; S2: in out Scalar) is <>;
  with function Log(S: Scalar) return Scalar is <>;
  with procedure Simple_Random(S: in out Scalar) is <>;
  with procedure QPower(Q: in Rational; S1: in Scalar; S2: in out Scalar) is <>;
  with procedure IPower(I: in Integer; S1: in Scalar; S2,Tmp: in out Scalar) is <>;
  with procedure Roots2(B,C: in Scalar; U1,U2,V: in out Scalar) is <>;
  with procedure Roots3(B,C,D: in Scalar; U0,U1,U2,V: in out Scalar) is <>;
  with function Pi return Scalar is <>;

  --- conversion and i/o
  with function Approx(S: Scalar) return Flt is <>;
  with procedure Assign(K: in Integer; S: in out Scalar) is <>;
  with procedure Assign(Q: in Rational; S: in out Scalar) is <>;
  with procedure Assign(R: in Flt; S: in out Scalar) is <>;
  with function Scal(I: Integer) return Scalar is <>;
  with function Scal(Q: Rational) return Scalar is <>;
  with function Scal(R: Flt) return Scalar is <>;
  with procedure Enclose(R1,R2: in Flt; S: in out Scalar) is <>;
  with function Enclose(R1,R2: Flt) return Scalar is <>;
  with procedure Show1(N: in String; S: in Scalar; Whatever: in Boolean := True) is <>;
  with procedure Show2(N: in String; S1,S2: in Scalar; Whatever: in Boolean := True) is <>;
  with procedure Put(F: in File_Type; S: in Scalar; Decimal: in Boolean := False) is <>;
  with procedure Get(F: in File_Type; S: in out Scalar; Decimal: in Boolean := False) is <>;

  --- misc
  with function Get_Precision(R: Scalar) return Positive is <>;
  with procedure Proper_Rounding is <>;
  with procedure Free_Cache(S: in Scalar) is <>;
  -------------------- end_include scalars

  with package SM is new Matrices (Scalar => Scalar);
  with package SV is new Vectors (Component => Scalar);

package Cylinders is

  package Fous1 is new Fouriers1 (PDeg => DegX, Scalar => Scalar, Rad => RadX, SV => SV, SM => SM);

  EDegX: Positive renames Fous1.EDeg;
  subtype FreqX is Fous1.Freq;
  subtype EFreqX is Fous1.EFreq;
  subtype FPolyX is Fous1.FPoly1;
  subtype ErrsX is Fous1.Errs1;
  subtype FScalar is Fous1.Fourier1;
  use Fous1;

  package SMF is new Matrices (Scalar => FScalar);
  package SVF is new Vectors (Component => FScalar);
  package Cyl is new CpCosSins1 (FDeg => DegY, WFac => WFac, WPow => One, Scalar => FScalar, SV => SVF);

  EDegY: Positive renames Cyl.EDeg;
  subtype FreqY is Integer range 1 .. DegY;
  subtype EFreqY is Cyl.EFreq;
  subtype FPolyY is Cyl.FPoly1;
  subtype ErrsY is Cyl.Errs1;
  subtype CAbs is Cyl.CAbs;
  subtype Cylinder is Cyl.CPCosSin1;
  use Cyl;

  LSums: constant Positive := IMax(LDeg,EDegY); --- save Sigma(0..LSums)
  KMax: constant Integer := 2*LSums;            --- first compute Sigma(KMax+1)
  subtype FltVec is Flt_Vectors.Vector;
  subtype Sums is FltVec(0 .. LSums);

  procedure Show_Cy_Param(Full: in Boolean := True);
  function Read(FileName: String; Decimal: Boolean := False) return FltVec;
  procedure Mult(S: in Scalar; A: in out Cylinder);

  type QData is         -- depending only on q
    record
      QDivKap: Scalar;  -- q/kappa
      QDivKap2: Scalar; -- (q/kappa)^2
      K0: Scalar;       -- (1-e^{-ql})q/(l*kappa^2)
      K1: Scalar;       -- (1+e^{-ql})q/(l*kappa^2)
      FourK0: Scalar;   -- 4*K0, or 2*K0 if WFac=0
      FourK1: Scalar;   -- 4*K1, or 2*K0 if WFac=0
      Sigma: Sums;      -- Sigma(t) = Sum_{k>=t} 1/(k^2+(q/kappa)^2)
    end record;

  procedure Init(Q: in Flt; O: in out QData);
  procedure Lqjn(O: in QData; J,N: in Positive; S: in out FScalar); -- for loops
  procedure Lqjn(Q: in Flt; J,N: in Positive; S: in out FScalar);   -- not for loops
  procedure Lq(Q: in Flt; P1: in Cylinder; P2: in out Cylinder);
  function LMaxNorm(Q: Natural) return Radius; -- needed by LTaylors
  function LVarRho(P: Positive) return Radius; -- needed by LTaylors
  procedure NegInvLap(A: in out Cylinder);
  procedure NegInvLap(A1: in Cylinder; A2: in out Cylinder);

  procedure Flip1(A: in out Cylinder);
  procedure Rotate(R: in Flt; A: in Cylinder; B: in out Cylinder);
  procedure ProjEven(A: in out Cylinder);
  procedure ProjOdd(A: in out Cylinder);
  procedure ProjSymm(N: in Positive; A: in out Cylinder);
  procedure ProjASymm(N: in Positive; A: in out Cylinder);
  procedure ProjConstant(A: in out Cylinder);

  package Radials is new CpCosSins1 (FDeg => DegY, WFac => WFac, WPow => One, Scalar => Scalar, SV => SV);
  subtype Radial is Radials.CpCosSin1;
  subtype RPoly is Radials.FPoly1;
  subtype RErrs is Radials.Errs1;

  procedure ValAtPi1(A: in Cylinder; V: in out Radial);
  procedure Val1(A: in Cylinder; X: in Scalar; V: in out Radial);
  procedure Val2(A: in Cylinder; Y: in Scalar; V: in out FScalar);
  procedure Val(A: in Cylinder; X,Y: in Scalar; S: in out Scalar);
  procedure IVal(A: in Cylinder; XL,XR,YC: in Flt; S: in out Scalar; Steps: in Positive);
  function IVal_Range(A: Cylinder; XL,XR,YC,ZMin,ZMax: Flt; Steps: Positive) return Boolean;
  procedure JValAt0(A: in Cylinder; YL,YR: in Flt; MinInf,MaxInf,MinSup,MaxSup: out Flt; Steps: in Positive);
  procedure JValAtPi(A: in Cylinder; YL,YR: in Flt; MinInf,MaxInf,MinSup,MaxSup: out Flt; Steps: in Positive);
  procedure JVal(A: in Cylinder; XC,YL,YR: in Flt; MinInf,MaxInf,MinSup,MaxSup: out Flt; Steps: in Positive);
  function JVal_Range(A: Cylinder; XC,YL,YR,ZMin,ZMax: Flt; Steps: Positive) return Boolean;
  function RVal_Range(A: Cylinder; XL,XR,YL,YR,ZMin,ZMax: Flt; IL,JL: Positive) return Boolean;
  procedure CheckMin(A: in Cylinder; XC,XL,XR,YC,YL,YR: in Flt; IL,JL: in Positive);
  procedure CheckMax(A: in Cylinder; XC,XL,XR,YC,YL,YR: in Flt; IL,JL: in Positive);
  procedure LevelStat(A: in Cylinder; Y,Cut: in Flt; Steps: in Positive);
  procedure Near_Cassini_Injective(R: in Flt);

  ----------------- LTaylors

  package PolyF is new Polynoms (Scalar => FScalar, SV => SVF);
  package LTays is new LTaylors(PDeg => LDeg, Scalar => FScalar, SV => SVF, SP => PolyF);
  subtype LTaylor is LTays.LTaylor;

  procedure Mult(S: in Scalar; P: in out LTaylor);
  procedure Prod(S: in Scalar; P1: in LTaylor; P2: in out LTaylor);
  procedure Apply(T: in LTaylor; A: in Cylinder; B: in out Cylinder);

  --------- |f|^2 for ...
  procedure PowerNegOne1(R: in Scalar; P: in out LTaylor; NLast: in Positive; PolyPsi: in Boolean);      --- f=(1-r\zeta)^{-1}
  procedure PowerNegOne2(R: in Scalar; P: in out LTaylor; NLast: in Positive; PolyPsi: in Boolean);      --- f=(1-r\zeta^2)^{-1}
  procedure PowerNegThreeHalf(R: in Scalar; P: in out LTaylor; NLast: in Positive; PolyPsi: in Boolean); --- f=(1-r\zeta^2)^{-3/2}

  --------- examples
  procedure Disk_Iso(R: in Flt; T: in out LTaylor; PolyPsi: in Boolean := False);
  procedure Disk_Iso_Plus(R: in Flt; T: in out LTaylor; PolyPsi: in Boolean := False);   --- also invludes L^2
  procedure InvEllipse(A: in Flt; T: in out LTaylor; PolyPsi: in Boolean := False);
  procedure InvEllipse_Plus(A: in Flt; T: in out LTaylor; PolyPsi: in Boolean := False); --- also includes L^2
  procedure Cassini(A: in Flt; T: in out LTaylor);
  procedure Cassini_Plus(A: in Flt; T: in out LTaylor);    --- also includes L^2
  procedure Near_Cassini(A: in Flt; T: in out LTaylor; NLast: in Positive);
  procedure Near_Cassini_Plus(A: in Flt; T: in out LTaylor; NLast: in Positive);    --- also includes L^2

  ------------- modes

  type CylinderMode is      --- a unit coeff or unit error
    record
      R: Radius;            --- domain parameter
      J: EFreqY;            --- CpCosSin1 index
      K: EFreqX;            --- Fourier1 index
      ErrY: Boolean;        --- error Fourier1
      ErrX: Boolean;        --- error CpCosSin1
      Simple: Boolean;      --- Extract used only in Linear.Contr.Prod
    end record;

  type CylinderModes is array (Positive range <>) of CylinderMode; -- should start with index 1

  procedure Norm(M: in CylinderMode; S: in out Scalar);
  procedure MaxNorm(M: in CylinderMode; S: out Radius);
  procedure Assign(M: in CylinderMode; Q: in out Cylinder);
  procedure AddProd(M: in CylinderMode; S: in Scalar; Q: in out Cylinder; Check: in Boolean);
  procedure Extract(M: in CylinderMode; Q: in out Cylinder; S: in out Scalar; Check: in Boolean);
  procedure Show(M: in CylinderMode; NewLine: in Boolean := True);
  function Make(RP: Radius; Siz: Positive; Simple,Trunc: Boolean) return CylinderModes;

  ------------------------ plotting

  function ApproxVal(X,Y: Flt; A: Cylinder) return Flt;
  function ApproxGrad2(X,Y: Flt; A: Cylinder) return Flt;

  --- |Phi'|^2
  function WPhi1(Y: Flt) return Flt;
  function WPhi2(X,Y,R: Flt; N: Integer) return Flt;
  function WPhi3(X,Y,R: Flt) return Flt;
  function WPhi4(X,Y,R: Flt) return Flt;
  function WPhi5(X,Y,R: Flt) return Flt;

  type Point is
    record
      X,Y,Z: Flt;
    end record;

  --- map Phi
  function Phi1(P: Point) return Point;
  function Phi2(P: Point; R: Flt; N: Integer) return Point;
  function Phi3(P: Point; R: Flt) return Point;
  function Phi4(P: Point; R: Flt) return Point;
  function Phi5(P: Point; R: Flt) return Point;
  function Phi6(P: Point; R: Flt) return Point;

  procedure Plot(P: in Point; F: in File_Type);

  Allow_Parallel: Boolean := True;

private

  SZero:      constant Scalar  := Scal(0);
  SHalf:      constant Scalar  := Scal(Half);
  SOne:       constant Scalar  := Scal(1);
  STwo:       constant Scalar  := Scal(2);
  SThree:     constant Scalar  := Scal(3);
  SFour:      constant Scalar  := Scal(4);
  SInfo:  constant Scalar_Info := Info(SZero);
  STrunc:     constant Boolean := SInfo.IsNumeric;
  Not_STrunc: constant Boolean := not STrunc;
  Ell:        constant Scalar  := Scal(EllR);
  Kappa:      constant Scalar  := Pi/Ell;
  InvKappa:   constant Scalar  := Ell/Pi;       -- 1/Kappa
  Kappa2:     constant Scalar  := Sqr(Kappa);   -- Kappa^2
  EllKap2:    constant Scalar  := Sqr(Pi)/Ell;  -- Ell*Kappa^2
  InvEllKap2: constant Scalar  := Ell/Sqr(Pi);  -- 1/(Ell*Kappa^2)
  AbsRot:     constant Natural := abs(Symm);

  function Sigma(QDivKap2: Scalar) return Sums;
  procedure LqC2C(O: in QData; C1: in FPolyY; A1: in CAbs; C2: in out FPolyY);
  procedure LqC2E(O: in QData; A1: in CAbs; E2: in out ErrsY);
  procedure LqE2E_Down(O: in QData; E1: in ErrsY; E2: in out ErrsY);
  procedure LqE2E_Up(O: in QData; E1: in ErrsY; E2: in out ErrsY);

  type PData is
    record
      ZC,ZE: Boolean;
      A: CAbs;
    end record;

  procedure Init(P: in Cylinder; O: in out PData);
  procedure Lq(OQ: in QData; OP: in PData; P1: in Cylinder; P2: in out Cylinder);

  procedure Plain_Val2(P: in Cylinder; Y: in Scalar; V: in out FScalar);
  procedure MultKappa(YL,YR: in Flt; KYL,KYR: out Flt);

  --- for Cassini
  NMax: constant Positive := 1024;
  package NewComb is new Comb(NMax => NMax, Scalar => Scalar);
  use NewComb;
  LMax: constant Natural       := LDeg;                 --- larger than necessary
  BVal: constant Scalar_Vector := AbsChoose(-1,2,LMax); --- coefficients B_n
  CVal: constant Scalar_Vector := AbsChoose(-3,2,LMax); --- coefficients C_n

  function FA(M: Natural; X: Radius) return Flt; --- upper bound on Sum_{n>=m} x^n
  function FB(M: Natural; X: Radius) return Flt; --- upper bound on Sum_{n>=m} B_n x^n
  function FC(M: Natural; X: Radius) return Flt; --- upper bound on Sum_{n>=m} C_n x^n

  --- modes etc
  procedure Extract_Coeff(M: in CylinderMode; Q: in out Cylinder; S: in out Scalar); --- used only in ContrFix
  procedure Fail(M: in CylinderMode; Q: in out Cylinder; S: in out Scalar);
  procedure Extract_ErrorF(M: in CylinderMode; Q: in out Cylinder; S: in out Scalar) renames Fail;
  procedure Extract_ErrorT(M: in CylinderMode; Q: in out Cylinder; S: in out Scalar) renames Fail;
  function SameType(M1,M2: CylinderMode) return Boolean;

  ThreeQuarters: constant Flt := Flt(3)/Flt(4);
  Approx_Pi: constant := 3.14159_26535_89793_23846_26433_83279_50288_41971_69399_37511;
  RKappa: constant Radius := Approx_Pi/EllR;                  -- approx Pi/Ell
  RInvKappa: constant Radius := EllR/Approx_Pi;               -- approx Ell/Pi

  TinyWFac: constant Boolean := WFac=Zero;              --- treat properly at the end
  TwoOrFour: constant Flt := Flt(Choose(TinyWFac,2,4)); --- 2 if TinyWFac, otherwise 4
  STwoOrFour: constant Scalar := Scal(TwoOrFour);

end Cylinders;
