with Globals;
use Globals;

package body Flts.Complexs is

  use Flt_EF;

  function "-"(Z: Complex) return Complex is
  begin
    return (-Z.Re,-Z.Im);
  end "-";

  function "+"(Z1,Z2: Complex) return Complex is
  begin
    return (Z1.Re+Z2.Re,Z1.Im+Z2.Im);
  end "+";

  function "-"(Z1,Z2: Complex) return Complex is
  begin
    return (Z1.Re-Z2.Re,Z1.Im-Z2.Im);
  end "-";

  function "*"(R: Flt; Z: Complex) return Complex is
  begin
    return (R*Z.Re,R*Z.Im);
  end "*";

  function "*"(Z1,Z2: Complex) return Complex is
  begin
    return (Z1.Re*Z2.Re-Z1.Im*Z2.Im,Z1.Re*Z2.Im+Z1.Im*Z2.Re);
  end "*";

  function "/"(R: Flt; Z: Complex) return Complex is
    S: constant Flt := R/(Z.Re*Z.Re+Z.Im*Z.Im);
  begin
    return (S*Z.Re,-S*Z.Im);
  end "/";

  function "/"(Z: Complex; R: Flt) return Complex is
  begin
    return (Z.Re/R,Z.Im/R);
  end "/";

  function "/"(Z1,Z2: Complex) return Complex is
    Q: constant Flt := Z2.Re*Z2.Re+Z2.Im*Z2.Im;
  begin
    return ((Z1.Im*Z2.Im+Z1.Re*Z2.Re)/Q,(Z1.Im*Z2.Re-Z1.Re*Z2.Im)/Q);
  end "/";

  function "**"(Z: Complex; I: Integer) return Complex is
  begin
    if I<0 then return COne/Z**(-I); end if;
    if I=0 then return COne; end if;
    declare
      J: Integer := I;
      U: Complex := Z;
      V: Complex := COne;
    begin
      loop
        if (J rem 2)=1 then V := U*V; end if;
        J := J/2;
        exit when J=0;
        U := U*U;
      end loop;
      return V;
    end;
  end "**";

  function "abs"(Z: Complex) return Flt is
  begin
    return Sqrt(Z.Re*Z.Re+Z.Im*Z.Im);
  end "abs";

  function Adjoint(Z: Complex) return Complex is
  begin
    return (Z.Re,-Z.Im);
  end Adjoint;

  function Sqrt(Z: Complex) return Complex is
    -- eispack style
    -- branch chosen so that W.Re>=0 and sign(W.Im)=sign(Z.Im)
    X: Flt renames Z.Re;
    Y: Flt renames Z.Im;
    S: Flt := Sqrt(Half*(Sqrt(X*X+Y*Y)+abs(X)));
    W: Complex;
  begin
    if X<Zero then
      if Y<Zero then S := -S; end if;
      W.Im := S;
      W.Re := Half*(X/S);
    else
      W.Re := S;
      W.Im := Half*(Y/S);
    end if;
    return W;
  end Sqrt;

  function Exp(Z: Complex) return Complex is
    R: constant Flt := Exp(Z.Re);
  begin
    if Z.Im=Zero then return (R,Zero); end if;
    return (R*Cos(Z.Im),R*Sin(Z.Im));
  end Exp;

  procedure Polar(X,Y: in Flt; Rad,Phi: out Flt) is
    --- Phi between -Pi and Pi
    Pi: constant := 3.14159_26535_89793_23846_26433_83279_50288_41971_69399_37511;
  begin
    if X<Zero then              -- X<0
      Polar(-X,-Y,Rad,Phi);
      if Phi>Zero then
        Phi := Phi-Pi;
      else
        Phi := Phi+Pi;
      end if;
    elsif X>Zero then           -- X>0
      if Y=Zero then
        Rad := X;
        Phi := Zero;
      else
        Rad := Sqrt(X*X+Y*Y);
        if X>abs(Y) then
          Phi := ArcSin(Y/Rad);  -- between -Pi/4 and Pi/4
        elsif Y>Zero then
          Phi := ArcCos(X/Rad);  -- between Pi/4 and Pi/2
        else
          Phi := -ArcCos(X/Rad); -- between -Pi/2 and -Pi/4
        end if;
      end if;
    elsif Y>Zero then            -- X=0, Y>0
      Rad := Y;
      Phi := Half*Pi;
    elsif Y<Zero then            -- X=0, Y<0
      Rad := -Y;
      Phi := NegHalf*Pi;
    else                         -- X=Y=0
      raise Undefined;
    end if;
  end Polar;

  function Log(Z: Complex) return Complex is
    --- principal branch
    W: Complex;
  begin
    Polar(Z.Re,Z.Im,W.Re,W.Im);
    W.Re := Log(W.Re);
    return W;
  end Log;

end Flts.Complexs;
