
generic

  XL: in Positive;           --- horizontal 0 .. XL
  YL: in Positive;           --- vertical 0 .. YL
  Cutoff: in Radius := Zero; --- no minima that close to 0

  with function Xof(M: Natural) return Flt; --- grid x-coordinate
  with function Yof(N: Natural) return Flt; --- grid y-coordinates
  with function Fun(X,Y: Flt) return Flt;   --- given function

package Flts.Minima is --- approximate local minima of Fun

  XFirst: constant Flt := Xof(0);
  XLast:  constant Flt := Xof(XL);
  YFirst: constant Flt := Yof(0);
  YLast:  constant Flt := Yof(YL);

  subtype MRange is Integer range 0 .. XL;
  subtype NRange is Integer range 0 .. YL;
  type Flt_Vec is array (Integer range <>) of Flt;
  type Flt_Mat is array (Integer range <>, Integer range <>) of Flt;
  subtype Values is Flt_Mat (MRange,NRange);

  procedure Fill(Z: out Values);                                         --- make grid Values
  function Fill return Values;                                           --- make grid Values
  function Rough_Local_Minima(Z: Values; Dist: Positive) return Flt_Vec; --- check grid Values
  function Rough_Local_Minima(Dist: Positive) return Flt_Vec;            --- check grid Values
  procedure Improve_Min(V: in out Flt_Vec; Eps: in Radius);              --- uses Nelder-Mead
  function Local_Minima(Z: Values; Dist: Positive; Eps: in Radius) return Flt_Vec; --- combined
  function Local_Minima(Dist: Positive; Eps: in Radius) return Flt_Vec;            --- combined

  procedure NegVal(V: in out Flt_Vec); --- when dealing with local Max
  procedure Show_List(N: in String; V: in Flt_Vec);
  procedure Write(FileName: in String; V: in Flt_Vec; Decimal: in Boolean := False);

  Fill_Parallel: Boolean := True;

private

  procedure Proper_Rounding;
  procedure Center(S: in out Flt);
  procedure Copy(S1: in Flt; S2: in out Flt);
  procedure Swap(S1,S2: in out Flt);
  procedure Add(S1: in Flt; S2: in out Flt);
  procedure Sub(S1: in Flt; S2: in out Flt);
  procedure Sum(S1,S2: in Flt; S3: in out Flt);
  procedure Diff(S1,S2: in Flt; S3: in out Flt);
  procedure Mult(R: in Flt; S: in out Flt);

end Flts.Minima;
