with Ada.Text_IO, Logicals, Ints, Rationals, Flts, Flt_Vectors, ScalInfo;
use Ada.Text_IO, Logicals, Ints, Rationals, Flts, ScalInfo;
with Recycling, Matrices, Vectors, Vectors.More, Vectors.Ops;

pragma Elaborate_All (Ada.Text_IO,Logicals,Ints,Rationals,Flt_Vectors);
pragma Elaborate_All (Recycling,Matrices,Vectors.More);

generic

  PDeg: in Positive;          --- max degree
  Rad:  in Radius  := Two;    --- default domain

  -------------------- begin_include scalars Scalar

  type Scalar is private;

  --- basic
  with function Info(Dummy: Scalar) return Scalar_Info is <>;
  with function IsSharp(S: Scalar) return Boolean is <>;
  with function IsZero(S: Scalar) return Boolean is <>;
  with function "="(S1,S2: Scalar) return Boolean is <>;
  with procedure SetZero(S: in out Scalar) is <>;
  with procedure Copy(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Swap(S1,S2: in out Scalar) is <>;

  --- sets
  with function Center0(S: Scalar) return Boolean is <>;
  with function Contains0(S: Scalar) return Logical is <>;
  with function Contains(S1,S2: Scalar) return Logical is <>; -- S1 constains S2
  with procedure BallAt0(R: in Flt; S: in out Scalar) is <>;
  with function BallAt0(R: Flt) return Scalar is <>;
  with procedure DiskAt0(R: in Flt; S: in out Scalar) is <>;
  with function DiskAt0(R: Flt) return Scalar is <>;
  with procedure ToErr(S: in out Scalar) is <>;
  with procedure ToErr(S1: in Scalar; S2: in out Scalar) is <>;
  with function ToErr(S: Scalar) return Scalar is <>;
  with procedure Center(S: in out Scalar) is <>;
  with procedure Center(S1: in Scalar; S2: in out Scalar) is <>;
  with function Center(S: Scalar) return Scalar is <>;
  with procedure CenterDisk(S: in out Scalar) is <>;
  with procedure CenterDisk(S1: in Scalar; S2: in out Scalar) is <>;
  with function CenterDisk(S: Scalar) return Scalar is <>;
  with procedure ModCenter(S: in out Scalar) is <>;
  with procedure ModCenter(S1: in Scalar; S2: in out Scalar) is <>;
  with function ModCenter(S: Scalar) return Scalar is <>;
  with procedure ModNumbers(S: in out Scalar) is <>;
  with procedure ModNumbers(S1: in Scalar; S2: in out Scalar) is <>;
  with function ModNumbers(S: Scalar) return Scalar is <>;
  with procedure ErrMult(R: in Radius; S: in out Scalar) is <>;
  with procedure Union(S1: in Scalar; S2: in out Scalar) is <>;
  with function Union(S1,S2: Scalar) return Scalar is <>;
  with procedure Intersection(S1: in Scalar; S2: in out Scalar; Empty: out Logical) is <>;

  --- order
  with function Sign(S: Scalar) return Integer is <>;
  with function Compare(S1,S2: Scalar) return Integer is <>;
  with function "<"(S1,S2: Scalar) return Boolean is <>;
  with function "<="(S1,S2: Scalar) return Boolean is <>;
  with function ">"(S1,S2: Scalar) return Boolean is <>;
  with function ">="(S1,S2: Scalar) return Boolean is <>;
  with procedure Min(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Min(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function Min(S1,S2: Scalar) return Scalar is <>;
  with procedure Max(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Max(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function Max(S1,S2: Scalar) return Scalar is <>;
  with function Inf(S: Scalar) return Flt is <>; -- inf of real part
  with function Sup(S: Scalar) return Flt is <>; -- sup of real part

  --- arithmetic
  with procedure Neg(S: in out Scalar) is <>;
  with procedure Neg(S1: in Scalar; S2: in out Scalar) is <>;
  with function "-"(S: Scalar) return Scalar is <>;
  with procedure Add(I: in Integer; S: in out Scalar) is <>;
  with procedure Add(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Sum(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function "+"(S1,S2: Scalar) return Scalar is <>;
  with procedure Sub(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Diff(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function "-"(S1,S2: Scalar) return Scalar is <>;
  with procedure Mult(R: in Flt; S,Tmp: in out Scalar) is <>;
  with procedure Mult(R: in Flt; S: in out Scalar) is <>;
  with procedure Prod(R: in Flt; S1: in Scalar; S2: in out Scalar) is <>;
  with function "*"(R: Flt; S: Scalar) return Scalar is <>;
  with procedure AddProd(R: in Flt; S1: in Scalar; S2,Tmp: in out Scalar) is <>;
  with procedure AddProd(R: in Flt; S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Mult(Q: in Rational; S: in out Scalar) is <>;
  with procedure Mult(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Prod(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function "*"(S1,S2: Scalar) return Scalar is <>;
  with procedure AddProd(S1,S2: in Scalar; S3,Tmp: in out Scalar) is <>;
  with procedure AddProd(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with procedure SubProd(S1,S2: in Scalar; S3,Tmp: in out Scalar) is <>;
  with procedure Div(R: in Flt; S,Tmp: in out Scalar) is <>;
  with procedure Div(R: in Flt; S: in out Scalar) is <>;
  with procedure Quot(S1: in Scalar; R: in Flt; S2: in out Scalar) is <>;
  with function "/"(S: Scalar; R: Flt) return Scalar is <>;
  with procedure Div(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Quot(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function "/"(S1,S2: Scalar) return Scalar is <>;
  with procedure Inv(S1: in Scalar; S2: in out Scalar) is <>;
  with function Inv(S: Scalar) return Scalar is <>;
  with function "**"(S: Scalar; I: Integer) return Scalar is <>;

  --- fun
  with function IsReal(S: Scalar) return Boolean is <>;
  with procedure Adjoint(S: in out Scalar) is <>;
  with procedure Adjoint(S1: in Scalar; S2: in out Scalar) is <>;
  with function Adjoint(S: Scalar) return Scalar is <>;
  with procedure Real_Part(S: in out Scalar) is <>;
  with procedure Imag_Part(S: in out Scalar) is <>;
  with procedure Eval0(S: in out Scalar) is <>;
  with function "abs"(S: Scalar) return Scalar is <>;
  with procedure Norm(S1: in Scalar; S2: in out Scalar) is <>;
  with function Norm(S: Scalar) return Scalar is <>;
  with function MaxNorm(S: Scalar) return Radius is <>;
  with procedure Sqr(S: in out Scalar) is <>;
  with function Sqr(S: Scalar) return Scalar is <>;
  with procedure Sqrt(S1: in Scalar; S2: in out Scalar) is <>;
  with function Sqrt(S: Scalar) return Scalar is <>;
  with procedure Root(K: in Positive; S1: in Scalar; S2: in out Scalar) is <>;
  with function Root(K: Positive; S: Scalar) return Scalar is <>;
  with procedure Exp(S1: in Scalar; S2: in out Scalar) is <>;
  with function Exp(S: Scalar) return Scalar is <>;
  with procedure ArcCos(S1: in Scalar; S2: in out Scalar) is <>;
  with function ArcCos(S: Scalar) return Scalar is <>;
  with procedure ArcSin(S1: in Scalar; S2: in out Scalar) is <>;
  with function ArcSin(S: Scalar) return Scalar is <>;
  with procedure Cos(S1: in Scalar; S2: in out Scalar) is <>;
  with function Cos(S: Scalar) return Scalar is <>;
  with procedure Sin(S1: in Scalar; S2: in out Scalar) is <>;
  with function Sin(S: Scalar) return Scalar is <>;
  with procedure Log(S1: in Scalar; S2: in out Scalar) is <>;
  with function Log(S: Scalar) return Scalar is <>;
  with procedure Simple_Random(S: in out Scalar) is <>;
  with procedure QPower(Q: in Rational; S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Roots2(B,C: in Scalar; U1,U2,V: in out Scalar) is <>;
  with procedure Roots3(B,C,D: in Scalar; U0,U1,U2,V: in out Scalar) is <>;

  --- conversion and i/o
  with function Approx(S: Scalar) return Flt is <>;
  with procedure Assign(K: in Integer; S: in out Scalar) is <>;
  with procedure Assign(Q: in Rational; S: in out Scalar) is <>;
  with procedure Assign(R: in Flt; S: in out Scalar) is <>;
  with function Scal(I: Integer) return Scalar is <>;
  with function Scal(Q: Rational) return Scalar is <>;
  with function Scal(R: Flt) return Scalar is <>;
  with procedure Enclose(R1,R2: in Flt; S: in out Scalar) is <>;
  with function Enclose(R1,R2: Flt) return Scalar is <>;
  with procedure Show1(N: in String; S: in Scalar; Whatever: in Boolean := True) is <>;
  with procedure Show2(N: in String; S1,S2: in Scalar; Whatever: in Boolean := True) is <>;
  with procedure Put(F: in File_Type; S: in Scalar; Decimal: in Boolean := False) is <>;
  with procedure Get(F: in File_Type; S: in out Scalar; Decimal: in Boolean := False) is <>;

  --- misc
  with function Get_Precision(R: Scalar) return Positive is <>;
  with procedure Proper_Rounding is <>;
  with procedure Free_Cache(S: in Scalar) is <>;
  -------------------- end_include scalars

  with package SM is new Matrices (Scalar => Scalar);
  with package SV is new Vectors (Component => Scalar);

package Fouriers1 is  -- Fourier series with non-constant coefficients

  package SV_More is new SV.More;
  package SV_Ops is new SV.Ops (Scalar => Scalar, SM => SM);

  EDeg: constant Positive := 2*PDeg;
  subtype Freq is Integer range -PDeg .. PDeg;
  subtype EFreq is Integer range -EDeg .. EDeg;
  subtype FPoly1 is SV.Vector(Freq);
  subtype Flt_Vector is Flt_Vectors.Vector;
  subtype Errs1 is Flt_Vector(EFreq);       -- assumes rounding up

  PolyFac:  Radius  := One;           -- default domain for polynomials is PolyFac*Rad
  Prod_Parallel:   Boolean := False;  -- run Product etc in parallel

  type Fourier1 is
    record
      C: FPoly1;               -- cos(K.) for K>=0 and sin(-K.) for K<0
      E: Errs1;                -- higher order frequency errors
      R: Radius := Rad;        -- domain of analyticity is |Im z|<Log(R)
    end record;

  procedure Show_F1_Param(Full: in Boolean := True);

  --------------- standard Scalar procedures

  --- basic
  function Info(Dummy: Fourier1) return Scalar_Info;
  function IsSharp(P: Fourier1) return Boolean;         --- polynomial with sharp coeffs
  function IsZero(P: Fourier1) return Boolean;
  function "="(P1,P2: Fourier1) return Boolean;
  procedure SetZero(P: in out Fourier1);
  procedure Copy(P1: in Fourier1; P2: in out Fourier1);
  procedure Swap(P1,P2: in out Fourier1);

  --- sets
  function Center0(P: Fourier1) return Boolean;           --- symmetric about 0
  function Contains0(P: Fourier1) return Logical;
  function Contains(P1,P2: Fourier1) return Logical;
  procedure BallAt0(R: in Flt; P: in out Fourier1);       --- 0-centered ball of radius |R|
  function BallAt0(R: Flt) return Fourier1;               --- 0-centered ball of radius |R|
  procedure DiskAt0(R: in Flt; S: in out Fourier1);
  function DiskAt0(R: Flt) return Fourier1;
  procedure ToErr(P: in out Fourier1);                    --- 0-centered set containing P
  procedure ToErr(P1: in Fourier1; P2: in out Fourier1);  --- 0-centered set containing P1
  function ToErr(P: Fourier1) return Fourier1;            --- 0-centered set containing P
  procedure Center(P: in out Fourier1);
  procedure Center(P1: in Fourier1; P2: in out Fourier1);
  function Center(P: Fourier1) return Fourier1;
  procedure CenterDisk(S: in out Fourier1);
  procedure CenterDisk(S1: in Fourier1; S2: in out Fourier1);
  function CenterDisk(S: Fourier1) return Fourier1;
  procedure ModNumbers(S: in out Fourier1);
  procedure ModNumbers(S1: in Fourier1; S2: in out Fourier1);
  function ModNumbers(S: Fourier1) return Fourier1;
  procedure ModCenter(P: in out Fourier1);
  procedure ModCenter(P1: in Fourier1; P2: in out Fourier1);
  function ModCenter(P: Fourier1) return Fourier1;            --- zero center
  procedure ErrMult(R: in Radius; P: in out Fourier1);
  procedure Union(P1: in Fourier1; P2: in out Fourier1);
  function Union(P1,P2: Fourier1) return Fourier1;
  procedure Intersection(P1: in Fourier1; P2: in out Fourier1; Empty: out Logical);

  --- order
  function Sign(P: Fourier1) return Integer;
  function Compare(P1,P2: Fourier1) return Integer;
  function "<"(P1,P2: Fourier1) return Boolean;
  function "<="(P1,P2: Fourier1) return Boolean;
  function ">="(P1,P2: Fourier1) return Boolean;
  function ">"(P1,P2: Fourier1) return Boolean;
  procedure Min(P1: in Fourier1; P2: in out Fourier1);
  procedure Min(P1,P2: in Fourier1; P3: in out Fourier1);
  function Min(P1,P2: Fourier1) return Fourier1;
  procedure Max(P1: in Fourier1; P2: in out Fourier1);
  procedure Max(P1,P2: in Fourier1; P3: in out Fourier1);
  function Max(P1,P2: Fourier1) return Fourier1;
  function Inf(P: Fourier1) return Flt;
  function Sup(P: Fourier1) return Flt;

  --- addition and multiplication etc.
  procedure Neg(P: in out Fourier1);
  procedure Neg(P1: in Fourier1; P2: in out Fourier1);
  function "-"(P: Fourier1) return Fourier1;
  procedure Add(I: in Integer; P: in out Fourier1);
  procedure Add(P1: in Fourier1; P2: in out Fourier1);
  procedure Sum(P1,P2: in Fourier1; P3: in out Fourier1);
  function "+"(P1,P2: Fourier1) return Fourier1;
  procedure Sub(P1: in Fourier1; P2: in out Fourier1);
  procedure Diff(P1,P2: in Fourier1; P3: in out Fourier1);
  function "-"(P1,P2: Fourier1) return Fourier1;
  procedure Mult(R: in Flt; P,Tmp: in out Fourier1);
  procedure Mult(R: in Flt; P: in out Fourier1; Tmp: in out Scalar);
  procedure Mult(R: in Flt; P: in out Fourier1);
  procedure Prod(R: in Flt; P1: in Fourier1; P2: in out Fourier1);
  function "*"(R: Flt; P: Fourier1) return Fourier1;
  procedure AddProd(R: in Flt; P1: in Fourier1; P2,Tmp: in out Fourier1);
  procedure AddProd(R: in Flt; P1: in Fourier1; P2: in out Fourier1; Tmp: in out Scalar);
  procedure AddProd(R: in Flt; P1: in Fourier1; P2: in out Fourier1);
  procedure Mult(Q: in Rational; P: in out Fourier1);
  procedure Mult(P1: in Fourier1; P2: in out Fourier1);
  procedure Prod(P1,P2: in Fourier1; P3: in out Fourier1);
  function "*"(P1,P2: Fourier1) return Fourier1;
  procedure AddProd(P1,P2: in Fourier1; P3,Tmp: in out Fourier1);
  procedure AddProd(P1,P2: in Fourier1; P3: in out Fourier1);
  procedure SubProd(P1,P2: in Fourier1; P3,Tmp: in out Fourier1);
  procedure Div(R: in Flt; P,Tmp: in out Fourier1);
  procedure Div(R: in Flt; P: in out Fourier1; Tmp: in out Scalar);
  procedure Div(R: in Flt; P: in out Fourier1);
  procedure Quot(P1: in Fourier1; R: in Flt; P2: in out Fourier1);
  function "/"(P: Fourier1; R: Flt) return Fourier1;
  procedure Div(P1: in Fourier1; P2: in out Fourier1);
  procedure Quot(P1,P2: in Fourier1; P3: in out Fourier1);
  function "/"(P1,P2: Fourier1) return Fourier1;
  procedure Inv(P: in out Fourier1);
  procedure Inv(P1: in Fourier1; P2: in out Fourier1);
  function Inv(P: Fourier1) return Fourier1;
  function "**"(P: Fourier1; I: Integer) return Fourier1;

  --- functions
  function IsReal(P: Fourier1) return Boolean;
  procedure Adjoint(P1: in Fourier1; P2: in out Fourier1);
  function Adjoint(P: Fourier1) return Fourier1;
  procedure Adjoint(P: in out Fourier1);
  procedure Real_Part(P: in out Fourier1);
  procedure Imag_Part(P: in out Fourier1);
  procedure Eval0(P: in out Fourier1);
  procedure Norm(P: in Fourier1; Q: in out Fourier1);
  function Norm(P: Fourier1) return Fourier1;
  function "abs"(P: Fourier1) return Fourier1 renames Norm;
  function MaxNorm(P: Fourier1) return Radius;
  procedure Sqr(P: in out Fourier1);
  function Sqr(P: Fourier1) return Fourier1;
  procedure Sqrt(P1: in Fourier1; P2: in out Fourier1);
  function Sqrt(P: Fourier1) return Fourier1;
  procedure Root(K: in Positive; P1: in Fourier1; P2: in out Fourier1);
  function Root(K: Positive; P: Fourier1) return Fourier1;
  procedure Exp(P1: in Fourier1; P2: in out Fourier1);
  function Exp(P: Fourier1) return Fourier1;
  procedure ArcCos(P1: in Fourier1; P2: in out Fourier1);
  function ArcCos(P: Fourier1) return Fourier1;
  procedure ArcSin(P1: in Fourier1; P2: in out Fourier1);
  function ArcSin(P: Fourier1) return Fourier1;
  procedure Cos(P1: in Fourier1; P2: in out Fourier1);
  function Cos(P: Fourier1) return Fourier1;
  procedure Sin(P1: in Fourier1; P2: in out Fourier1);
  function Sin(P: Fourier1) return Fourier1;
  procedure Log(P1: in Fourier1; P2: in out Fourier1);
  function Log(P: Fourier1) return Fourier1;
  procedure Simple_Random(P: in out Fourier1);
  procedure QPower(Q: in Rational; P1: in Fourier1; P2: in out Fourier1);
  procedure Roots2(B,C: in Fourier1; U1,U2,V: in out Fourier1);
  procedure Roots3(B,C,D: in Fourier1; U0,U1,U2,V: in out Fourier1);

  --- conversion and i/o
  function Approx(P: Fourier1) return Flt;
  procedure Assign(I: in Integer; P: in out Fourier1);
  procedure Assign(Q: in Rational; P: in out Fourier1);
  procedure Assign(R: in Flt; P: in out Fourier1);
  function Scal(I: Integer) return Fourier1;
  function Scal(Q: Rational) return Fourier1;
  function Scal(R: Flt) return Fourier1;
  procedure Enclose(R1,R2: in Flt; P: in out Fourier1);
  function Enclose(R1,R2: Flt) return Fourier1;
  procedure Show1(N: in String; P: in Fourier1; Hide0: in Boolean := True);
  procedure Show2(N: in String; P1,P2: in Fourier1; Hide0: in Boolean := True);
  procedure Put(F: in File_Type; P: in Fourier1; Decimal: in Boolean := False);
  procedure Get(F: in File_Type; P: in out Fourier1; Decimal: in Boolean := False);
  procedure Write(FileName: in String; P: in Fourier1; Decimal: in Boolean := False);
  procedure Read(FileName: in String; P: in out Fourier1; Decimal: in Boolean := False);

  --- misc
  function Get_Precision(P: Fourier1) return Positive;
  procedure Proper_Rounding(Dummy: in Fourier1);
  procedure Free_Cache(Dummy: in Fourier1);

  --------------- other standard procedures
  function IsPoly(P: Fourier1) return Boolean;
  function Frequency0(P: Fourier1) return Boolean;
  function IsEven(P: Fourier1) return Boolean;
  function IsOdd(P: Fourier1) return Boolean;
  procedure ProjEven(P: in out Fourier1);                 --- Cos series
  procedure ProjOdd(P: in out Fourier1);                  --- Sin series
  procedure ProjConstant(P: in out Fourier1);             --- P.C(I)=0 if I>0
  procedure ProjSymm(N: in Positive; P: in out Fourier1); --- P.C(I)=0 unless I=0 (mod N)
  procedure ProjASymm(N: in Positive; P: in out Fourier1);--- P.C(I)=0 unless I=N (mod 2N)
  procedure Flip(P: in out Fourier1);
  procedure Rotate(R: in Flt; P1: in Fourier1; P2: in out Fourier1);

  function Rho(P: Fourier1) return Radius;
  procedure SetRho(R: in Radius; P: in out Fourier1; Check: in Boolean := True);
  procedure SetZero(R: in Radius; P: in out Fourier1);

  procedure LiftErrs(P: in out Fourier1; F: in Natural := 0);   -- make general errors start at frequency F
  procedure Assign(S: in Scalar; P: in out Fourier1);
  function Scal(S: Scalar) return Fourier1;

  procedure Average(P: in Fourier1; S: in out Scalar);
  function Average(P: Fourier1) return Scalar;
  procedure Average0(P: in out Fourier1);
  procedure Coeff(I: in Integer; P: in Fourier1; S: in out Scalar);
  function Coeff(I: Integer; P: Fourier1) return Scalar;
  procedure AddCoeff(I: in Integer; S: in Scalar; P: in out Fourier1);    -- P[I] := P[I] + S
  procedure AddBall(R: in Radius; P: in out Fourier1);
  procedure Add(S: in Scalar; P: in out Fourier1);
  procedure Mult(S: in Scalar; P: in out Fourier1);                       -- P := S*P
  procedure Prod(S: in Scalar; P1: in Fourier1; P2: in out Fourier1);     -- P2 := S*P1;
  function "*"(S: Scalar; P: Fourier1) return Fourier1;                   -- return S*P
  procedure AddProd(S: in Scalar; P1: in Fourier1; P2: in out Fourier1);  -- P2 := P2+S*P1

  procedure CSPowers(X: in Scalar; C: in out FPoly1);
  procedure Val(P: in Fourier1; X: in Scalar; Y: in out Scalar; Tmp: in out FPoly1);
  procedure Val(P: in Fourier1; X: in Scalar; Y: in out Scalar);
  function Val(P: Fourier1; S: Scalar) return Scalar;
  procedure Val(P1: in Fourier1; S: in Scalar; P2: in out Fourier1);
  procedure Norm1(P: in Fourier1; S: in out Scalar);
  function Norm1(P: Fourier1) return Scalar;
  procedure Monom(I: in Integer; B: in Scalar; P: in out Fourier1);        -- B*CosSin(I*X)
  function Monom(I: Integer; B: Flt) return Fourier1;
  procedure AplusMonom(I: in Integer; A,B: in Scalar; P: in out Fourier1); -- A+B*CosSin(I*X)
  function AplusMonom(I: Integer; A,B: Flt) return Fourier1;
  procedure HiUnitBall(I: in Integer; P: in out Fourier1; R: in Radius := Rad);
  procedure Compose(C: in FPoly1; P1: in Fourier1; P2: in out Fourier1);
  procedure CosSin(P: in Fourier1; PC,PS: in out Fourier1);
  procedure Der(C: in out FPoly1);
  procedure Der2(C: in out FPoly1);
  procedure Der(P: in out Fourier1; R: in Radius);
  procedure AntiDer(C: in out FPoly1; CheckAverage: in Boolean := True);
  procedure AntiDer(P: in out Fourier1; CheckAverage: in Boolean := True);
  procedure AntiDer(P1: in Fourier1; P2: in out Fourier1; CheckAverage: in Boolean := True);
  function Integral(P: Fourier1; S1,S2: Scalar) return Scalar;
  procedure InvHelm(M: in Scalar; F: in out Fourier1);                         --- 1/(-D^2+M)
  procedure InvHelm(M: in Scalar; F1: in Fourier1; F2: in out Fourier1);       --- 1/(-D^2+M)

  --- value stuff
  procedure ValAt0(P: in Fourier1; S: in out Scalar);
  procedure ValAtPi(P: in Fourier1; S: in out Scalar);
  procedure IVal(P: in Fourier1; XL,XR: in Flt; Z: in out Scalar; Steps: in Positive);
  function IVal_Range(P: Fourier1; XL,XR,ZMin,ZMax: Flt; Steps: Positive) return Boolean;
  procedure LevelStat(P: in Fourier1; Cut: in Flt; Steps: in Positive);

  --- purely numerical
  function ApproxVal(X: Flt; P: Fourier1) return Flt;
  function ApproxValDer(X: Flt; P: Fourier1) return Flt;
  function GuessRho(P: Fourier1) return Radius;
  procedure Test_Some_Ops;

  type Fourier1_Access is access Fourier1;

private

  subtype Weights is Flt_Vector(0 .. EDeg);

  procedure Show_F1_Rounding;

  procedure AddProd(S: in Scalar; C1: in FPoly1; C2: in out FPoly1; Tmp: in out Scalar);
  procedure SubProd(S: in Scalar; C1: in FPoly1; C2: in out FPoly1; Tmp: in out Scalar);
  function QuasiDeg(C: FPoly1) return Natural;
  function QuasiDeg(E: Errs1) return Natural;
  function UpSum(E: Errs1) return Radius;
  function UpSumEven(E: Errs1) return Radius;
  function UpSumOdd(E: Errs1) return Radius;
  function IsEven(C: FPoly1) return Boolean;
  function IsOdd(C: FPoly1) return Boolean;
  procedure ErrProd(P1,P2: in Fourier1; R3: in Radius; E3: out Errs1);
  procedure Mult(P1: in Fourier1; P2,Tmp: in out Fourier1);
  procedure Norm1(R: in Radius; C: in FPoly1; S: in out Scalar);
  function TrueF(E: Errs1) return Natural;
  function UpInvNormFac(R: Radius; I: Integer) return Radius;
  function UpInvNormFacs(R: Radius) return Weights;
  function ValAt0(R: Radius; E: Errs1) return Radius;
  function Val(R: Radius; E: Errs1) return Radius;
  function UpNormFac(R: Radius; I: Integer) return Radius;
  procedure AddErr(I: in Integer; R: in Flt; E: in out Errs1);
  procedure ValAt0(P: in Fourier1; S: in out Scalar; D: in Natural; Is_Poly: in Boolean);
  procedure EvenCoeff(I: in Natural; P: in Fourier1; S: in out Scalar);
  procedure OddCoeff(I: in Negative; P: in Fourier1; S: in out Scalar);
  function Simple(P: Fourier1) return Boolean;
  procedure NewQuot(S0,S1: in Scalar; Q: in out Flt; N: in out Integer);

  procedure Exp_Split(P1: in Fourier1; P2: in out Fourier1; S: in out Scalar);
  procedure CosSin_Split(P1: in Fourier1; P2: in out Fourier1; C,S: in out Scalar);
  procedure Unused1(P1: in Fourier1; P2: in out Fourier1; S: in out Scalar);
  procedure Unused2(Q: in Rational; P1: in Fourier1; P2: in out Fourier1; S: in out Scalar);
  function ExpTerms(R: Radius; Prec: Positive) return Natural;
  procedure Approx_Inv(P: in out Fourier1);
  procedure Approx_Sqrt(P: in out Fourier1);
  procedure Approx_Root(K: in Positive; P: in out Fourier1);
  procedure Approx_ArcCos(P: in out Fourier1);
  procedure Approx_ArcSin(P: in out Fourier1);

  SZero:      constant Scalar      := Scal(0);
  SHalf:      constant Scalar      := Scal(Half);
  SOne:       constant Scalar      := Scal(1);
  SThree:     constant Scalar      := Scal(3);
  SInfo:      constant Scalar_Info := Info(SZero);
  STrunc:     constant Boolean     := SInfo.IsNumeric;
  Not_STrunc: constant Boolean     := not STrunc;
  Std:     constant Boolean     := SInfo.IsStandard;

  FInfo: constant Scalar_Info := (RepType    => SInfo.RepType,
                                  IsStandard => SInfo.IsStandard,
                                  IsNumeric  => SInfo.IsNumeric,
                                  IsComplex  => SInfo.IsComplex,
                                  NReps      => SInfo.NReps*(2*PDeg+1),
                                  RMode      => SInfo.RMode);

  package Alloc is new Recycling (Item => Fourier1);
  subtype LF_Pointer is Alloc.LD_Pointer;
  Pool: Alloc.Protected_Data;

end Fouriers1;
