with Globals, Strings, Ints;
use Globals, Strings, Ints;

pragma Elaborate_All (Globals,Strings,Ints);

package body LTaylors is

  use SP;

  protected body Protected_Initialize is
    procedure Initialize is
    begin
      for Q in 0 .. WLast loop OpNorm(Q) := LMaxNorm(Q); end loop;
      VarRho := LVarRho(WLast+1);
    end Initialize;
  end Protected_Initialize;

  procedure Show_OpNorms is
    use Flt_EF;
    R,MinR: Flt;
  begin
    if VarRho=Zero then OpNorms.Initialize; end if;
    R:= OpNorm(1);
    Show1("  1   ",R);
    MinR:= R;
    for Q in 2 .. WLast loop
      R:= Exp(Log(OpNorm(Q))/Flt(Q)); -- Q-th root of OpNorm(Q)
      if R<MinR then
        Show1(Strng(Q,3) & " i ",R);
        MinR:= R;
      else
        Show1(Strng(Q,3) & "   ",R);
      end if;
      Show0("These are the q-th roots of |L^q|, i means improved.");
    end loop;
    Show1("Higher order q-th roots of |L^q|: ",VarRho);
  end Show_OpNorms;

  --- basic

  function IsConst(P: LTaylor) return Boolean is
    C: Poly1 renames P.C;
  begin
    for I in 1 .. PDeg loop
      if not IsZero(C(I)) then return False; end if;
    end loop;
    return STrunc or else P.E=Zero;
  end IsConst;

  function IsPoly(P: LTaylor) return Boolean is
  begin
    return STrunc or else P.E=Zero;
  end IsPoly;

  function IsZero(P: LTaylor) return Boolean is
  begin
    return IsZero(P.C) and then (STrunc or else P.E=Zero);
  end IsZero;

  procedure SetZero(P: in out LTaylor) is
  begin
    SetZero(P.C);
    P.E := Zero;
  end SetZero;

  procedure Copy(P1: in LTaylor; P2: in out LTaylor) is
  begin
    Copy(P1.C,P2.C);
    if Not_STrunc then P2.E := P1.E; end if;
  end Copy;

  procedure Swap(P1,P2: in out LTaylor) is
    E: Radius;
  begin
    Swap(P1.C,P2.C);
    if Not_STrunc then
      E := P1.E;
      P1.E := P2.E;
      P2.E := E;
    end if;
  end Swap;

  --- addition and multiplication etc.

  procedure Neg(P: in out LTaylor) is
  begin
    Neg(P.C);
  end Neg;

  procedure Neg(P1: in LTaylor; P2: in out LTaylor) is
  begin
    Neg(P1.C,P2.C);
    if Not_STrunc then P2.E := P1.E; end if;
  end Neg;

  procedure Add(I: in Integer; P: in out LTaylor) is
  begin
    Add(I,P.C(0));
  end Add;

  procedure Add(P1: in LTaylor; P2: in out LTaylor) is
  begin
    Add(P1.C,P2.C);
    if Not_STrunc then P2.E := P1.E+P2.E; end if;
  end Add;

  procedure Sum(P1,P2: in LTaylor; P3: in out LTaylor) is
  begin
    Sum(P1.C,P2.C,P3.C);
    if Not_STrunc then P3.E := P1.E+P2.E; end if;
  end Sum;

  procedure Sub(P1: in LTaylor; P2: in out LTaylor) is
  begin
    Sub(P1.C,P2.C);
    if Not_STrunc then P2.E := P1.E+P2.E; end if;
  end Sub;

  procedure Diff(P1,P2: in LTaylor; P3: in out LTaylor) is
  begin
    Diff(P1.C,P2.C,P3.C);
    if Not_STrunc then P3.E := P1.E+P2.E; end if;
  end Diff;

  procedure Mult(R: in Flt; P: in out LTaylor) is
    C: Poly1 renames P.C;
    Tmp: Scalar;
  begin
    for I in Power loop Mult(R,C(I),Tmp); end loop;
    if Not_STrunc then P.E := abs(R)*P.E; end if;
  end Mult;

  procedure Mult(S: in Scalar; P: in out LTaylor) is
    C: Poly1 renames P.C;
  begin
    for I in Power loop Mult(S,C(I)); end loop;
    if Not_STrunc then P.E := MaxNorm(S)*P.E; end if;
  end Mult;

  procedure Prod(R: in Flt; P1: in LTaylor; P2: in out LTaylor) is
    C1: Poly1 renames P1.C;
    C2: Poly1 renames P2.C;
  begin
    for I in Power loop Prod(R,C1(I),C2(I)); end loop;
    if Not_STrunc then P2.E := abs(R)*P1.E; end if;
  end Prod;

  procedure Prod(S: in Scalar; P1: in LTaylor; P2: in out LTaylor) is
    C1: Poly1 renames P1.C;
    C2: Poly1 renames P2.C;
  begin
    for I in Power loop Prod(S,C1(I),C2(I)); end loop;
    if Not_STrunc then P2.E := MaxNorm(S)*P1.E; end if;
  end Prod;

  procedure AddProd(S: in Scalar; P1: in LTaylor; P2: in out LTaylor) is
    C1: Poly1 renames P1.C;
    C2: Poly1 renames P2.C;
    Tmp: Scalar;
  begin
    for I in Power loop AddProd(S,C1(I),C2(I),Tmp); end loop;
    if Not_STrunc then P2.E := P2.E+MaxNorm(S)*P1.E; end if;
  end AddProd;

  procedure SubProd(S: in Scalar; P1: in LTaylor; P2: in out LTaylor) is
    C1: Poly1 renames P1.C;
    C2: Poly1 renames P2.C;
    Tmp: Scalar;
  begin
    for I in Power loop SubProd(S,C1(I),C2(I),Tmp); end loop;
    if Not_STrunc then P2.E := P2.E+MaxNorm(S)*P1.E; end if;
  end SubProd;

  procedure Mult(P1: in LTaylor; P2,Tmp: in out LTaylor) is
  begin
    SetZero(Tmp);
    AddProd(P1,P2,Tmp);
    Copy(Tmp,P2);
  end Mult;

  procedure Mult(P1: in LTaylor; P2: in out LTaylor) is
    P3: LTaylor;
  begin
    SetZero(P3);
    AddProd(P1,P2,P3);
    Copy(P3,P2);
  end Mult;

  procedure Prod(P1,P2: in LTaylor; P3: in out LTaylor) is
  begin
    SetZero(P3);
    AddProd(P1,P2,P3);
  end Prod;

  procedure CCAddProd(C1,C2: in Poly1; A1,A2: CAbs; P3: in out LTaylor) is
    C3: Poly1 renames P3.C;
    E3: Flt renames P3.E;
    Tmp: Scalar;
  begin
    for I in Power loop
      if A1(I)/=Zero then
        for J in Power loop
          if A2(J)/=Zero then
            if I+J <= PDeg then
              AddProd(C1(I),C2(J),C3(I+J),Tmp);
            elsif Not_STrunc then
              E3 := E3+A1(I)*A2(J)*OpNorm(I+J);
            end if;
          end if;
        end loop;
      end if;
    end loop;
  end CCAddProd;

  procedure CEAddProd(A1: in CAbs; E2: in Radius; E3: in out Radius) is
  begin
    for I in Power loop
      if A1(I)/=Zero then
        E3 := E3+A1(I)*OpNorm(I)*E2;
      end if;
    end loop;
  end CEAddProd;

  procedure AddProd(P1,P2: in LTaylor; P3: in out LTaylor) is
  begin
    if not (IsZero(P1) or else IsZero(P2)) then
      declare
        C1: Poly1 renames P1.C;
        C2: Poly1 renames P2.C;
        A1,A2: CAbs;
      begin
        for I in Power loop
          A1(I) := MaxNorm(C1(I));
          A2(I) := MaxNorm(C2(I));
        end loop;
        if VarRho=Zero then OpNorms.Initialize; end if;
        CCAddProd(C1,C2,A1,A2,P3);
        if Not_STrunc then
          if P2.E/=Zero then
            CEAddProd(A1,P2.E,P3.E);
          end if;
          if P1.E/=Zero then
            CEAddProd(A2,P1.E,P3.E);
            if P2.E/=Zero then P3.E := P3.E+P1.E*P2.E; end if;
          end if;
        end if;
      end;
    end if;
  end AddProd;

  procedure SubProd(P1,P2: in LTaylor; P3: in out LTaylor) is
  begin
    Neg(P3);
    AddProd(P1,P2,P3);
    Neg(P3);
  end SubProd;

  procedure LMult(Q: in Power; P: in out LTaylor) is
    --- multiply by L^Q
    C: Poly1 renames P.C;
    E: Radius renames P.E;
  begin
    if Q>0 then
      if Not_STrunc then
        if VarRho=Zero then OpNorms.Initialize; end if;
        for I in 1 .. Q loop
          if not IsZero(C(PDeg1-I)) then
            E := E+MaxNorm(C(PDeg1-I))*OpNorm(PDeg1);
          end if;
        end loop;
      end if;
      for I in reverse Q .. PDeg loop
        Copy(C(I-Q),C(I));
      end loop;
      for I in 0 .. Q-1 loop SetZero(C(I)); end loop;
    end if;
  end LMult;

  procedure LProd(Q: in Power; P1: in LTaylor; P2: in out LTaylor) is
    --- multiply by L^Q
    C1: Poly1 renames P1.C;
    C2: Poly1 renames P2.C;
    E2: Radius renames P2.E;
  begin
    if Q=0 then
      Copy(P1,P2);
    else
      if Not_STrunc then
        E2 := P1.E;
        if VarRho=Zero then OpNorms.Initialize; end if;
        for I in 1 .. Q loop
          if not IsZero(C1(PDeg1-I)) then
            E2 := E2+MaxNorm(C1(PDeg1-I))*OpNorm(PDeg1);
          end if;
        end loop;
      end if;
      for I in reverse Q .. PDeg loop
        Copy(C1(I-Q),C2(I));
      end loop;
      for I in 0 .. Q-1 loop SetZero(C2(I)); end loop;
    end if;
  end LProd;

  --- functions

  procedure Inv(P1: in LTaylor; P2: in out LTaylor) is
    procedure Fac(Dummy: in Positive; S: in out Scalar) is
      pragma Unreferenced(Dummy);
    begin
      Assign(-1,S);
    end Fac;
    procedure InvSeries is new Series (Fac => Fac);
    Decreasing: constant Boolean := False;
    OnePlus: constant Boolean := P1.C(0)=SOne;
    S: Scalar;
    P: LTaylor;
  begin
    if OnePlus then
      Copy(P1,P);
    else
      Inv(P1.C(0),S);                       -- S=1/P1(0)
      Prod(S,P1,P);
    end if;
    SetZero(P.C(0));                        -- P=P1/P1(0)-1
    InvSeries(P,P2,Decreasing);             -- 1/(1+P)
    if not OnePlus then Mult(S,P2); end if;
  end Inv;

  procedure Exp(P1: in LTaylor; P2: in out LTaylor) is
    procedure Fac(I: in Positive; S: in out Scalar) is begin Quot(SOne,Flt(I),S); end Fac;
    procedure ExpSeries is new Series (Fac => Fac);
    Decreasing: constant Boolean := True;
    S: Scalar;
    P: LTaylor;
  begin
    Copy(P1,P);
    SetZero(P.C(0));
    ExpSeries(P,P2,Decreasing);        -- Exp(P)
    if not IsZero(P1.C(0)) then
      Exp(P1.C(0),S);
      Mult(S,P2);
    end if;
  end Exp;

  procedure Log(P1: in LTaylor; P2: in out LTaylor) is
    --- not optimal with generic Series
    procedure Fac(I: in Positive; S: in out Scalar) is
      R: constant Rational := (-I)/(I+1);
    begin
      Assign(R,S);
    end Fac;
    procedure LogSeries is new Series (Fac => Fac);
    Decreasing: constant Boolean := False;
    OnePlus: constant Boolean := P1.C(0)=SOne;
    S: Scalar;
    P: LTaylor;
  begin
    if OnePlus then
      Copy(P1,P);
    else
      Inv(P1.C(0),S);           -- S=1/P1(0)
      Prod(S,P1,P);
    end if;
    SetZero(P.C(0));            -- P=P1/P1(0)-1
    LogSeries(P,P2,Decreasing); -- Log(1+P)
    if not OnePlus then
      Log(P1.C(0),S);
      Add(S,P2.C(0));
    end if;
  end Log;

  procedure QPower(Q: in Rational; P1: in LTaylor; P2: in out LTaylor) is
    N: constant Integer := INum(Q);
    D: constant Integer := IDen(Q);
    procedure Fac(I: in Positive; S: in out Scalar) is
      R: constant Rational := (N-D*(I-1))/(D*I);
    begin
      Assign(R,S);
    end Fac;
    procedure QPowerSeries is new Series (Fac => Fac);
    Decreasing: constant Boolean := False;
    OnePlus: constant Boolean := P1.C(0)=SOne;
    S: Scalar;
    P: LTaylor;
  begin
    if abs(N)>D then raise Not_Implemented; end if;
    if OnePlus then
      Copy(P1,P);
    else
      Inv(P1.C(0),S);              -- S=1/P1(0)
      Prod(S,P1,P);
    end if;
    SetZero(P.C(0));               -- P=P1/P1(0)-1
    QPowerSeries(P,P2,Decreasing); -- (1+P)^Q
    if not OnePlus then
      QPower(Q,P1.C(0),S);
      Mult(S,P2);
    end if;
  end QPower;

  function MaxNorm(P: LTaylor) return Radius is
    C: Poly1 renames P.C;
    R: Radius;
  begin
    if STrunc then R := Zero; else R := P.E; end if;
    if VarRho=Zero then OpNorms.Initialize; end if;
    for I in reverse Power loop
      if not IsZero(C(I)) then
        R := R+MaxNorm(C(I))*OpNorm(I);
      end if;
    end loop;
    return R;
  end MaxNorm;

  --- conversion and i/o

  procedure Assign(R: in Flt; P: in out LTaylor) is
  begin
    SetZero(P);
    Assign(R,P.C(0));
  end Assign;

  procedure Assign(S: in Scalar; P: in out LTaylor) is
  begin
    SetZero(P);
    Copy(S,P.C(0));
  end Assign;

  procedure AddCoeff(I: in Natural; S: in Scalar; P: in out LTaylor) is
  begin
    if IsZero(S) then
      null;
    elsif I <= PDeg then
      Add(S,P.C(I));
    elsif Not_STrunc then
      if VarRho=Zero then OpNorms.Initialize; end if;
      if I <= WLast then
        P.E := P.E+MaxNorm(S)*OpNorm(I);
      else
        P.E := P.E+MaxNorm(S)*NPower(VarRho,I);
      end if;
    end if;
  end AddCoeff;

  procedure Show1(N: in String; P: in LTaylor; Hide0: in Boolean := True) is
  begin
    if not IsZero(P) then
      Show1(N & "C ",P.C);
      if Not_STrunc then
        Show1(N & "E ",P.E);
      end if;
    elsif not Hide0 then
      Show0(N & "0");
    end if;
  end Show1;

  procedure Put(F: in File_Type; P: in LTaylor; Decimal: in Boolean := False) is
    C: Poly1 renames P.C;
    Deg: constant Integer := EffDeg(C);
  begin
    Put(F,Deg);
    for I in 0 .. Deg loop
      Put(F,C(I),Decimal);
    end loop;
    Put(F,PDeg1);
    if STrunc then
      Put(F,Zero,Decimal);
    else
      Put(F,P.E,Decimal);
    end if;
  end Put;

  procedure Get(F: in File_Type; P: in out LTaylor; Decimal: in Boolean := False) is
    Deg,EDeg: Integer := 0;
    S: Scalar;
  begin
    SetZero(P);
    Get(F,Deg);
    for I in 0 .. Deg loop
      Get(F,S,Decimal);
      AddCoeff(I,S,P);
    end loop;
    Get(F,EDeg);
    Get(F,P.E);
    if P.E=Zero then
      null;
    elsif STrunc then
      P.E := Zero;
    elsif EDeg<PDeg1 then
      raise Not_Implemented; -- no errors below PDeg1
    end if;
  end Get;

  procedure Write(FileName: in String; P: in LTaylor; Decimal: in Boolean := False) is
    F: File_Type;
  begin
    if Verbosity>0 then Show0("Writing " & FileName); end if;
    Create(F,Out_File,FileName);
    Put(F,P,Decimal);
    Close(F);
  end Write;

  procedure Read(FileName: in String; P: in out LTaylor; Decimal: in Boolean := False) is
    F: File_Type;
  begin
    if Verbosity>1 then Show0("Reading " & FileName); end if;
    Open(F,In_File,FileName);
    Get(F,P,Decimal);
    Close(F);
  end Read;

  ---------------------------------- series

  procedure Series(X: in LTaylor; Y: in out LTaylor; Decreasing: in Boolean) is
    R,S,C: Radius;
    CN,Tmp: Scalar;
    P: array(1 .. PDeg) of LTaylor; --- powers of X
  begin
    if not IsZero(X.C(0)) then raise Not_Implemented; end if;
    Copy(X,P(1));
    for N in 2 .. PDeg loop Prod(P(N/2),P(N-N/2),P(N)); end loop;
    if Not_STrunc then
      R := MaxNorm(P(PDeg));        --- norm of X^PDeg
      if R>=One then
        Show1("Series: R =",R);
        raise Domain_Error;
      end if;
      S := R;
      for N in reverse 1 .. PDeg-1 loop S := S+MaxNorm(P(N)); end loop;
      S := S+One;                   --- sum of norms of X^N for N<=PDeg
    end if;
    Assign(1,CN);
    for N in 1 .. PDeg loop
      Fac(N,Tmp);
      Mult(Tmp,CN);
      Mult(CN,P(N));
    end loop;
    SetZero(Y);
    for N in reverse 1 .. PDeg loop Add(P(N),Y); end loop;
    Assign(1,Y.C(0));
    if Not_STrunc then
      if Decreasing then C := MaxNorm(CN); else C := One; end if;
      Y.E := Y.E+C*S*(-R)/(R-One);
    end if;
  end Series;

end LTaylors;
