with Ada.Text_IO, Logicals, Rationals, Flts, Flt_Vectors, ScalInfo;
use Ada.Text_IO, Logicals, Rationals, Flts, ScalInfo;
with Vectors, Polynoms;

pragma Elaborate_All (Ada.Text_IO,Logicals,Rationals,Flt_Vectors,ScalInfo);
pragma Elaborate_All (Vectors,Polynoms);

generic

  PDeg: in Natural;      --- max degree

  -------------------- begin_include scalars Scalar

  type Scalar is private;
  with procedure Proper_Rounding is <>;

  --- basic
  with function Info(Dummy: Scalar) return Scalar_Info is <>;
  with function IsSharp(S: Scalar) return Boolean is <>;
  with function IsZero(S: Scalar) return Boolean is <>;
  with function "="(S1,S2: Scalar) return Boolean is <>;
  with procedure SetZero(S: in out Scalar) is <>;
  with procedure Copy(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Swap(S1,S2: in out Scalar) is <>;

  --- sets
  with function Center0(S: Scalar) return Boolean is <>;
  with function Contains0(S: Scalar) return Logical is <>;
  with function Contains(S1,S2: Scalar) return Logical is <>;
  with procedure BallAt0(R: in Flt; S: in out Scalar) is <>;
  with function BallAt0(R: Flt) return Scalar is <>;
  with procedure DiskAt0(R: in Flt; S: in out Scalar) is <>;
  with procedure ToErr(S: in out Scalar) is <>;
  with procedure ToErr(S1: in Scalar; S2: in out Scalar) is <>;
  with function ToErr(S: Scalar) return Scalar is <>;
  with procedure Center(S: in out Scalar) is <>;
  with procedure Center(S1: in Scalar; S2: in out Scalar) is <>;
  with function Center(S: Scalar) return Scalar is <>;
  with procedure ModCenter(S: in out Scalar) is <>;
  with procedure ModCenter(S1: in Scalar; S2: in out Scalar) is <>;
  with function ModCenter(S: Scalar) return Scalar is <>;
  with procedure ErrMult(R: in Radius; S: in out Scalar) is <>;
  with procedure Union(S1: in Scalar; S2: in out Scalar) is <>;
  with function Union(S1,S2: Scalar) return Scalar is <>;
  with procedure Intersection(S1: in Scalar; S2: in out Scalar; Empty: out Logical) is <>;

  --- order
  with function Sign(S: Scalar) return Integer is <>;
  with function Compare(S1,S2: Scalar) return Integer is <>;
  with function "<"(S1,S2: Scalar) return Boolean is <>;
  with function "<="(S1,S2: Scalar) return Boolean is <>;
  with function ">"(S1,S2: Scalar) return Boolean is <>;
  with function ">="(S1,S2: Scalar) return Boolean is <>;
  with procedure Min(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Min(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function Min(S1,S2: Scalar) return Scalar is <>;
  with procedure Max(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Max(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function Max(S1,S2: Scalar) return Scalar is <>;
  with function Inf(S: Scalar) return Flt is <>; -- inf of real part
  with function Sup(S: Scalar) return Flt is <>; -- sup of real part

  --- arithmetic
  with procedure Neg(S: in out Scalar) is <>;
  with procedure Neg(S1: in Scalar; S2: in out Scalar) is <>;
  with function "-"(S: Scalar) return Scalar is <>;
  with procedure Add(I: in Integer; S: in out Scalar) is <>;
  with procedure Add(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Sum(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function "+"(S1,S2: Scalar) return Scalar is <>;
  with procedure Sub(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Diff(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function "-"(S1,S2: Scalar) return Scalar is <>;
  with procedure Mult(R: in Flt; S,Tmp: in out Scalar) is <>;
  with procedure Mult(R: in Flt; S: in out Scalar) is <>;
  with procedure Prod(R: in Flt; S1: in Scalar; S2: in out Scalar) is <>;
  with function "*"(R: Flt; S: Scalar) return Scalar is <>;
  with procedure AddProd(R: in Flt; S1: in Scalar; S2,Tmp: in out Scalar) is <>;
  with procedure AddProd(R: in Flt; S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Mult(Q: in Rational; S: in out Scalar) is <>;
  with procedure Mult(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Prod(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function "*"(S1,S2: Scalar) return Scalar is <>;
  with procedure AddProd(S1,S2: in Scalar; S3,Tmp: in out Scalar) is <>;
  with procedure AddProd(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with procedure SubProd(S1,S2: in Scalar; S3,Tmp: in out Scalar) is <>;
  with procedure Div(R: in Flt; S,Tmp: in out Scalar) is <>;
  with procedure Div(R: in Flt; S: in out Scalar) is <>;
  with procedure Quot(S1: in Scalar; R: in Flt; S2: in out Scalar) is <>;
  with function "/"(S: Scalar; R: Flt) return Scalar is <>;
  with procedure Div(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Quot(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function "/"(S1,S2: Scalar) return Scalar is <>;
  with procedure Inv(S1: in Scalar; S2: in out Scalar) is <>;
  with function Inv(S: Scalar) return Scalar is <>;
  with function "**"(S: Scalar; I: Integer) return Scalar is <>;

  --- fun
  with function IsReal(S: Scalar) return Boolean is <>;
  with procedure Adjoint(S: in out Scalar) is <>;
  with procedure Adjoint(S1: in Scalar; S2: in out Scalar) is <>;
  with function Adjoint(S: Scalar) return Scalar is <>;
  with procedure Real_Part(S: in out Scalar) is <>;
  with procedure Imag_Part(S: in out Scalar) is <>;
  with procedure Eval0(S: in out Scalar) is <>;
  with function "abs"(S: Scalar) return Scalar is <>;
  with procedure Norm(S1: in Scalar; S2: in out Scalar) is <>;
  with function Norm(S: Scalar) return Scalar is <>;
  with function MaxNorm(S: Scalar) return Radius is <>;
  with procedure Sqr(S: in out Scalar) is <>;
  with function Sqr(S: Scalar) return Scalar is <>;
  with procedure Sqrt(S1: in Scalar; S2: in out Scalar) is <>;
  with function Sqrt(S: Scalar) return Scalar is <>;
  with procedure Root(K: in Positive; S1: in Scalar; S2: in out Scalar) is <>;
  with function Root(K: Positive; S: Scalar) return Scalar is <>;
  with procedure Exp(S1: in Scalar; S2: in out Scalar) is <>;
  with function Exp(S: Scalar) return Scalar is <>;
  with procedure ArcCos(S1: in Scalar; S2: in out Scalar) is <>;
  with function ArcCos(S: Scalar) return Scalar is <>;
  with procedure ArcSin(S1: in Scalar; S2: in out Scalar) is <>;
  with function ArcSin(S: Scalar) return Scalar is <>;
  with procedure Cos(S1: in Scalar; S2: in out Scalar) is <>;
  with function Cos(S: Scalar) return Scalar is <>;
  with procedure Sin(S1: in Scalar; S2: in out Scalar) is <>;
  with function Sin(S: Scalar) return Scalar is <>;
  with procedure Log(S1: in Scalar; S2: in out Scalar) is <>;
  with function Log(S: Scalar) return Scalar is <>;
  with procedure Simple_Random(S: in out Scalar) is <>;
  with procedure QPower(Q: in Rational; S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Roots2(B,C: in Scalar; U1,U2,V: in out Scalar) is <>;
  with procedure Roots3(B,C,D: in Scalar; U0,U1,U2,V: in out Scalar) is <>;

  --- conversion and i/o
  with function Approx(S: Scalar) return Flt is <>;
  with procedure Assign(K: in Integer; S: in out Scalar) is <>;
  with procedure Assign(Q: in Rational; S: in out Scalar) is <>;
  with procedure Assign(R: in Flt; S: in out Scalar) is <>;
  with function Scal(I: Integer) return Scalar is <>;
  with function Scal(Q: Rational) return Scalar is <>;
  with function Scal(R: Flt) return Scalar is <>;
  with procedure Enclose(R1,R2: in Flt; S: in out Scalar) is <>;
  with function Enclose(R1,R2: Flt) return Scalar is <>;
  with procedure Show1(N: in String; S: in Scalar; Whatever: in Boolean := True) is <>;
  with procedure Show2(N: in String; S1,S2: in Scalar; Whatever: in Boolean := True) is <>;
  with procedure Put(F: in File_Type; S: in Scalar; Decimal: in Boolean := False) is <>;
  with procedure Get(F: in File_Type; S: in out Scalar; Decimal: in Boolean := False) is <>;

  --- misc
  with function Get_Precision(R: Scalar) return Positive is <>;

  -------------------- end_include scalars

  with function LMaxNorm(Q: Natural) return Radius is <>; -- upper bound on ||L^q||
  with function LVarRho(P: Positive) return Radius is <>; -- upper bound on ||L^q||^{1/q} for q>=p

  with package SV is new Vectors (Component => Scalar);
  with package SP is new Polynoms (Scalar => Scalar, SV => SV);

package LTaylors is      --- Power series of an operator L

  pragma Elaborate_Body;

  subtype Power is Integer range 0 .. PDeg;
  subtype Poly1 is SP.Polynom1(Power);
  subtype FltVec is Flt_Vectors.Vector;

  procedure Show_OpNorms;

  type LTaylor is
    record
      C: Poly1;           -- coefficients
      E: Radius := Zero;  -- higher order error
    end record;

  --------------- standard Scalar procedures

  --- basic
  function IsConst(P: LTaylor) return Boolean;
  function IsPoly(P: LTaylor) return Boolean;
  function IsZero(P: LTaylor) return Boolean;
  procedure SetZero(P: in out LTaylor);
  procedure Copy(P1: in LTaylor; P2: in out LTaylor);
  procedure Swap(P1,P2: in out LTaylor);

  --- addition and multiplication etc.
  procedure Neg(P: in out LTaylor);
  procedure Neg(P1: in LTaylor; P2: in out LTaylor);
  procedure Add(I: in Integer; P: in out LTaylor);
  procedure Add(P1: in LTaylor; P2: in out LTaylor);
  procedure Sum(P1,P2: in LTaylor; P3: in out LTaylor);
  procedure Sub(P1: in LTaylor; P2: in out LTaylor);
  procedure Diff(P1,P2: in LTaylor; P3: in out LTaylor);
  procedure Mult(R: in Flt; P: in out LTaylor);
  procedure Mult(S: in Scalar; P: in out LTaylor);
  procedure Prod(R: in Flt; P1: in LTaylor; P2: in out LTaylor);
  procedure Prod(S: in Scalar; P1: in LTaylor; P2: in out LTaylor);
  procedure AddProd(S: in Scalar; P1: in LTaylor; P2: in out LTaylor);
  procedure SubProd(S: in Scalar; P1: in LTaylor; P2: in out LTaylor);
  procedure Mult(P1: in LTaylor; P2,Tmp: in out LTaylor);
  procedure Mult(P1: in LTaylor; P2: in out LTaylor);
  procedure Prod(P1,P2: in LTaylor; P3: in out LTaylor);
  procedure AddProd(P1,P2: in LTaylor; P3: in out LTaylor);
  procedure SubProd(P1,P2: in LTaylor; P3: in out LTaylor);
  procedure LMult(Q: in Power; P: in out LTaylor);                  -- multiply by L^Q
  procedure LProd(Q: in Power; P1: in LTaylor; P2: in out LTaylor); -- multiply by L^Q

  --- functions
  procedure Inv(P1: in LTaylor; P2: in out LTaylor);
  procedure Exp(P1: in LTaylor; P2: in out LTaylor);
  procedure Log(P1: in LTaylor; P2: in out LTaylor);
  procedure QPower(Q: in Rational; P1: in LTaylor; P2: in out LTaylor);
  function MaxNorm(P: LTaylor) return Radius;

  --- conversion and i/o
  procedure Assign(R: in Flt; P: in out LTaylor);
  procedure Assign(S: in Scalar; P: in out LTaylor);
  procedure AddCoeff(I: in Natural; S: in Scalar; P: in out LTaylor);
  procedure Show1(N: in String; P: in LTaylor; Hide0: in Boolean := True);
  procedure Put(F: in File_Type; P: in LTaylor; Decimal: in Boolean := False);
  procedure Get(F: in File_Type; P: in out LTaylor; Decimal: in Boolean := False);
  procedure Write(FileName: in String; P: in LTaylor; Decimal: in Boolean := False);
  procedure Read(FileName: in String; P: in out LTaylor; Decimal: in Boolean := False);

private

  protected type Protected_Initialize is
    procedure Initialize;
  end Protected_Initialize;

  PDeg1: constant Positive := PDeg+1;
  WLast: constant Positive := 2*PDeg1;
  VarRho: Radius := Zero;         --- upper bound on ||L^q||^{1/q} for q > WLast
  OpNorm: FltVec(0 .. WLast);     --- upper bound on ||L^q|| for q <= Wast
  OpNorms: Protected_Initialize;

  subtype CAbs is FltVec(Power);

  procedure CCAddProd(C1,C2: in Poly1; A1,A2: CAbs; P3: in out LTaylor);
  procedure CEAddProd(A1: in CAbs; E2: in Radius; E3: in out Radius);

  generic
    with procedure Fac(I: in Positive; S: in out Scalar); --- assuming bounded in modulus by 1
  procedure Series(X: in LTaylor; Y: in out LTaylor; Decreasing: in Boolean);

  SOne:    constant Scalar      := Scal(One);
  SInfo:   constant Scalar_Info := Info(SOne);
  STrunc:  constant Boolean     := SInfo.IsNumeric;
  Not_STrunc: constant Boolean  := not STrunc;

end LTaylors;
