with Flts;
use Flts;

pragma Elaborate_all (Flts);

generic

  Dim: in Positive := 3;

  type Scalar is private;

  with procedure Proper_Rounding is <>;
  with procedure Center(S: in out Scalar) is <>;
  with procedure Copy(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Swap(S1,S2: in out Scalar) is <>;
  with procedure Add(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Sub(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Sum(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with procedure Diff(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with procedure Mult(R: in Flt; S: in out Scalar) is <>;

package N_M is --- find approximate minimum of F via Nelder-Mead

  subtype VRange is Integer range 1 .. Dim;
  type Vec is array (VRange) of Scalar;        --- point in R^Dim
  Empty: constant Flt := Flt'Last;

  type VecFlt is
    record
      X: Vec;
      Y: Flt := Empty;
    end record;

  Dim1: constant Positive := Dim+1;
  subtype VRange1 is Integer range 1 .. Dim1;
  type VecFltVec is array (VRange1) of VecFlt; --- polytope in R^Dim

  generic
    with function F(X: Vec) return Flt;
    with function Done(Q: VecFltVec) return Boolean; --- evaluated when YMin has decreased
  procedure FindMin(P: in out VecFltVec);            --- input points need not have values

  Alpha: Flt := One;
  Gamma: Flt := Two;
  Rho:   Flt := Half;
  Sigma: Flt := Half;

  Fill_Parallel: Boolean := False;
  Verbose: Boolean := False;

private

  procedure Copy(P1: in VecFlt; P2: in out VecFlt);
  procedure Swap(P1,P2: in out VecFlt);
  procedure Sort(P: in out VecFltVec);
  procedure Centroid(P: in VecFltVec; P0: in out VecFlt);
  procedure Shrink(P: in out VecFltVec);
  function NumEmpty(P: VecFltVec) return Natural;

  generic
    with function F(X: Vec) return Flt;
  procedure Fill(P: in out VecFltVec);

end N_M;
