with Flts, Globals, Reps, Params;
use Flts, Globals, Reps, Params;
with Params.Names,Matrices,Vectors,Cylinders, Flts.Minima;

pragma Elaborate_All (Flts,Globals,Reps,Params.Names,Matrices,Vectors,Cylinders,Flts.Minima);

procedure Num_Extrema(P: in Param) is

  package PN is new Params.Names(P => P, Trunc => True);
  use PN;

  subtype Scalar is Rep;

  package SM is new Matrices (Scalar => Scalar);
  package SV is new Vectors (Component => Scalar);
  package Cyls is new Cylinders(DegX => DegX, DegY => DegY, LDeg => LDeg, WFac => WFac, RadX => RadX,
                        EllR => EllR, Scalar => Scalar, SM => SM, SV => SV, Ref => Ref, Symm => Symm);
  use Cyls,Cyls.Cyl;

  XL:  constant Positive := 2**9;
  YL:  constant Positive := 2**8;
  Pi:  constant := 3.14159_26535_89793_23846_26433_83279_50288_41971_69399_37511;
  Pi2: constant Flt := Two*Pi;
  DX: constant Flt := Pi2/Flt(XL);
  DY: constant Flt := EllR/Flt(YL);
  Offset: constant Flt := Pi2*Flt(-4)/Flt(9); --- try to avoid peaks at 0 or XL

  Eps: constant Radius := 1.0E-15;            --- precision for Local_Min and Local_Max
  Cut: constant Radius := 1.0E-1;             --- no local extrema that close to zero

  function Xof(M: Natural) return Flt is
  begin
    if M=XL then return Offset+Pi2; else return Offset+Flt(M)*DX; end if;
  end Xof;

  function Yof(N: Natural) return Flt is
  begin
    if N=YL then return EllR; else return Flt(N)*DY; end if;
  end Yof;

  C: constant Cylinder := Read(Input_Dir & Fix_Name);

  function PlusVal(X,Y: Flt) return Flt is
  begin
    return ApproxVal(X,Y,C);
  end PlusVal;

  function MinusVal(X,Y: Flt) return Flt is
  begin
    return -ApproxVal(X,Y,C);
  end MinusVal;

  procedure Local_Min is
    package Flts_Min is new Flts.Minima (XL => XL, YL => YL, Cutoff => Cut, Xof => Xof, Yof => Yof, Fun => PlusVal);
    use Flts_Min;
    Dist: constant Positive := 3;
    V: constant Flt_Vec := Local_Minima(Dist,Eps);
  begin
    Show_List("local min",V);
    Write(Output_Dir & Prefix & "-min.approx",V);
  end Local_Min;

  procedure Local_Max is
    package Flts_Max is new Flts.Minima (XL => XL, YL => YL, Cutoff => Cut, Xof => Xof, Yof => Yof, Fun => MinusVal);
    use Flts_Max;
    Dist: constant Positive := 3;
    V: Flt_Vec := Local_Minima(Dist,Eps);
  begin
    NegVal(V);
    Show_List("local max",V);
    Write(Output_Dir & Prefix & "-max.approx",V);
  end Local_Max;

begin
  NTasks := 6;
  Verbosity := 2;
  Show("",P);

  Local_Max;
  Local_Min;

end Num_Extrema;
