with Flts,Ints;
use Flts,Ints;

pragma Elaborate_All (Flts);

package Params is

   subtype String4 is String(1..4);

   type Param is
      record
         Prefix: String4;       --- name
         Map:   Positive;
         DegX:  Positive;       --- max degree periodic
         DegY:  Positive;       --- max degree "radial"
         LDeg:  Natural;        --- LTaylor degree
         WFac:  Radius;         --- weight factor is (1+WFac*N)
         EllR:  Radius;         --- e^(-ell) <  R < 1
         RadX:  Radius;
         NN:    Integer;        --- parameter in psi
         RR:    Radius;         --- parameter in psi
         Ref:   MinusOneZeroOne;  --- odd, no reflection, even
         Symm:  Integer;        ---  symmetry with respect to 2pi/Symm rotations, 0 means radial
         Mat_PackSize:  Positive;
         Contr_Packsize: Natural;
      end record;

   ANN1: constant Param := (Prefix => "ANN1",  ---- 1
                            Map  => 1,
                            DegX => 20,
                            DegY => 50,
                            LDeg => 10,
                            WFac => Half,
                            EllR => One,
                            RadX => Flt(33)/Flt(32),
                            NN   => 0,
                            RR   => Zero,
                            Ref => 1,
                            Symm => 0,
                            Mat_PackSize => 12,
                            Contr_Packsize => 16);

   ANN3: constant Param := (Prefix => "ANN3",  ---- 2
                            Map  => 1,
                            DegX => 50,
                            DegY => 50,
                            LDeg => 0,
                            WFac => Half,
                            EllR => One,
                            RadX => Flt(33)/Flt(32),
                            NN   => 0,
                            RR   => Zero,
                            Ref => 1,
                            Symm => 1,
                            Mat_PackSize => 16,
                            Contr_Packsize => 20);

   ANN7: constant Param := (Prefix => "ANN7",  ----  3
                            Map  => 1,
                            DegX => 200,
                            DegY => 100,
                            LDeg => 0,
                            WFac => Half,
                            EllR => Quarter,
                            RadX => Flt(33)/Flt(32),
                            NN   => 0,
                            RR   => Zero,
                            Ref => 1,
                            Symm => 12,
                            Mat_PackSize => 26,
                            Contr_Packsize => 63);

   ANN8: constant Param := (Prefix => "ANN8",  ----  4
                            Map  => 1,
                            DegX => 120,
                            DegY => 100,
                            LDeg => 0,
                            WFac => Half,
                            EllR => Half,
                            RadX => Flt(33)/Flt(32),
                            NN   => 0,
                            RR   => Zero,
                            Ref => 1,
                            Symm => -3,
                            Mat_PackSize => 20,
                            Contr_Packsize => 32);

   E3B1: constant Param := (Prefix => "E3B1", ----  5
                            Map  => 2,
                            DegX => 34,
                            DegY => 50,
                            LDeg => 0,
                            WFac => Half,
                            EllR => One,
                            RadX => Flt(33)/Flt(32),
                            NN   => 3,
                            RR   => Flt(461)*Half**9,   ----- 0.900391...
                            Mat_PackSize => 16,
                            Ref => 1,
                            Symm => 1,
                            Contr_Packsize => 32);

   EX3F: constant Param := (Prefix => "EX3F", --- 6
                            Map  => 2,
                            DegX => 110,
                            DegY => 110,
                            LDeg => 0,
                            WFac => Half,
                            EllR => Half,
                            RadX => Flt(33)/Flt(32),
                            NN   => 3,
                            RR   => Three*Half**2,
                            Ref => 1,
                            Symm => 3,
                            Mat_PackSize => 20,
                            Contr_Packsize => 56);

   EX3B: constant Param := (Prefix => "EX3B",  ---- 7
                            Map  => 2,
                            DegX => 50,
                            DegY => 60,
                            LDeg => 0,
                            WFac => Half,
                            EllR => One,
                            RadX => Flt(33)/Flt(32),
                            NN   => 3,
                            RR   => Half**2,
                            Ref  => -1,
                            Symm => 1,
                            Mat_PackSize => 20,
                            Contr_Packsize => 24);

   EX3C: constant Param := (Prefix => "EX3C", --- 8
                            Map  => 2,
                            DegX => 100,
                            DegY => 80,
                            LDeg => 0,
                            WFac => Half,
                            EllR => Flt(25)*Half**5,
                            RadX => Flt(33)/Flt(32),
                            NN   => 3,
                            RR   => Three*Half**3,
                            Ref => 1,
                            Symm => 1,
                            Mat_PackSize => 18,
                            Contr_Packsize => 32);

   E3M3: constant Param := (Prefix => "E3M3",  --- 9
                            Map  => 2,
                            DegX => 20,
                            DegY => 50,
                            LDeg => 0,
                            WFac => Half,
                            EllR => One,
                            RadX => Flt(33)/Flt(32),
                            NN   => -3,
                            RR   => Flt(51)*Half**10,
                            Ref => 1,
                            Symm => 3,
                            Mat_PackSize => 16,
                            Contr_Packsize => 20);

   E36E: constant Param := (Prefix => "E36E", --- 10
                            Map  => 2,
                            DegX => 100,
                            DegY => 150,
                            LDeg => 0,
                            WFac => Half,
                            EllR => Half,
                            RadX => Flt(33)/Flt(32),
                            NN   => 6,
                            RR   => Half,
                            Ref  => 0,
                            Symm => 1,
                            Mat_PackSize => 24,
                            Contr_Packsize => 60);

   ISO3: constant Param := (Prefix => "ISO3",  --- 11
                            Map  => 3,
                            DegX => 30,
                            DegY => 50,
                            LDeg => 50,
                            WFac => Half,
                            EllR => One,
                            RadX => Flt(33)/Flt(32),
                            NN   => 0,
                            RR   => Half**2,  ---- r = -RR
                            Ref => 1,
                            Symm => 1,
                            Mat_PackSize => 16,
                            Contr_Packsize => 24);

   ISO2: constant Param := (Prefix => "ISO2",  --- 12
                            Map  => 3,
                            DegX => 50,
                            DegY => 50,
                            LDeg => 50,
                            WFac => Half,
                            EllR => One,
                            RadX => Flt(33)/Flt(32),
                            NN   => 0,
                            RR   => Half**2,  ---- r = -RR
                            Ref => 1,
                            Symm => 1,
                            Mat_PackSize => 16,
                            Contr_Packsize => 44);

   ISO1: constant Param := (Prefix => "ISO1", --- 13
                            Map  => 3,
                            DegX => 50,
                            DegY => 50,
                            LDeg => 50,
                            WFac => Half,
                            EllR => Two,
                            RadX => Flt(33)/Flt(32),
                            NN   => 0,
                            RR   => Three*Half**3,  ---- r = -RR
                            Ref => 1,
                            Symm => 1,
                            Mat_PackSize => 16,
                            Contr_Packsize => 36);

   IEL3: constant Param := (Prefix => "IEL3", ---  14
                            Map  => 5,
                            DegX => 48,
                            DegY => 100,
                            LDeg => 150,
                            WFac => Half,
                            EllR => One,
                            RadX => Flt(33)/Flt(32),
                            NN   => 0,
                            RR   => Flt(13)*Half**5,   ---- 0.40625
                            Ref => 1,
                            Symm => 1,
                            Mat_PackSize => 16,
                            Contr_Packsize => 52);

   IEL2: constant Param := (Prefix => "IEL2",  --- 15
                            Map  => 5,
                            DegX => 48,
                            DegY => 100,
                            LDeg => 200,
                            WFac => Half,
                            EllR => One,
                            RadX => Flt(33)/Flt(32),
                            NN   => 0,
                            RR   => Flt(21)*Half**6,  ----  0.328125
                            Ref => 1,
                            Symm => 2,
                            Mat_PackSize => 16,
                            Contr_Packsize => 52);

   CAS1: constant Param := (Prefix => "CAS1",  --- 16
                            Map  => 4,
                            DegX => 40,
                            DegY => 100,
                            LDeg => 100,
                            WFac => Half,
                            EllR => One,
                            RadX => Flt(33)/Flt(32),
                            NN   => 0,
                            RR   => Flt(11)*Half**5,  ----  0.34375
                            Ref => 1,
                            Symm => 2,
                            Mat_PackSize => 16,
                            Contr_Packsize => 32);

   NCS3: constant Param := (Prefix => "NCS3",  --- 17
                            Map  => 6,
                            DegX => 100,
                            DegY => 100,
                            LDeg => 180,
                            WFac => Half,
                            EllR => One,
                            RadX => Flt(33)/Flt(32),
                            NN   => 50,
                            RR   => Three*Quarter,
                            Ref => 1,
                            Symm => -1,
                            Mat_PackSize => 24,
                            Contr_Packsize => 66);

   ALLX: constant Param := (Prefix => "ALLX",  -----  dummy
                            Map  => 2,
                            DegX => 20,
                            DegY => 50,
                            LDeg => 0,
                            WFac => Half,
                            EllR => One,
                            RadX => Flt(33)/Flt(32),
                            NN   => -7,
                            RR   => Three*Half**12,
                            Ref => 1,
                            Symm => 1,
                            Mat_PackSize => 16,
                            Contr_Packsize => 20);

   type Param_List is array(Natural range <>) of Param;

   PList: constant Param_List(0..17) := (ALLX,ANN1,ANN3,ANN7,ANN8,E3B1,E3M3,E36E,
                                         EX3B,EX3C,EX3F,ISO1,ISO2,ISO3,CAS1,NCS3,IEL2,IEL3);

   procedure Show(N: in String; P: in Param);
   function IndexOf(S: String) return Natural;

end Params;
