with ScalInfo, Flts, Vectors.Sorting, Polynoms.Soil;
use ScalInfo, Flts;

pragma Elaborate_All (ScalInfo,Flts,Vectors.Sorting,Polynoms.Soil);

generic

  with function Info(Dummy: Scalar) return Scalar_Info is <>;
  with function Sign(S: Scalar) return Integer is <>;
  with procedure Center(S: in out Scalar) is <>;
  with procedure AddProd(S1,S2: in Scalar; S3,Tmp: in out Scalar) is <>;
  with procedure BallAt0(R: in Flt; S: in out Scalar) is <>;
  with procedure Inv(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Quot(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with procedure Norm(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Sqrt(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Roots2(B,C: in Scalar; U1,U2,V: in out Scalar) is <>;

  with procedure Assign(R: in Flt; S: in out Scalar) is <>;
  with function Scal(R: Flt) return Scalar is <>;
  with function MaxNorm(S: Scalar) return Radius is <>;
  with function Approx(S: Scalar) return Flt is <>;
  with procedure Mult(R: in Flt; S: in out Scalar) is <>;
  with procedure Mult(R: in Flt; S,Tmp: in out Scalar) is <>;
  with procedure Prod(R: in Flt; S1: in Scalar; S2: in out Scalar) is <>;

  with procedure Proper_Rounding is <>;
  with procedure Put_Numeric(N: in Positive; S: in Scalar) is <>;
  with procedure Get_Numeric(N: in Positive; S: in out Scalar) is <>;

package Polynoms.Roots is

  package SP_Soil is new Polynoms.Soil;
  package SV_Sorting is new SV.Sorting;

  procedure Norm(X,Y: in Scalar; S,Tmp: in out Scalar);                -- S := Sqrt(X*X+Y*Y)
  procedure Sqrt(Zr,Zi: in Scalar; Wr,Wi: in out Scalar);              -- W := Sqrt(Z)  complex
  procedure QSolve(Zr,Zi: in Scalar; Wr1,Wi1,Wr2,Wi2: in out Scalar);  -- solve W*W-2*Z*W+1=0  complex

  --- a Symplectic_Root is H=(W+1/W)/2 for a root W
  procedure Real_Root(P: in Polynom1; WMin,WMax: in Flt; W: in out Scalar; Eps: in Radius);
  procedure Real_Roots(P: in Polynom1; WMin,WMax: in Flt; W: in out Vector; Eps: in Radius);
  procedure Circle_Roots(P: in Polynom1; Hr: in out Vector; Eps: in Radius);
  procedure Sym_QFactor(P: in Polynom1; UMin,UMax: in Flt; U: in out Scalar; Q: in out Polynom1; Eps: in Radius);
  procedure Sym_QFactor(P: in Polynom1; UMin,UMax: in Flt; U: in out Scalar; Eps: in Radius);
  procedure Symplectic_Roots(P: in Polynom1; Hr,Hi: in out Vector; Eps: in Radius);
  procedure Gen_Roots(P: in Polynom1; Wr,Wi: in out Vector; D: out Natural); -- general, D is the number of roots found

  --- for testing etc
  procedure SymProd(P1: in Polynom1; P2: in out Polynom1);                   -- P2(x) := (x^D)P1(1/x)*P1(x)
  function SymErr(P: Polynom1) return Radius;                                -- measure of asymmetry
  procedure RandomRoots(W: in out Vector; Fac: in Flt := One);
  procedure RandomRoots(NSimple: in Natural; Wr,Wi: in out Vector; RFac,IFac: in Flt := One);
  procedure RandomReal(W: in out Vector; P: in out Polynom1; Fac: in Flt := One);           -- real roots
  procedure RandomCC(Wr,Wi: in out Vector; P: in out Polynom1; RFac,IFac: in Flt := One);   -- cc roots
  procedure Random(NSimple: in Natural; Wr,Wi: in out Vector; P: in out Polynom1; RFac,IFac: in Flt := One); -- NSimple real roots
  procedure RandomPoly(U,V: in out Vector; P: in out Polynom1; UFac,VFac: in Flt := Three); -- using QFactors
  procedure RandomSymPoly(U: in out Vector; P: in out Polynom1; Fac: in Flt := Two);        -- using QFactors, V(I)=1

  NFind: Positive renames SP_Soil.Bis.NFind;

end Polynoms.Roots;
