with ScalInfo, Flts, Bisect, Bisect.PSign, Small_Vectors;
use ScalInfo, Flts;

pragma Elaborate_All (ScalInfo,Flts,Bisect.PSign,Small_Vectors);

generic

  with function Info(Dummy: Scalar) return Scalar_Info is <>;
  with function Sign(S: Scalar) return Integer is <>;
  with procedure Center(S: in out Scalar) is <>;
  with procedure Assign(R: in Flt; S: in out Scalar) is <>;
  with function Scal(R: Flt) return Scalar is <>;
  with function MaxNorm(S: Scalar) return Radius is <>;
  with procedure BallAt0(R: in Flt; S: in out Scalar) is <>;
  with function Approx(S: Scalar) return Flt is <>;
  with procedure Mult(R: in Flt; S: in out Scalar) is <>;
  with procedure Inv(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Quot(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with procedure Roots2(B,C: in Scalar; U1,U2,V: in out Scalar) is <>;

package Polynoms.Soil is --- needed by Polynoms.Roots

  subtype Poly is Polynom1;

  procedure FindRealRoot(P: in Poly; W: in out Scalar; Err: out Radius); -- Newton1
  procedure FindRealRoots(P: in Poly; W: in out Vector);
  procedure FindQFactor(P: in Poly; U,V: in out Scalar; Q: in out Poly; Err: out Radius);  -- needs approx U,V
  procedure FindQFactors(P: in Poly; U,V: in out Vector; A,B: in out Scalar);              -- needs approx U,V
  procedure FindRoots(P: in Poly; Wr,Wi: in out Vector; D: in Natural);                    -- needs approx Wr,Wi

  --- symplectic polynomials: roots come in pairs (r,1/r)
  procedure CircleRoots(P: in Poly; Hr: in out Vector);         -- needs approximate Hr (real parts of roots)
  procedure FindSymQFactor(P: in Poly; U0: in out Scalar; Q: in out Poly; Err: out Radius); -- needs approx U
  procedure FindSymQFactors(P: in Poly; U: in out Vector);                                  -- needs approx U

  ----------------------------
  ----- purely numerical ----- (eps is for bisection only)
  ----------------------------

  --- 1d bisection and 1d Newton
  procedure RoughRealRoot(P: in Poly; WMin,WMax: in Flt; W: in out Scalar; Err: in out Radius); -- bisection
  procedure NumRealRoot(P: in Poly; WMin,WMax: in Flt; W: in out Scalar; Eps: in Radius);       -- bisection, then Newton
  procedure NumRealRoots(P: in Poly; WMin,WMax: in Flt; W: in out Vector; Eps: in Radius);      -- bisection, then Newton

  --- symplectic roots on the circle
  procedure NumCircleRoots(P: in Poly; Hr: in out Vector; Eps: in Radius);           -- first SymCosPoly, then NumRealRoots

  --- for symplectic roots on the circle or real line (a SymQFactor is a factor x*x+U*x+1)
  procedure RoughSymQFactor(P: in Poly; UMin,UMax: in Flt; U0: in out Scalar; Tmp: in out Poly; Err: in out Radius); -- bisection
  procedure NumSymQFactor(P: in Poly; UMin,UMax: in Flt; U: in out Scalar; Eps: in Radius);   -- bisection, then Newton
  procedure NumSymQFactors(P: in Poly; UMin,UMax: in Flt; U: in out Vector; Eps: in Radius);  -- bisection, then Newton

  package SMV is new Small_Vectors (Scalar => Scalar, MS => MuSc);
  package Bis is new Bisect(Scalar => Scalar);
  package Bis_PSign is new Bis.PSign;

  Scal_Zero: constant Scalar := Scal(Zero);
  Scal_Half: constant Scalar := Scal(Half);
  Scal_One:  constant Scalar := Scal(One);
  Scal_Two:  constant Scalar := Scal(Two);
  STrunc:    constant Boolean;

  No_Contraction, Off_Circle: exception;

private

  procedure RealRoot1(P0,P1: in Scalar; W: in out Scalar);
  procedure RealRoots2(P0,P1,P2: in Scalar; W1,W2: in out Scalar);
  function Right_Sign(P: Poly) return Integer;
  function Left_Sign(P: Poly) return Integer;
  procedure FindQFactor2(P0,P1,P2: in Scalar; U,V: in out Scalar; Q: in out Poly);
  procedure FindSymQFactor2(P0,P1,P2: in Scalar; U: in out Scalar; Q: in out Poly);

  SInfo:     constant Scalar_Info := Info(Scal_Zero);
  STrunc:    constant Boolean     := SInfo.IsNumeric;

end Polynoms.Soil;
