with Ada.Text_IO, Integer_Types;
use Ada.Text_IO;

pragma Elaborate_All (Ada.Text_IO,Integer_Types);

package Rationals is  --- not for major computations

  subtype LInt is Integer_Types.LInt; -- Long_Long_Integer
  subtype LNat is Integer_Types.LNat; -- LInt range 0 .. LInt'Last
  subtype LPos is Integer_Types.LPos; -- LInt range 1 .. LInt'Last

  package LInt_IO is new Ada.Text_IO.Integer_IO (LInt);

  function Sqr(I: Lint) return LInt;
  function GCD(N,D: LInt) return LInt;
  function Width(I: LInt; Base: Positive := 10) return Positive;
  function Strng(I: LInt; W: Positive) return String;
  function Strng(I: LInt) return String;
  function Value(S: String) return LInt;
  procedure Show1(N: in String; I: in LInt; NewLine: in Boolean := True);
  function Prompt(N: String) return LInt;
  procedure Put(F: in File_Type; I: in LInt; Dummy: in Boolean := False);
  procedure Get(F: in File_Type; I: in out LInt; Dummy: in Boolean := False);

  type Rational is private;

  function "-"(Q: Rational) return Rational;
  function "+"(P,Q: Rational) return Rational;
  function "-"(P,Q: Rational) return Rational;
  function "*"(P,Q: Rational) return Rational;
  function "/" (P,Q: Rational) return Rational;
  function "mod"(P,Q: Rational) return Rational;
  function Inv(Q: Rational) return Rational;
  function Sqr(Q: Rational) return Rational;
  function IsZero(Q: Rational) return Boolean;
  function Sign(Q: Rational) return Integer;
  function "abs"(Q: Rational) return Rational;
  function "<"(P,Q: Rational) return Boolean;

  function LRat(N: Lint; D: LInt := 1) return Rational;
  function LNum(Q: Rational) return Lint;
  function LDen(Q: Rational) return Lint;
  function LFloor(Q: Rational) return LInt;
  procedure LFloorPlus(Q: in out Rational; C: out LInt);
  function LCeiling(Q: Rational) return LInt;
  function "mod"(P: Rational; I: LInt) return Rational;
  ---
  function LSum(I: LInt; Q: Rational) return Rational;
  function LSum(Q: Rational; I: LInt) return Rational;
  function LDiff(I: LInt; Q: Rational) return Rational;
  function LDiff(Q: Rational; I: LInt) return Rational;
  function LProd(I: LInt; Q: Rational) return Rational;
  function LProd(Q: Rational; I: LInt) return Rational;
  function LQuot(I: LInt; Q: Rational) return Rational;
  function LQuot(Q: Rational; I: LInt) return Rational;

  function IRat(N: Integer; D: Integer := 1) return Rational;
  function INum(Q: Rational) return Integer;
  function IDen(Q: Rational) return Integer;
  function IFloor(Q: Rational) return Integer;
  function ICeiling(Q: Rational) return Integer;
  function "mod"(P: Rational; I: Integer) return Rational;
  ---
  function ISum(I: Integer; Q: Rational) return Rational;
  function ISum(Q: Rational; I: Integer) return Rational;
  function IDiff(I: Integer; Q: Rational) return Rational;
  function IDiff(Q: Rational; I: Integer) return Rational;
  function IProd(I: Integer; Q: Rational) return Rational;
  function IProd(Q: Rational; I: Integer) return Rational;
  function IPower(Q: Rational; I: Integer) return Rational;
  function IQuot(I: Integer; Q: Rational) return Rational;
  function IQuot(Q: Rational; I: Integer) return Rational;

  function "+"(I: Integer; Q: Rational) return Rational renames ISum;
  function "+"(Q: Rational; I: Integer) return Rational renames ISum;
  function "-"(I: Integer; Q: Rational) return Rational renames IDiff;
  function "-"(Q: Rational; I: Integer) return Rational renames IDiff;
  function "*"(I: Integer; Q: Rational) return Rational renames IProd;
  function "*"(Q: Rational; I: Integer) return Rational renames IProd;
  function "**"(Q: Rational; I: Integer) return Rational renames IPower;
  function "/"(N,D: Integer) return Rational renames IRat;
  function "/"(I: Integer; Q: Rational) return Rational renames IQuot;
  function "/"(Q: Rational; I: Integer) return Rational renames IQuot;

  function Strng(Q: Rational) return String;
  function Value(S: String) return Rational;
  procedure Normalize(Q: in out Rational); --- should not be necessary
  procedure Show1(N: in String; Q: in Rational; NewLine: in Boolean := True);
  function Prompt(N: String) return Rational;
  function GetArg return Rational;
  procedure Put(F: in File_Type; Q: in Rational; Dummy: in Boolean := False);
  procedure Get(F: in File_Type; Q: in out Rational; Dummy: in Boolean := False);

  use Integer_Types;
  subtype IMat2 is Int_Mat(OneTwo,OneTwo);

  function EvenOverOdd(Q: Rational) return Boolean;
  function OddOverEven(Q: Rational) return Boolean;
  function OddOverOdd(Q: Rational) return Boolean;
  function ContFrac_Length(Q: Rational) return Natural;
  function ContFrac(Q: Rational) return Int_Vec; --- starting with index 0
  procedure PnQn(C: in Int_Vec; Pn0,Pn1,Qn0,Qn1: out LInt);
  function PnQn(C: Int_Vec) return IMat2;   -- [Pn0,Pn1;Qn0,Qn1]
  function PnQn(Q: Rational) return IMat2;  -- [Pn0,Pn1;Qn0,Qn1]

  Rat_Zero,Rat_One,Rat_NegOne,Rat_Two,Rat_Half: constant Rational;

private

  type Rational is
    record
      Num: LInt;
      Den: LNat;
    end record;

  function "/"(N,D: LInt) return Rational renames LRat;
  function Normalize(Q: Rational) return Rational; --- should not be needed

  Rat_Zero:   constant Rational := ( 0,1);
  Rat_One:    constant Rational := ( 1,1);
  Rat_NegOne: constant Rational := (-1,1);
  Rat_Two:    constant Rational := ( 2,1);
  Rat_Half:   constant Rational := ( 1,2);

end Rationals;
