with Polynoms, Polynoms.Roots;

pragma Elaborate_All (Polynoms.Roots);

generic

  with function Sign(S: Scalar) return Integer is <>;
  with procedure Mult(R: in Flt; S,Tmp: in out Scalar) is <>;
  with procedure Roots2(B,C: in Scalar; U1,U2,V: in out Scalar) is <>;

  with procedure Put_Numeric(N: in Positive; S: in Scalar) is <>;
  with procedure Get_Numeric(N: in Positive; S: in out Scalar) is <>;

package ScalVectors.Eigenv is

  package SP is new Polynoms (Scalar => Scalar, SV => SV);
  package SP_Roots is new SP.Roots;

  subtype Polynom1 is SP.Polynom1;
  subtype EigenPolys is Matrix; -- range (0 .. N,0 .. N)

  procedure CharPolyTri(A: in Matrix; P: in out Polynom1);     -- characteristic polynomial of a tridiagonal matrix
  procedure EigenPolyHess(A: in Matrix; E: in out EigenPolys); -- E(0..N,*) are polynomials of degree <= N, for A(1..N,1..N)
  procedure GetCharPoly(E: in EigenPolys; P: in out Polynom1); -- last row of E
  procedure QSolve(Hr,Hi: in Vector; Wr,Wi: in out Vector);    -- solve (W+1/W)/2=H

  --- these need approx eigenvalues
  procedure EigenSymm(A: in Matrix; W: in out Vector);                    -- needs approx W
  procedure EigenSymm(A: in Matrix; W: in out Vector; V: in out Matrix);  -- needs approx W
  procedure EigenCircle(A: in Matrix; Hr: in out Vector);                 -- needs approx Hr (real part)

  --- these use SP_Roots to find approx solutions (eps is for bisection)
  --- a Sym_QFactor is a factor x*x+U*x+1
  --- a Symplectic_Root is H=(W+1/W)/2 for a root W
  procedure EigenSymm(A: in Matrix; WMin,WMax: in Flt; W: in out Vector; Eps: in Radius);
  procedure EigenSymm(A: in Matrix; W: in out Vector; Eps: in Radius);                    -- determines WMin,WMax via MaxNorm2
  procedure EigenSymm(A: in Matrix; WMin,WMax: in Flt; W: in out Vector; V: in out Matrix; Eps: in Radius);
  procedure EigenSymm(A: in Matrix; W: in out Vector; V: in out Matrix; Eps: in Radius);  -- determines WMin,WMax via MaxNorm2
  procedure EigenCircle(A: in Matrix; Hr: in out Vector; Eps: in Radius);
  procedure EigenSymplectic(A: in Matrix; Wr,Wi: in out Vector; Vr,Vi: in out Matrix; Eps: in Radius);

  ----------------------------
  ----- purely numerical -----
  ----------------------------

  procedure SortEigen(W: in out Vector) renames Default_Sort;
  procedure SortEigen(W: in out Vector; V: in out Matrix);
  procedure SortEigen(Eps: in Flt; Wr,Wi: in out Vector) renames Default_Sort;
  procedure SortEigen(Eps: in Flt; Wr,Wi: in out Vector; Vr,Vi: in out Matrix);

  procedure EigenQuality(A: in Matrix; W: in Scalar; V: in Vector);           -- norm1 of A*V-W*V
  procedure EigenQuality(A: in Matrix; Wr,Wi: in Scalar; Vr,Vi: in Vector);
  procedure EigenQuality(Ar,Ai: in Matrix; Wr,Wi: in Scalar; Vr,Vi: in Vector);
  procedure EigenQuality(A: in Matrix; W: in Vector; V: in Matrix);
  procedure EigenQuality(A: in Matrix; Wr,Wi: in Vector; Vr,Vi: in Matrix);
  procedure EigenQuality(Ar,Ai: in Matrix; Wr,Wi: in Vector; Vr,Vi: in Matrix);

private

   procedure CheckPolyDim(D: in Natural; P: in Polynom1);

end ScalVectors.Eigenv;
