with Globals, Newton;

pragma Elaborate_All (Globals,Newton);

package body Small_Vectors is

  -------------------- Vector2

  procedure Center2(V: in out Vector2) is
  begin
    Center(V(1));
    Center(V(2));
  end Center2;

  procedure Neg2(V: in out Vector2) is
  begin
    Neg(V(1));
    Neg(V(2));
  end Neg2;

  procedure Copy2(U: in Vector2; V: in out Vector2) is
  begin
    Copy(U(1),V(1));
    Copy(U(2),V(2));
  end Copy2;

  procedure Add2(U: in Vector2; V: in out Vector2) is
  begin
    Add(U(1),V(1));
    Add(U(2),V(2));
  end Add2;

  procedure Sum2(U,V: in Vector2; W: in out Vector2) is
  begin
    Sum(U(1),V(1),W(1));
    Sum(U(2),V(2),W(2));
  end Sum2;

  procedure Sub2(U: in Vector2; V: in out Vector2) is
  begin
    Sub(U(1),V(1));
    Sub(U(2),V(2));
  end Sub2;

  procedure Diff2(U,V: in Vector2; W: in out Vector2) is
  begin
    Diff(U(1),V(1),W(1));
    Diff(U(2),V(2),W(2));
  end Diff2;

  procedure Mult2(S: in Scalar; U: in out Vector2) is
  begin
    Mult(S,U(1));
    Mult(S,U(2));
  end Mult2;

  procedure Prod2(S: in Scalar; U: in Vector2; V: in out Vector2) is
  begin
    Prod(S,U(1),V(1));
    Prod(S,U(2),V(2));
  end Prod2;

  procedure AddProd2(S: in Scalar; U: in Vector2; V: in out Vector2) is
  begin
    AddProd(S,U(1),V(1));
    AddProd(S,U(2),V(2));
  end AddProd2;

  procedure Dot2(U,V: in Vector2; S: in out Scalar) is
  begin
    Prod(U(1),V(1),S);
    AddProd(U(2),V(2),S);
  end Dot2;

  function Dot2(U,V: Vector2) return Scalar is
    S: Scalar;
  begin
    Prod(U(1),V(1),S);
    AddProd(U(2),V(2),S);
    return S;
  end Dot2;

  procedure Vol2(U1,U2: in Vector2; S: in out Scalar) is
    --- det[U1 U2]
    Tmp: Scalar;
  begin
    Prod(U1(1),U2(2),S);
    Prod(U1(2),U2(1),Tmp);
    Sub(Tmp,S);
  end Vol2;

  function Vol2(U1,U2: Vector2) return Scalar is
    S,Tmp: Scalar;
  begin
    Prod(U1(1),U2(2),S);
    Prod(U1(2),U2(1),Tmp);
    Sub(Tmp,S);
    return S;
  end Vol2;

  function MaxNorm(V: Vector2) return Radius is
  begin
    return MaxNorm(V(1))+MaxNorm(V(2));
  end MaxNorm;

  -------------------- Matrix2

  procedure Center2(A: in out Matrix2) is
  begin
    Center(A(1,1));
    Center(A(1,2));
    Center(A(2,1));
    Center(A(2,2));
  end Center2;

  procedure Neg2(A: in out Matrix2) is
  begin
    Neg(A(1,1));
    Neg(A(1,2));
    Neg(A(2,1));
    Neg(A(2,2));
  end Neg2;

  procedure Copy2(A: in Matrix2; B: in out Matrix2) is
  begin
    Copy(A(1,1),B(1,1));
    Copy(A(1,2),B(1,2));
    Copy(A(2,1),B(2,1));
    Copy(A(2,2),B(2,2));
  end Copy2;

  procedure Transpose2(A: in out Matrix2) is
  begin
    Swap(A(1,2),A(2,1));
  end Transpose2;

  procedure Transpose2(A: in Matrix2; B: in out Matrix2) is
  begin
    Copy(A(1,1),B(1,1));
    Copy(A(1,2),B(2,1));
    Copy(A(2,1),B(1,2));
    Copy(A(2,2),B(2,2));
  end Transpose2;

  procedure SetRows2(U1,U2: in Vector2; A: in out Matrix2) is
  begin
    Copy(U1(1),A(1,1));
    Copy(U1(2),A(1,2));
    Copy(U2(1),A(2,1));
    Copy(U2(2),A(2,2));
  end SetRows2;

  procedure GetRows2(A: in Matrix2; U1,U2: in out Vector2) is
  begin
    Copy(A(1,1),U1(1));
    Copy(A(1,2),U1(2));
    Copy(A(2,1),U2(1));
    Copy(A(2,2),U2(2));
  end GetRows2;

  procedure SetColumns2(U1,U2: in Vector2; A: in out Matrix2) is
  begin
    Copy(U1(1),A(1,1));
    Copy(U1(2),A(2,1));
    Copy(U2(1),A(1,2));
    Copy(U2(2),A(2,2));
  end SetColumns2;

  procedure GetColumns2(A: in Matrix2; U1,U2: in out Vector2) is
  begin
    Copy(A(1,1),U1(1));
    Copy(A(2,1),U1(2));
    Copy(A(1,2),U2(1));
    Copy(A(2,2),U2(2));
  end GetColumns2;

  procedure Add2(A: in Matrix2; B: in out Matrix2) is
  begin
    Add(A(1,1),B(1,1));
    Add(A(1,2),B(1,2));
    Add(A(2,1),B(2,1));
    Add(A(2,2),B(2,2));
  end Add2;

  procedure Sum2(A,B: in Matrix2; C: in out Matrix2) is
  begin
    Sum(A(1,1),B(1,1),C(1,1));
    Sum(A(1,2),B(1,2),C(1,2));
    Sum(A(2,1),B(2,1),C(2,1));
    Sum(A(2,2),B(2,2),C(2,2));
  end Sum2;

  procedure Sub2(A: in Matrix2; B: in out Matrix2) is
  begin
    Sub(A(1,1),B(1,1));
    Sub(A(1,2),B(1,2));
    Sub(A(2,1),B(2,1));
    Sub(A(2,2),B(2,2));
  end Sub2;

  procedure Diff2(A,B: in Matrix2; C: in out Matrix2) is
  begin
    Diff(A(1,1),B(1,1),C(1,1));
    Diff(A(1,2),B(1,2),C(1,2));
    Diff(A(2,1),B(2,1),C(2,1));
    Diff(A(2,2),B(2,2),C(2,2));
  end Diff2;

  procedure Mult2(S: in Scalar; A: in out Matrix2) is
  begin
    Mult(S,A(1,1));
    Mult(S,A(1,2));
    Mult(S,A(2,1));
    Mult(S,A(2,2));
  end Mult2;

  function MaxNorm(A: Matrix2) return Radius is
  begin
    return RMax(MaxNorm(A(1,1))+MaxNorm(A(2,1)),
                MaxNorm(A(1,2))+MaxNorm(A(2,2)));
  end MaxNorm;

  procedure Prod2(A: in Matrix2; U: in Vector2; V: in out Vector2) is
  begin
    Prod(A(1,1),U(1),V(1));
    AddProd(A(1,2),U(2),V(1));
    Prod(A(2,1),U(1),V(2));
    AddProd(A(2,2),U(2),V(2));
  end Prod2;

  procedure Prod2(A,B: in Matrix2; C: in out Matrix2) is
  begin
    -- C(1,1)
    Prod(A(1,1),B(1,1),C(1,1));
    AddProd(A(1,2),B(2,1),C(1,1));
    -- C(1,2)
    Prod(A(1,1),B(1,2),C(1,2));
    AddProd(A(1,2),B(2,2),C(1,2));
    -- C(2,1)
    Prod(A(2,1),B(1,1),C(2,1));
    AddProd(A(2,2),B(2,1),C(2,1));
    -- C(2,2)
    Prod(A(2,1),B(1,2),C(2,2));
    AddProd(A(2,2),B(2,2),C(2,2));
  end Prod2;

  procedure Tr2(A: in Matrix2; S: in out Scalar) is
  begin
    Sum(A(1,1),A(2,2),S);
  end Tr2;

  function Tr2(A: Matrix2) return Scalar is
    S: Scalar;
  begin
    Sum(A(1,1),A(2,2),S);
    return S;
  end Tr2;

  procedure Det2(A: in Matrix2; S: in out Scalar) is
    Tmp: Scalar;
  begin
    Prod(A(1,1),A(2,2),S);
    Prod(A(1,2),A(2,1),Tmp);
    Sub(Tmp,S);
  end Det2;

  function Det2(A: Matrix2) return Scalar is
    S,Tmp: Scalar;
  begin
    Prod(A(1,1),A(2,2),S);
    Prod(A(1,2),A(2,1),Tmp);
    Sub(Tmp,S);
    return S;
  end Det2;

  procedure Adj2(A: in Matrix2; B: in out Matrix2) is
  begin
    Copy(A(2,2),B(1,1));
    Copy(A(1,1),B(2,2));
    Neg(A(1,2),B(1,2));
    Neg(A(2,1),B(2,1));
  end Adj2;

  procedure Inv2(A: in Matrix2; B: in out Matrix2) is
    Tmp: Scalar renames B(1,1);
    S: Scalar;
  begin
    Det2(A,Tmp);
    Inv(Tmp,S);
    Prod(S,A(2,2),B(1,1));
    Prod(S,A(1,1),B(2,2));
    Neg(S);
    Prod(S,A(1,2),B(1,2));
    Prod(S,A(2,1),B(2,1));
  end Inv2;

  procedure Quot2(B: in Vector2; A: in Matrix2; X: in out Vector2) is
    Tmp: Scalar renames X(1);
    S: Scalar;
  begin
    Det2(A,Tmp);
    Inv(Tmp,S);
    Prod(A(1,2),B(2),X(1));
    Neg(X(1));
    AddProd(A(2,2),B(1),X(1));
    Mult(S,X(1));
    Prod(A(2,1),B(1),X(2));
    Neg(X(2));
    AddProd(A(1,1),B(2),X(2));
    Mult(S,X(2));
  end Quot2;

  procedure Quot2(B,A: in Matrix2; X: in out Matrix2) is
    Tmp: Scalar renames X(1,1);
    S: Scalar;
  begin
    Det2(A,Tmp);
    Inv(Tmp,S);
    -- X(I,1)
    Prod(A(1,2),B(2,1),X(1,1));
    Neg(X(1,1));
    AddProd(A(2,2),B(1,1),X(1,1));
    Mult(S,X(1,1));
    Prod(A(2,1),B(1,1),X(2,1));
    Neg(X(2,1));
    AddProd(A(1,1),B(2,1),X(2,1));
    Mult(S,X(2,1));
    -- X(I,2)
    Prod(A(1,2),B(2,2),X(1,2));
    Neg(X(1,2));
    AddProd(A(2,2),B(1,2),X(1,2));
    Mult(S,X(1,2));
    Prod(A(2,1),B(1,2),X(2,2));
    Neg(X(2,2));
    AddProd(A(1,1),B(2,2),X(2,2));
    Mult(S,X(2,2));
  end Quot2;

  -------------------- Vector3

  procedure Center3(V: in out Vector3) is
  begin
    Center(V(1));
    Center(V(2));
    Center(V(3));
  end Center3;

  procedure Neg3(V: in out Vector3) is
  begin
    Neg(V(1));
    Neg(V(2));
    Neg(V(3));
  end Neg3;

  procedure Copy3(U: in Vector3; V: in out Vector3) is
  begin
    Copy(U(1),V(1));
    Copy(U(2),V(2));
    Copy(U(3),V(3));
  end Copy3;

  procedure Add3(U: in Vector3; V: in out Vector3) is
  begin
    Add(U(1),V(1));
    Add(U(2),V(2));
    Add(U(3),V(3));
  end Add3;

  procedure Sum3(U,V: in Vector3; W: in out Vector3) is
  begin
    Sum(U(1),V(1),W(1));
    Sum(U(2),V(2),W(2));
    Sum(U(3),V(3),W(3));
  end Sum3;

  procedure Sub3(U: in Vector3; V: in out Vector3) is
  begin
    Sub(U(1),V(1));
    Sub(U(2),V(2));
    Sub(U(3),V(3));
  end Sub3;

  procedure Diff3(U,V: in Vector3; W: in out Vector3) is
  begin
    Diff(U(1),V(1),W(1));
    Diff(U(2),V(2),W(2));
    Diff(U(3),V(3),W(3));
  end Diff3;

  procedure Mult3(S: in Scalar; U: in out Vector3) is
  begin
    Mult(S,U(1));
    Mult(S,U(2));
    Mult(S,U(3));
  end Mult3;

  procedure Prod3(S: in Scalar; U: in Vector3; V: in out Vector3) is
  begin
    Prod(S,U(1),V(1));
    Prod(S,U(2),V(2));
    Prod(S,U(3),V(3));
  end Prod3;

  procedure AddProd3(S: in Scalar; U: in Vector3; V: in out Vector3) is
  begin
    AddProd(S,U(1),V(1));
    AddProd(S,U(2),V(2));
    AddProd(S,U(3),V(3));
  end AddProd3;

  procedure Dot3(U,V: in Vector3; S: in out Scalar) is
  begin
    Prod(U(1),V(1),S);
    AddProd(U(2),V(2),S);
    AddProd(U(3),V(3),S);
  end Dot3;

  function Dot3(U,V: Vector3) return Scalar is
    S: Scalar;
  begin
    Prod(U(1),V(1),S);
    AddProd(U(2),V(2),S);
    AddProd(U(3),V(3),S);
    return S;
  end Dot3;

  procedure DiagProd3(B,U: in Vector3; BU: in out Vector3) is
  begin
    Prod(B(1),U(1),BU(1));
    Prod(B(2),U(2),BU(2));
    Prod(B(3),U(3),BU(3));
  end DiagProd3;

  procedure Vol3(U1,U2,U3: in Vector3; S: in out Scalar) is
    --- det[U1 U2 U3]
    Tmp: Scalar;
  begin
    Prod(U1(1),U2(2),Tmp);
    Prod(Tmp,U3(3),S);
    Prod(U2(1),U3(2),Tmp);
    AddProd(Tmp,U1(3),S);
    Prod(U3(1),U1(2),Tmp);
    AddProd(Tmp,U2(3),S);
    Prod(U3(1),U2(2),Tmp);
    Neg(Tmp);
    AddProd(Tmp,U1(3),S);
    Prod(U1(1),U3(2),Tmp);
    Neg(Tmp);
    AddProd(Tmp,U2(3),S);
    Prod(U2(1),U1(2),Tmp);
    Neg(Tmp);
    AddProd(Tmp,U3(3),S);
  end Vol3;

  function Vol3(U1,U2,U3: Vector3) return Scalar is
    S: Scalar;
  begin
    Vol3(U1,U2,U3,S);
    return S;
  end Vol3;

  procedure Cross3(U,V: in Vector3; W: in out Vector3) is
  begin
    Prod(U(3),V(2),W(1));
    Neg(W(1));
    AddProd(U(2),V(3),W(1));
    Prod(U(1),V(3),W(2));
    Neg(W(2));
    AddProd(U(3),V(1),W(2));
    Prod(U(2),V(1),W(3));
    Neg(W(3));
    AddProd(U(1),V(2),W(3));
  end Cross3;

  procedure Cross3(U,V,W: in Vector3; X: in out Vector3) is
    S: Scalar;
  begin
    Dot3(U,W,S);
    Prod3(S,V,X);
    Dot3(U,V,S);
    Neg(S);
    AddProd3(S,W,X);
  end Cross3;

  function MaxNorm(V: Vector3) return Radius is
  begin
    return MaxNorm(V(1))+MaxNorm(V(2))+MaxNorm(V(3));
  end MaxNorm;

  -------------------- Matrix3

  procedure Center3(A: in out Matrix3) is
  begin
    Center(A(1,1));
    Center(A(1,2));
    Center(A(1,3));
    Center(A(2,1));
    Center(A(2,2));
    Center(A(2,3));
    Center(A(3,1));
    Center(A(3,2));
    Center(A(3,3));
  end Center3;

  procedure Neg3(A: in out Matrix3) is
  begin
    Neg(A(1,1));
    Neg(A(1,2));
    Neg(A(1,3));
    Neg(A(2,1));
    Neg(A(2,2));
    Neg(A(2,3));
    Neg(A(3,1));
    Neg(A(3,2));
    Neg(A(3,3));
  end Neg3;

  procedure Copy3(A: in Matrix3; B: in out Matrix3) is
  begin
    Copy(A(1,1),B(1,1));
    Copy(A(1,2),B(1,2));
    Copy(A(1,3),B(1,3));
    Copy(A(2,1),B(2,1));
    Copy(A(2,2),B(2,2));
    Copy(A(2,3),B(2,3));
    Copy(A(3,1),B(3,1));
    Copy(A(3,2),B(3,2));
    Copy(A(3,3),B(3,3));
  end Copy3;

  procedure Transpose3(A: in out Matrix3) is
  begin
    Swap(A(1,2),A(2,1));
    Swap(A(1,3),A(3,1));
    Swap(A(2,3),A(3,2));
  end Transpose3;

  procedure Transpose3(A: in Matrix3; B: in out Matrix3) is
  begin
    Copy(A(1,1),B(1,1));
    Copy(A(1,2),B(2,1));
    Copy(A(1,3),B(3,1));
    Copy(A(2,1),B(1,2));
    Copy(A(2,2),B(2,2));
    Copy(A(2,3),B(3,2));
    Copy(A(3,1),B(1,3));
    Copy(A(3,2),B(2,3));
    Copy(A(3,3),B(3,3));
  end Transpose3;

  procedure SetRows3(U1,U2,U3: in Vector3; A: in out Matrix3) is
  begin
    Copy(U1(1),A(1,1));
    Copy(U1(2),A(1,2));
    Copy(U1(3),A(1,3));
    Copy(U2(1),A(2,1));
    Copy(U2(2),A(2,2));
    Copy(U2(3),A(2,3));
    Copy(U3(1),A(3,1));
    Copy(U3(2),A(3,2));
    Copy(U3(3),A(3,3));
  end SetRows3;

  procedure GetRows3(A: in Matrix3; U1,U2,U3: in out Vector3) is
  begin
    Copy(A(1,1),U1(1));
    Copy(A(1,2),U1(2));
    Copy(A(1,3),U1(3));
    Copy(A(2,1),U2(1));
    Copy(A(2,2),U2(2));
    Copy(A(2,3),U2(3));
    Copy(A(3,1),U3(1));
    Copy(A(3,2),U3(2));
    Copy(A(3,3),U3(3));
  end GetRows3;

  procedure SetColumns3(U1,U2,U3: in Vector3; A: in out Matrix3) is
  begin
    Copy(U1(1),A(1,1));
    Copy(U1(2),A(2,1));
    Copy(U1(3),A(3,1));
    Copy(U2(1),A(1,2));
    Copy(U2(2),A(2,2));
    Copy(U2(3),A(3,2));
    Copy(U3(1),A(1,3));
    Copy(U3(2),A(2,3));
    Copy(U3(3),A(3,3));
  end SetColumns3;

  procedure GetColumns3(A: in Matrix3; U1,U2,U3: in out Vector3) is
  begin
    Copy(A(1,1),U1(1));
    Copy(A(2,1),U1(2));
    Copy(A(3,1),U1(3));
    Copy(A(1,2),U2(1));
    Copy(A(2,2),U2(2));
    Copy(A(3,2),U2(3));
    Copy(A(1,3),U3(1));
    Copy(A(2,3),U3(2));
    Copy(A(3,3),U3(3));
  end GetColumns3;

  procedure Add3(A: in Matrix3; B: in out Matrix3) is
  begin
    Add(A(1,1),B(1,1));
    Add(A(1,2),B(1,2));
    Add(A(1,3),B(1,3));
    Add(A(2,1),B(2,1));
    Add(A(2,2),B(2,2));
    Add(A(2,3),B(2,3));
    Add(A(3,1),B(3,1));
    Add(A(3,2),B(3,2));
    Add(A(3,3),B(3,3));
  end Add3;

  procedure Sum3(A,B: in Matrix3; C: in out Matrix3) is
  begin
    Sum(A(1,1),B(1,1),C(1,1));
    Sum(A(1,2),B(1,2),C(1,2));
    Sum(A(1,3),B(1,3),C(1,3));
    Sum(A(2,1),B(2,1),C(2,1));
    Sum(A(2,2),B(2,2),C(2,2));
    Sum(A(2,3),B(2,3),C(2,3));
    Sum(A(3,1),B(3,1),C(3,1));
    Sum(A(3,2),B(3,2),C(3,2));
    Sum(A(3,3),B(3,3),C(3,3));
  end Sum3;

  procedure Sub3(A: in Matrix3; B: in out Matrix3) is
  begin
    Sub(A(1,1),B(1,1));
    Sub(A(1,2),B(1,2));
    Sub(A(1,3),B(1,3));
    Sub(A(2,1),B(2,1));
    Sub(A(2,2),B(2,2));
    Sub(A(2,3),B(2,3));
    Sub(A(3,1),B(3,1));
    Sub(A(3,2),B(3,2));
    Sub(A(3,3),B(3,3));
  end Sub3;

  procedure Diff3(A,B: in Matrix3; C: in out Matrix3) is
  begin
    Diff(A(1,1),B(1,1),C(1,1));
    Diff(A(1,2),B(1,2),C(1,2));
    Diff(A(1,3),B(1,3),C(1,3));
    Diff(A(2,1),B(2,1),C(2,1));
    Diff(A(2,2),B(2,2),C(2,2));
    Diff(A(2,3),B(2,3),C(2,3));
    Diff(A(3,1),B(3,1),C(3,1));
    Diff(A(3,2),B(3,2),C(3,2));
    Diff(A(3,3),B(3,3),C(3,3));
  end Diff3;

  procedure Mult3(S: in Scalar; A: in out Matrix3) is
  begin
    Mult(S,A(1,1));
    Mult(S,A(1,2));
    Mult(S,A(1,3));
    Mult(S,A(2,1));
    Mult(S,A(2,2));
    Mult(S,A(2,3));
    Mult(S,A(3,1));
    Mult(S,A(3,2));
    Mult(S,A(3,3));
  end Mult3;

  function MaxNorm(A: Matrix3) return Radius is
  begin
    return RMax(MaxNorm(A(1,1))+MaxNorm(A(2,1))+MaxNorm(A(3,1)),
                MaxNorm(A(1,2))+MaxNorm(A(2,2))+MaxNorm(A(3,2)),
                MaxNorm(A(1,3))+MaxNorm(A(2,3))+MaxNorm(A(3,3)));
  end MaxNorm;

  procedure Prod3(A: in Matrix3; U: in Vector3; V: in out Vector3) is
  begin
    -- V(1)
    Prod(A(1,1),U(1),V(1));
    AddProd(A(1,2),U(2),V(1));
    AddProd(A(1,3),U(3),V(1));
    -- V(2)
    Prod(A(2,1),U(1),V(2));
    AddProd(A(2,2),U(2),V(2));
    AddProd(A(2,3),U(3),V(2));
    -- V(3)
    Prod(A(3,1),U(1),V(3));
    AddProd(A(3,2),U(2),V(3));
    AddProd(A(3,3),U(3),V(3));
  end Prod3;

  procedure Prod3(A,B: in Matrix3; C: in out Matrix3) is
  begin
    -- C(1,1)
    Prod(A(1,1),B(1,1),C(1,1));
    AddProd(A(1,2),B(2,1),C(1,1));
    AddProd(A(1,3),B(3,1),C(1,1));
    -- C(1,2)
    Prod(A(1,1),B(1,2),C(1,2));
    AddProd(A(1,2),B(2,2),C(1,2));
    AddProd(A(1,3),B(3,2),C(1,2));
    -- C(1,3)
    Prod(A(1,1),B(1,3),C(1,3));
    AddProd(A(1,2),B(2,3),C(1,3));
    AddProd(A(1,3),B(3,3),C(1,3));
    -- C(2,1)
    Prod(A(2,1),B(1,1),C(2,1));
    AddProd(A(2,2),B(2,1),C(2,1));
    AddProd(A(2,3),B(3,1),C(2,1));
    -- C(2,2)
    Prod(A(2,1),B(1,2),C(2,2));
    AddProd(A(2,2),B(2,2),C(2,2));
    AddProd(A(2,3),B(3,2),C(2,2));
    -- C(2,3)
    Prod(A(2,1),B(1,3),C(2,3));
    AddProd(A(2,2),B(2,3),C(2,3));
    AddProd(A(2,3),B(3,3),C(2,3));
    -- C(3,1)
    Prod(A(3,1),B(1,1),C(3,1));
    AddProd(A(3,2),B(2,1),C(3,1));
    AddProd(A(3,3),B(3,1),C(3,1));
    -- C(3,2)
    Prod(A(3,1),B(1,2),C(3,2));
    AddProd(A(3,2),B(2,2),C(3,2));
    AddProd(A(3,3),B(3,2),C(3,2));
    -- C(3,3)
    Prod(A(3,1),B(1,3),C(3,3));
    AddProd(A(3,2),B(2,3),C(3,3));
    AddProd(A(3,3),B(3,3),C(3,3));
  end Prod3;

  procedure ColCross3(A: in Matrix3; U: in Vector3; B: in out Matrix3) is
  begin
    -- B(I,1)
    Prod(A(3,1),U(2),B(1,1));
    Neg(B(1,1));
    AddProd(A(2,1),U(3),B(1,1));
    Prod(A(1,1),U(3),B(2,1));
    Neg(B(2,1));
    AddProd(A(3,1),U(1),B(2,1));
    Prod(A(2,1),U(1),B(3,1));
    Neg(B(3,1));
    AddProd(A(1,1),U(2),B(3,1));
    -- B(I,2)
    Prod(A(3,2),U(2),B(1,2));
    Neg(B(1,2));
    AddProd(A(2,2),U(3),B(1,2));
    Prod(A(1,2),U(3),B(2,2));
    Neg(B(2,2));
    AddProd(A(3,2),U(1),B(2,2));
    Prod(A(2,2),U(1),B(3,2));
    Neg(B(3,2));
    AddProd(A(1,2),U(2),B(3,2));
    -- B(I,3)
    Prod(A(3,3),U(2),B(1,3));
    Neg(B(1,3));
    AddProd(A(2,3),U(3),B(1,3));
    Prod(A(1,3),U(3),B(2,3));
    Neg(B(2,3));
    AddProd(A(3,3),U(1),B(2,3));
    Prod(A(2,3),U(1),B(3,3));
    Neg(B(3,3));
    AddProd(A(1,3),U(2),B(3,3));
  end ColCross3;

  procedure CrossSandwich3(U: in Vector3; A: in out Matrix3) is
    --- AX = U x X x U
    S1: Scalar renames A(2,3);
    S2: Scalar renames A(3,1);
    S3: Scalar renames A(1,2);
  begin
    -- diag
    Prod(U(1),U(1),S1);
    Prod(U(2),U(2),S2);
    Prod(U(3),U(3),S3);
    Sum(S1,S2,A(3,3));
    Sum(S2,S3,A(1,1));
    Sum(S3,S1,A(2,2));
    -- off-diag
    Prod(U(1),U(2),S3); -- A(1,2)
    Neg(S3);
    Copy(S3,A(2,1));
    Prod(U(3),U(1),S2); -- A(3,1)
    Neg(S2);
    Copy(S2,A(1,3));
    Prod(U(2),U(3),S1); -- A(2,3)
    Neg(S1);
    Copy(S1,A(3,2));
  end CrossSandwich3;

  procedure Tr3(A: in Matrix3; S: in out Scalar) is
  begin
    Sum(A(1,1),A(2,2),S);
    Add(A(3,3),S);
  end Tr3;

  function Tr3(A: Matrix3) return Scalar is
    S: Scalar;
  begin
    Sum(A(1,1),A(2,2),S);
    Add(A(3,3),S);
    return S;
  end Tr3;

  procedure Det3(A: in Matrix3; S: in out Scalar) is
    Tmp: Scalar;
  begin
    Prod(A(1,1),A(2,2),Tmp);
    Prod(Tmp,A(3,3),S);
    Prod(A(1,2),A(2,3),Tmp);
    AddProd(Tmp,A(3,1),S);
    Prod(A(1,3),A(2,1),Tmp);
    AddProd(Tmp,A(3,2),S);
    Prod(A(1,3),A(2,2),Tmp);
    Neg(Tmp);
    AddProd(Tmp,A(3,1),S);
    Prod(A(1,1),A(2,3),Tmp);
    Neg(Tmp);
    AddProd(Tmp,A(3,2),S);
    Prod(A(1,2),A(2,1),Tmp);
    Neg(Tmp);
    AddProd(Tmp,A(3,3),S);
  end Det3;

  function Det3(A: Matrix3) return Scalar is
    S: Scalar;
  begin
    Det3(A,S);
    return S;
  end Det3;

  procedure Adj3(A: in Matrix3; B: in out Matrix3) is
    S1,S2: Scalar;
  begin
    Prod(A(2,2),A(3,3),S1); Prod(A(3,2),A(2,3),S2); Diff(S1,S2,B(1,1));
    Prod(A(3,2),A(1,3),S1); Prod(A(1,2),A(3,3),S2); Diff(S1,S2,B(1,2));
    Prod(A(1,2),A(2,3),S1); Prod(A(1,3),A(2,2),S2); Diff(S1,S2,B(1,3));
    Prod(A(2,3),A(3,1),S1); Prod(A(2,1),A(3,3),S2); Diff(S1,S2,B(2,1));
    Prod(A(1,1),A(3,3),S1); Prod(A(3,1),A(1,3),S2); Diff(S1,S2,B(2,2));
    Prod(A(2,1),A(1,3),S1); Prod(A(1,1),A(2,3),S2); Diff(S1,S2,B(2,3));
    Prod(A(2,1),A(3,2),S1); Prod(A(2,2),A(3,1),S2); Diff(S1,S2,B(3,1));
    Prod(A(3,1),A(1,2),S1); Prod(A(1,1),A(3,2),S2); Diff(S1,S2,B(3,2));
    Prod(A(1,1),A(2,2),S1); Prod(A(1,2),A(2,1),S2); Diff(S1,S2,B(3,3));
  end Adj3;

  procedure Inv3(A: in Matrix3; B: in out Matrix3) is
    D,InvD: Scalar;
  begin
    Adj3(A,B);
    Prod(A(1,1),B(1,1),D);
    AddProd(A(1,2),B(2,1),D);
    AddProd(A(1,3),B(3,1),D); --  det = a11*b11 + a12*b21 + a13*b31
    Inv(D,InvD);
    Mult3(InvD,B);
  end Inv3;

  function Inv3(A: Matrix3) return Matrix3 is
    B: Matrix3;
  begin
    Inv3(A,B);
    return B;
  end Inv3;

  -------------------- Newton-related stuff

  procedure Add(I: in Integer; S: in out Scalar) is
  begin
    Add(Scal(Flt(I)),S);
  end Add;

  procedure Add(I: in Integer; A: in out Matrix2) is
    S: constant Scalar := Scal(Flt(I));
  begin
    Add(S,A(1,1));
    Add(S,A(2,2));
  end Add;

  procedure BallAt0(R: in Flt; V: in out Vector2) is
  begin
    BallAt0(R,V(1));
    BallAt0(R,V(2));
  end BallAt0;

  procedure BallAt0(R: in Flt; A: in out Matrix2) is
  begin
    BallAt0(R,A(1,1));
    BallAt0(R,A(1,2));
    BallAt0(R,A(2,1));
    BallAt0(R,A(2,2));
  end BallAt0;

  procedure Copy(A: in Matrix2; B: in out Matrix2) renames Copy2;
  procedure Center(V: in out Vector2) renames Center2;
  procedure Center(A: in out Matrix2) renames Center2;
  procedure Neg(A: in out Matrix2) renames Neg2;
  procedure Add(U: in Vector2; V: in out Vector2) renames Add2;
  procedure Add(A: in Matrix2; B: in out Matrix2) renames Add2;
  procedure Sub(U: in Vector2; V: in out Vector2) renames Sub2;
  procedure Prod(A,B: in Matrix2; C: in out Matrix2) renames Prod2;
  procedure Quot(B: in Vector2; A: in Matrix2; X: in out Vector2) renames Quot2;
  procedure Quot(B,A: in Matrix2; X: in out Matrix2) renames Quot2;

  -------------------- Newton stuff

  package Newton1 is new Newton(Numeric => STrunc, Scalar => Scalar, Operator => Scalar);
  package Newton2 is new Newton(Numeric => STrunc, Scalar => Vector2, Operator => Matrix2);

  procedure Find01(S: in out Scalar; Steps: in Natural := 64; Verb: in Natural := 0) is
    use Globals, Newton1;
    procedure Find is new Find0 (TF => TF);
  begin
    Find(S,Steps,Verb);
  exception
    when No_Contraction => raise Not_Found;
  end Find01;

  procedure FindZero1(S: in out Scalar; Steps: in Natural := 64; Verb: in Natural := 0) is
    use Globals, Newton1;
    procedure Find is new FindZero (F => F, DF => DF);
  begin
    Find(S,Steps,Verb);
  exception
    when No_Contraction => raise Not_Found;
  end FindZero1;

  procedure Find02(V: in out Vector2; Steps: in Natural := 64; Verb: in Natural := 0) is
    use Globals, Newton2;
    procedure Find is new Find0 (TF => TF);
  begin
    Find(V,Steps,Verb);
  exception
    when No_Contraction => raise Not_Found;
  end Find02;

  procedure FindZero2(V: in out Vector2; Steps: in Natural := 64; Verb: in Natural := 0) is
    use Globals, Newton2;
    procedure Find is new FindZero (F => F, DF => DF);
  begin
    Find(V,Steps,Verb);
  exception
    when No_Contraction => raise Not_Found;
  end FindZero2;

  procedure QFindZero2(DerF: in Matrix2; R: in Radius; V: in out Vector2; Steps: in Natural := 64; Verb: in Natural := 0) is
    use Globals, Newton2;
    procedure DF(Dummy: in Vector2; A: in out Matrix2) is
      pragma Unreferenced(Dummy);
    begin
      Copy2(DerF,A);
    end DF;
    procedure Find is new FindZero (F => F, DF => DF);
    V0: Vector2;
  begin
    Copy2(V,V0);
    Find(V,Steps,Verb);
    if not STrunc then
      Sub2(V,V0);
      if MaxNorm(V0)>R then raise Domain_Error; end if;
    end if;
  exception
    when No_Contraction => raise Not_Found;
  end QFindZero2;

end Small_Vectors;
